/*
    Copyright (C) 1992-1994
                    Welson R. Jacometti

    All rights reserved.

    This program is  free software; you can redistribute unmodified
    versions of it as long as you do not  profit from that, neither
    by selling it nor  by   including it  with other products   for
    sale.  This copyright  notice  must  remain in all  distributed
    copies.

    You are  allowed to modify your  personal copy  of this program
    (as long as you do not distribute any changed versions).

    You may,  of course, distribute  along with it, your own source
    code or libraries of algorithms that run under GeoLab.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

    If you must contact the authors, you may write to:
                     geolab-request@dcc.unicamp.br
*/

/*
 * Este header tem como funcao facilitar o cadastro de tipos desconhecidos
 * pelo GeoLab. Um novo tipo pode ser criado a qualquer momento e passar
 * a ser usado desde que obedeca-se um conjunto reduzido de normas, que
 * sao:
 *
 *      1. Todo objeto geometrico e' completamente definido por duas
 *	   classes, uma para a implementacao da funcionalidade do objeto
 *	   com relacao apenas `as suas componentes geometricas e que
 *	   deriva da super-classe Obj (vide Obj.h) e outra para a 
 *	   incorporacao que possibilitem a manipulacao grafica do
 *	   objeto, derivada da super-classe GraphicObj (vide Obj.h)
 *
 *	2. Objetos externos nao podem, em condicoes normais, ser construidos
 *	   iterativamente em funcao da falta de conhecimento da interface
 *	   do "mode" adequado a ser selecionado para um objeto desconhecido.
 *	   Entretanto, atraves das facilidades "external_event_handler" 
 *	   disponiveis em World, tratadores externos especificos para
 *	   manipulacao iterativa de objetos desconhecidos podem ser
 *	   acrescentados atraves de algoritmos externos.
 *
 *	3. Com excecao dos metodos de criacao iterativa (vide 2), todos
 *	   os outros metodos virtuais herdados das super-classes devem
 *	   ser implementados para que, mesmo desconhecidos, possam ser
 *	   executadas algumas funcoes da interface sobre eles (isto inclui
 *	   armazenamento em memoria secundaria, descrito em maiores detalhes
 *	   em algum outro lugar).
 *	
 *	4. IMPORTANTE: A cada novo tipo introduzido, um identificados inteiro
 *	   deve ser a ele atribuido atraves deste modulo (isto melhora as
 *	   condicoes de consistencia destes valores). 
 *         
 *         Os valores de 0 a 1000 devem ser reservados para futuras incrementacoes 
 *         do ambiente. Outros valores podem ser usados livremente. Convenciona-se 
 *         que o nome do identificador deve ser formado pelo nome da classe 
 *         (assim como aparece apos a palavra reservada 'class`) seguida de 
 *         um 'underscore` e da letra 't`. Por exemplo:
 *
 *         class Point2D    ->   identificador Point2D_t
 *
 *      5. Com excecao a alguns objetos gerados (e conhecidos) pelo GeoLab,
 *         objetos que permitem representar variacoes de um mesmo objeto
 *         (como por exemplo Poligono, que pode ser generico, simples, em forma
 *         de estrela e convexo) devem ser caracterizados atraves de 'type` e 'subtype`.
 *         E' importante que os subtipos tambem sejam cadastrados, praticamente
 *         da mesma forma que os tipos, para que nao ocorram conflitos nos 
 *         algoritmos externos assim como no GeoLab. Por exemplo, tomemos o tipo
 *         'QuadEdge` que pode ser usado para representar um diagrama de Voronoi,
 *         uma triangulacao, ou uma subdivisao planar qualquer. No esquema
 *         proposto, o cadastro neste arquivo ficaria da seguinte forma:
 *
 *         const ObjType QuadEdge_t  = 5000;
 *         const ObjType QVononoi_t  = 5001;
 *         const ObjType QDelaunay_t = 5002;
 *          
 *         <...>
 *
 *         Observe que os subtipos QVoronoi_t e QDelaunay_t sao usados para
 *         informar que o que esta' representado e' um diagrama de Voronoi 
 *         (triangulacao de Delaunay) e que, alem disso, a estrutura usada e'
 *         uma QuadEdge. Isto e' muito importante, pois os algoritmos que vao
 *         manipular um diagrama de Voronoi podem exigir que ele seja dado
 *         atraves de uma QuadEdge e nao de outra estrutura de dados. Ate' que
 *         uma estrutura geral o suficiente surja, este sera' o mecanismo que
 *         possibilitara' que varias representacoes sejam usadas e intercambiadas,
 *         atraves de algoritmos externos.
 *
 */

// Tipos basicos mapeados em objetos geometricos
// para permitir trocas de informacoes mais genericas
// entre algoritmos externos e o ambiente

#ifndef _EXTERNALOBJ_ID_DEFINED
#define _EXTERNALOBJ_ID_DEFINED

const ObjType Message_t				= 1001;
const ObjType Boolean_t				= 1002;
const ObjType Integer_t				= 1003;
const ObjType Real_t				= 1004;
const ObjType QuadEdge_t			= 1005;
const ObjType DelaunayTriangulation_t		= 1006;
const ObjType VoronoiDiagram_t			= 1007;
const ObjType Arrow2D_t				= 1008;
const ObjType FarthestNeighborsTriangulation_t	= 1009;
const ObjType FarthestNeighborsVoronoiDiagram_t = 1010;
const ObjType Pslg_t                            = 1011;
const ObjType XString_t                         = 1024;
const ObjType SpanningTree_t			= 1025;
const ObjType MinimumSpanningTree_t		= 1026;

#endif ~_EXTERNALOBJ_ID_DEFINED;








