/*
    Copyright (C) 1992-1994
                    Welson R. Jacometti

    All rights reserved.

    This program is  free software; you can redistribute unmodified
    versions of it as long as you do not  profit from that, neither
    by selling it nor  by   including it  with other products   for
    sale.  This copyright  notice  must  remain in all  distributed
    copies.

    You are  allowed to modify your  personal copy  of this program
    (as long as you do not distribute any changed versions).

    You may,  of course, distribute  along with it, your own source
    code or libraries of algorithms that run under GeoLab.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

    If you must contact the authors, you may write to:
                     geolab-request@dcc.unicamp.br
*/

/*
 *
 * Square.h - definicao das classes Square e GraphicSquare
 *
 *            link flag -lm
 *
 *
 *    	Data	Hora  	Nome     		Modificacao/Motivo
 *
 *    07.01.92  13:40 	Welson R. Jacometti  	Arquivo criado
 *
 */

#ifndef _SQUARE_DEFINED
#define _SQUARE_DEFINED

#include <iostream.h>
#include <math.h>
#include "global.h"
#include "utilities.h"
#include "World.h"
#include "Obj.h"
#include "Rectangle.h"
#include "Point2D.h"

// forward declaration
class GraphicSquare;

class Square : public Rectangle {
  public:
    // construtores
    Square(char *s = "Square", ObjType t = Generic_t, bool v = false) : Rectangle(s, t, v) {}
    Square(double x1, double y1, double x2, double y2) : Rectangle("Square") { set(x1, y1, x2, y2); }
    Square( Point2D &a,  Point2D &b) : Rectangle("Square") { set(a, b); }
    Square( Point2D &p, double width, double height) : Rectangle("Square") { set(p, width, height); }
    Square( Square &s) : Rectangle(s.what_is(), s.volati, s.stype) { set(s.P1(), s.P2()); }
    
    //destrutor
    ~Square() {}				
    
    // construtor virtual (simulado) que estabelece o relacionamento
    //
    //                   Square --> GraphicSquare
    
    GraphicObj* new_GraphicObj();
    
    // funcao para auto-reproducao de um objeto - necessario para possibilitar 
    // que a interface manipule objetos externos ao geolab (alem disso torna
    // mais homogeneo o tratamento de objetos conhecidos e desconhecidos)

    Obj* reproduct()  {
    	return new Square(*this);
    }
        
    // query & set functions
    void setP1(double, double) { 
        warning("Square.h", "setP1", "not a valid call");
    }
    
    void setP2(double, double) {
        warning("Square.h", "setP2", "not a valid call");
    }
    
    void set(double x1, double y1, double x2, double y2) {
      if (zero(fabs(x2 - x1) - fabs(y2 - y1))) {
        p1.setXY(x1, y1);
        p2.setXY(x2, y2);
      }
    }

    void set(Point2D a, Point2D b) {
    	set(a.X(), a.Y(), b.X(), b.Y());
    }

    void set(Point2D p, double width, double /* height */) {
    	set(p.X(), p.Y(), p.X() + width, p.Y() + width);
    }
    
    // geometric functions

    bool is_square() { return true; }
        
    ObjType type()  { return Square_t; }
    
    void print()  { cout << "Square vertices : (" << p1.X() << "," << p1.Y() << ") (" << 
                                                          p2.X() << "," << p2.Y() << ") (" <<
                                                          p3.X() << "," << p3.Y() << ") (" << 
                                                          p4.X() << "," << p4.Y() << ")" << endl; }
};

/*
 *
 * GraphicSquare
 *
 *
 */

class GraphicSquare : public GraphicRectangle {
  public:
    // construtores
    GraphicSquare() : GraphicRectangle(new Square()) {}
    GraphicSquare(Square *s) : GraphicRectangle(s) {}
        
    //destrutor
    ~GraphicSquare() {}	

    void int_start(World *, double initial_X, double initial_Y, Int_Mode mode = LeftButton);    
    void int_change(World *, double new_X, double new_Y, Int_Mode mode = LeftButton);
    void int_end(World *, double final_X, double final_Y, Int_Mode mode = LeftButton);

    void resize_start(World *, double X, double Y, Int_Mode mode = LeftButton);    
};

typedef Square* Square_ptr;
typedef GraphicSquare* GraphicSquare_ptr;

#endif ~_SQUARE_DEFINED;



