/*
 * draw_polygon.h - Desenhador de Poligonos
 *
 *
 *    	Data	Hora  	Nome     		Modificacao/Motivo
 *
 *    13.03.93  17:10 	Cesar N. Gon 	  	Arquivo criado
 *
 */

#ifndef _DRAW_POLYGON_DEFINED
#define _DRAW_POLYGON_DEFINED

#include <GEOLAB/External.h>		// obrigatorio a todo modulo externo
#include <GEOLAB/Algorithm.h>
#include <GEOLAB/Point2D.h>
#include <GEOLAB/Segment2D.h>
#include <GEOLAB/Polygon.h>
#include <GEOLAB/line_settings.h>

icon_bits draw_polygon_icon[] = {
#include "draw_polygon.icon"
};

class draw_polygon : public ExternalObjectMode {
  public:

    ResultObjects binder(const SelectedObjects &, const GraphicObjects &, World *, int animation_depth = 0, int reference_depth = 0) {
	ResultObjects R;
	start = false;
	R.push(P);
	return R;
    }
    
    void activator(World *w) {
	start = false;
	edge1.set_all(w, FOREGROUND, BACKGROUND, 0, 0, false);
	edge2.set_all(w, FOREGROUND, BACKGROUND, 0, get_line_style_from_index(DASHED), false);
    }
    
    void deactivator(World *w) {	
	if (start) {
	    edge1.draw(w, XorMode);
	    edge2.draw(w, XorMode);
	    EO_exec_binder(w);
	}
    }

    bool click_select_down(World *w, double x, double y) {
	if (!start) {
	    P = new Polygon();
	    P->closed(true);
	    ((Segment2D *)edge2.obj)->set(x,y,x,y);
	    start = true;
	}
	else {
	    edge1.draw(w, XorMode);
	    edge2.draw(w, XorMode);
	    ((Segment2D *)edge1.obj)->setP2(x,y);
	    ((Segment2D *)edge2.obj)->setP1(x,y);
	    edge1.draw(w, XorMode);
	    edge2.draw(w, XorMode);
	}
	
	P->push(new Point2D(x,y));
	((Segment2D *)edge1.obj)->set(x,y,x,y);
	return true;
    }
    
    bool double_click(World *w, double x, double y) {
	edge1.draw(w, XorMode);
	edge2.draw(w, XorMode);
	if (start) 
	    EO_exec_binder(w);
	return true;
    }
    
    bool loc_move(World *w, double x, double y) {
	if (start) {
	    edge1.draw(w, XorMode);
	    edge2.draw(w, XorMode);
	    ((Segment2D *)edge1.obj)->setP2(x,y);
	    ((Segment2D *)edge2.obj)->setP1(x,y);
	    edge1.draw(w, XorMode);
	    edge2.draw(w, XorMode);
	}
	return true;
    }

    icon_bits *icon() {return draw_polygon_icon;}

  private:
    bool start;
    Polygon *P;
    GraphicSegment2D edge1, edge2;    
};

Algorithm* draw_polygon_link();

#endif ~_DRAW_POLYGON_DEFINED;




