/*
    Copyright (C) 1992-1994
                    Welson R. Jacometti

    All rights reserved.

    This program is  free software; you can redistribute unmodified
    versions of it as long as you do not  profit from that, neither
    by selling it nor  by   including it  with other products   for
    sale.  This copyright  notice  must  remain in all  distributed
    copies.

    You are  allowed to modify your  personal copy  of this program
    (as long as you do not distribute any changed versions).

    You may,  of course, distribute  along with it, your own source
    code or libraries of algorithms that run under GeoLab.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

    If you must contact the authors, you may write to:
                     geolab-request@dcc.unicamp.br
*/

/*
 *
 * Boolean.h
 *
 *    	Data	Hora  	Nome     		Modificacao/Motivo
 *
 *    20.01.92  13:40 	Welson R. Jacometti  	Arquivo criado
 *
 */

#ifndef _BOOLEAN_DEFINED
#define _BOOLEAN_DEFINED

#include <stdio.h>
#include <iostream.h>
#include "global.h"
#include "World.h"
#include "Obj.h"
#include "ExternalObj_id.h"	// obrigatorio em todo objeto externo

// forward declaration
class GraphicBoolean;
 
class Boolean : public Obj {
  public:
    // construtores
    Boolean(char *s = "Boolean", ObjType t = Generic_t, bool v = false) : Obj(s, t, v) { set(false); }
    Boolean(bool b) : Obj("Boolean") { set(b); }
    Boolean( Boolean &b) : Obj(b.what_is(), b.stype, b.volati) { set(b.get()); }
        
    //destrutor
    ~Boolean() {}

    // construtor virtual (simulado) que estabelece o relacionamento
    //
    //                   Boolean --> GraphicBoolean
    
    GraphicObj* new_GraphicObj();
    
    // funcao para auto-reproducao de um objeto - necessario para possibilitar 
    // que a interface manipule objetos externos ao geolab (alem disso torna
    // mais homogeneo o tratamento de objetos conhecidos e desconhecidos)

    Obj* reproduct()  {
    	return new Boolean(*this);
    }
        
    // query & set functions
    void set(bool b) { val = b; }
    bool get()  { return val; }
    
    // inherited methods redefinition
    ObjType type()  { return Boolean_t; }
    
    void print()  { cout << "Boolean: " << (val ? "true" : "false") << endl; }

    void draw(World *w, GC gc = 0) ;

  protected:
    bool val;
};

/*
 *
 * GraphicBoolean
 *
 *
 */

class GraphicBoolean : public GraphicObj {
  public:
    // construtores
    GraphicBoolean() : GraphicObj(new Boolean()) {}
    GraphicBoolean(Boolean *b) : GraphicObj(b) {}
    
    //destrutor
    ~GraphicBoolean() {}
    
    // inherited methods redefinition
    void draw(World *, DrawMode);
};

typedef Boolean* Boolean_ptr;
typedef GraphicBoolean* GraphicBoolean_ptr;

#endif ~_BOOLEAN_DEFINED;




