/*
    Copyright (C) 1992-1994
                    Welson R. Jacometti

    All rights reserved.

    This program is  free software; you can redistribute unmodified
    versions of it as long as you do not  profit from that, neither
    by selling it nor  by   including it  with other products   for
    sale.  This copyright  notice  must  remain in all  distributed
    copies.

    You are  allowed to modify your  personal copy  of this program
    (as long as you do not distribute any changed versions).

    You may,  of course, distribute  along with it, your own source
    code or libraries of algorithms that run under GeoLab.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

    If you must contact the authors, you may write to:
                     geolab-request@dcc.unicamp.br
*/

/*
 *
 * Coords2D.C - implementacao dos metodos e funcoes de apoio das classes
 *		PlaneCoords e PlaneMatrix.
 *  
 *              link flag (-lm)
 *
 *
 *    	Data	Hora  	Nome     		Modificacao/Motivo
 *
 *    31.10.91  13:45	Welson R. Jacometti	Arquivo Criado
 *
 */

#include "Coords2D.h"

// PlaneCoords methods

double PlaneCoords::operator * (PlaneCoords c) const {
    return coords[0]*c[0] + coords[1]*c[1] + coords[2]*c[2];
}

void PlaneCoords::print() {
    cout << "<" << coords[0] << " " << coords[1] << " " << coords[2] << ">" << endl;
}

void PlaneCoords::load(FILE *f) {
    fscanf(f, "%lf %lf %lf\n", &(coords[0]), &(coords[1]), &(coords[2]));
}

void PlaneCoords::save(FILE *f) {
    fprintf(f, "%lf %lf %lf\n", coords[0], coords[1], coords[2]);
}

// PlaneMatrix methods

PlaneMatrix PlaneMatrix::operator * (const PlaneMatrix &m) {
    int i, j, k;
    PlaneMatrix res;
    	    
    for(i=0; i < 3; i++)
      for(j=0; j < 3; j++) {
        res[i][j] = 0.0;
        for(k=0; k < 3; k++)
          res[i][j] += matrix[i][k] * m[k][j];
      }
	
    return res;				    
}

void PlaneMatrix::ident() {
    matrix[0][0] = 1; matrix[0][1] = 0; matrix[0][2] = 0;
    matrix[1][0] = 0; matrix[1][1] = 1; matrix[1][2] = 0;
    matrix[2][0] = 0; matrix[2][1] = 0; matrix[2][2] = 1;
}

void PlaneMatrix::trans(double tx, double ty) {
    PlaneMatrix mt;
    
    mt.maketrans(tx, ty);
    *this = *this * mt;
}

void PlaneMatrix::rotat(double angle) {
   PlaneMatrix mr;
   
   mr.makerotat(angle);
   *this = *this * mr;
}

void PlaneMatrix::print() {
    cout << "|" << matrix[0][0] << " " << matrix[0][1] << " " << matrix[0][2] << "|" << endl;
    cout << "|" << matrix[1][0] << " " << matrix[1][1] << " " << matrix[1][2] << "|" << endl;
    cout << "|" << matrix[2][0] << " " << matrix[2][1] << " " << matrix[2][2] << "|" << endl;
}

void PlaneMatrix::maketrans(double tx, double ty) {
    (*this).ident();
    matrix[2][0] = tx;
    matrix[2][1] = ty;
}

void PlaneMatrix::makerotat(double angle) {
    double co, si;
    co = cos(angle/180*PI);
    si = sin(angle/180*PI);

    (*this).ident();
    matrix[0][0] = co;
    matrix[0][1] = si;
    matrix[1][0] = -si;
    matrix[1][1] = co;
}

// tools

PlaneCoords cross(PlaneCoords &c1, PlaneCoords &c2) {
    PlaneCoords res;
    
    res[0] = c1[1]*c2[2] - c1[2]*c2[1];
    res[1] = c1[2]*c2[0] - c1[0]*c2[2];
    res[2] = c1[0]*c2[1] - c1[1]*c2[0];
        
    return res;
}

PlaneCoords multmatrix(PlaneCoords &c, PlaneMatrix &m) {
    int i, j;
    PlaneCoords res;
    
    for (i=0; i < 3; i++) {
      res[i] = 0;
      for (j=0; j < 3; j++)
        res[i] += c[j] * m[i][j];
    }
    
    return res;
}       


