/*
    Copyright (C) 1992-1994
                    Welson R. Jacometti

    All rights reserved.

    This program is  free software; you can redistribute unmodified
    versions of it as long as you do not  profit from that, neither
    by selling it nor  by   including it  with other products   for
    sale.  This copyright  notice  must  remain in all  distributed
    copies.

    You are  allowed to modify your  personal copy  of this program
    (as long as you do not distribute any changed versions).

    You may,  of course, distribute  along with it, your own source
    code or libraries of algorithms that run under GeoLab.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

    If you must contact the authors, you may write to:
                     geolab-request@dcc.unicamp.br
*/

/*
 *
 * Coords3D.C - implementacao dos metodos e funcoes de apoio das classes
 *		SpaceCoords e TransfMatrix.
 *  
 *              link flag (-lm)
 *
 *
 *    	Data	Hora  	Nome     		Modificacao/Motivo
 *
 *    31.10.91  13:45	Welson R. Jacometti	Arquivo Criado
 *
 */

#include "Coords3D.h"

// SpaceCoords methods

SpaceCoords::SpaceCoords( SpaceCoords &c) {
    *this = c;
}

SpaceCoords& SpaceCoords::operator = (SpaceCoords &c) {
    for (int i=0; i < 4; i++)
      coords[i] = c[i];
    
    return *this;
}

double SpaceCoords::operator * (SpaceCoords &c) {
    return coords[0]*c[0] + coords[1]*c[1] + coords[2]*c[2];
}

SpaceCoords SpaceCoords::operator + ( SpaceCoords &c) {
    SpaceCoords res;

    res[0] = coords[0] + c[0];
    res[1] = coords[1] + c[1];
    res[2] = coords[2] + c[2];
    res[3] = coords[3];
    
    return res;
}

SpaceCoords SpaceCoords::operator - ( SpaceCoords &c)  {
    SpaceCoords res;

    res[0] = coords[0] - c[0];
    res[1] = coords[1] - c[1];
    res[2] = coords[2] - c[2];
    res[3] = coords[3];
        
    return res;
}

double SpaceCoords::normalize() {
    double d;
    
    d = sqrt(coords[0]*coords[0] + coords[1]*coords[1] + coords[2]*coords[2]);
    coords[0] /= d;
    coords[1] /= d;
    coords[2] /= d;
    
    return d;
}

int SpaceCoords::drop() {
    double x, y, z;
  
    x = fabs(coords[0]);
    y = fabs(coords[1]);
    z = fabs(coords[2]);
  
    if (x > y)
      if (x > z)
        return Space_XCoord;
      else
        return Space_ZCoord;
    else 
      if (y > z)
        return Space_YCoord;
      else
        return Space_ZCoord;
}

ostream & operator<<(ostream &os, SpaceCoords &sc) {
  return os << "<" << sc.coords[0] << " " << sc.coords[1] << " " 
                   << sc.coords[2] << " " << sc.coords[3] << ">" << endl;
}


// TransfMatrix methods

TransfMatrix::TransfMatrix( TransfMatrix &m) {
    *this = m;
}

TransfMatrix& TransfMatrix::operator = ( TransfMatrix &m) {
    for (int i=0; i < 4; i++)
      for (int j=0; j < 4; j++)
        matrix[i][j] = m[i][j];
    
    return *this;
}

TransfMatrix TransfMatrix::operator * ( TransfMatrix &m) {
    int i, j, k;
    TransfMatrix res;
    	    
    for(i=0; i < 4; i++)
      for(j=0; j < 4; j++) {
        res[i][j] = 0.0;
        for(k=0; k < 4; k++)
          res[i][j] += matrix[i][k] * m[k][j];
      }
	
    return res;				    
}

void TransfMatrix::ident() {
    matrix[0][0] = 1; matrix[0][1] = 0; matrix[0][2] = 0; matrix[0][3] = 0;
    matrix[1][0] = 0; matrix[1][1] = 1; matrix[1][2] = 0; matrix[1][3] = 0;
    matrix[2][0] = 0; matrix[2][1] = 0; matrix[2][2] = 1; matrix[2][3] = 0;
    matrix[3][0] = 0; matrix[3][1] = 0; matrix[3][2] = 0; matrix[3][3] = 1;
}

void TransfMatrix::trans(double tx, double ty, double tz) {
    TransfMatrix mt;
    
    mt.maketrans(tx, ty, tz);
    *this = *this * mt;
}

void TransfMatrix::rotat(double rx, double ry, double rz) {
   TransfMatrix mr;
   
   mr.makerotat(rx, ry, rz);
   *this = *this * mr;
}

ostream & operator<<(ostream &os, TransfMatrix &tm) {
    os << "<<" << tm.matrix[0][0] << " " << tm.matrix[0][1] << " "
	       << tm.matrix[0][2] << " " << tm.matrix[0][3] << ">" << endl;
    os << " <" << tm.matrix[1][0] << " " << tm.matrix[1][1] << " "
	       << tm.matrix[1][2] << " " << tm.matrix[1][3] << ">" << endl;
    os << " <" << tm.matrix[2][0] << " " << tm.matrix[2][1] << " "
	       << tm.matrix[2][2] << " " << tm.matrix[2][3] << ">" << endl;
    os << " <" << tm.matrix[3][0] << " " << tm.matrix[3][1] << " "
	       << tm.matrix[3][2] << " " << tm.matrix[3][3] << ">>" << endl;
    return os;
}

void TransfMatrix::maketrans(double tx, double ty, double tz) {
    (*this).ident();
    matrix[3][0] = tx;
    matrix[3][1] = ty;
    matrix[3][2] = tz;
}

void TransfMatrix::makerotat(double rx, double ry, double rz) {
    TransfMatrix mx, my, mz;
    
    (*this).ident();
    mx.makexrotat(rx);
    my.makeyrotat(ry);
    mz.makezrotat(rz);
    
    *this = *this * mx;
    *this = *this * my;
    *this = *this * mz;
}

void TransfMatrix::makexrotat(double rx) {
    double co, si;
    
    co = cos(rx/180*PI);
    si = sin(rx/180*PI);
    (*this).ident();
    
    matrix[1][1] = co;
    matrix[1][2] = si;
    matrix[2][1] = -si;
    matrix[2][2] = co;
}

void TransfMatrix::makeyrotat(double ry) {
    double co, si;
    
    co = cos(ry/180*PI);
    si = sin(ry/180*PI);
    (*this).ident();
    
    matrix[2][2] = co;
    matrix[2][0] = si;
    matrix[0][2] = -si;
    matrix[0][0] = co;
}

void TransfMatrix::makezrotat(double rz) {
    double co, si;
    
    co = cos(rz/180*PI);
    si = sin(rz/180*PI);
    (*this).ident();
    
    matrix[0][0] = co;
    matrix[0][1] = si;
    matrix[1][0] = -si;
    matrix[1][1] = co;
}

// tools

bool null( SpaceCoords &c, double tol) {
    if (zero(c[0], tol) && zero(c[1], tol) && zero(c[2], tol))
      return true;
    
    return false;
}

bool equal( SpaceCoords &c1,  SpaceCoords &c2, double tol) {
    SpaceCoords tmp;
    
    tmp = c1 - c2;
    return null(tmp, tol);
}

SpaceCoords cross( SpaceCoords &c1,  SpaceCoords &c2) {
    SpaceCoords res;
    
    res[0] = c1[1]*c2[2] - c1[2]*c2[1];
    res[1] = c1[2]*c2[0] - c1[0]*c2[2];
    res[2] = c1[0]*c2[1] - c1[1]*c2[0];
    res[3] = c1[3];
        
    return res;
}

SpaceCoords multmatrix( SpaceCoords &c,  TransfMatrix &m) {
    int i, j;
    SpaceCoords res;
    
    for (i=0; i < 4; i++) {
      res[i] = 0;
      for (j=0; j < 4; j++)
        res[i] += c[j] * m[i][j];
    }
    
    return res;
}       



    



