/*
    Copyright (C) 1992-1994
                    Welson R. Jacometti

    All rights reserved.

    This program is  free software; you can redistribute unmodified
    versions of it as long as you do not  profit from that, neither
    by selling it nor  by   including it  with other products   for
    sale.  This copyright  notice  must  remain in all  distributed
    copies.

    You are  allowed to modify your  personal copy  of this program
    (as long as you do not distribute any changed versions).

    You may,  of course, distribute  along with it, your own source
    code or libraries of algorithms that run under GeoLab.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

    If you must contact the authors, you may write to:
                     geolab-request@dcc.unicamp.br
*/

/*
 * Diameter.h - algoritmos para o calculo do diametro de um conjunto de
 *              pontos, de poligonos convexos, em forma de estrela e simples.
 * 
 *
 *    	Data	Hora  	Nome     		Modificacao/Motivo
 *
 *    12.05.92  09:30 	Welson R. Jacometti  	Arquivo criado
 *
 */

#ifndef _DIAMETER_DEFINED
#define _DIAMETER_DEFINED
 
#include "External.h"		// obrigatorio a todo modulo externo
#include "Algorithm.h"
#include "Point2D.h"
#include "Segment2D.h"
#include "Polygon.h"

#ifndef _Point2D_ptr_list
#define _Point2D_ptr_list
declare(list,Point2D_ptr)
#endif ~_Point2D_ptr_list;

#ifndef _Segment2D_ptr_list
#define _Segment2D_ptr_list
declare(list,Segment2D_ptr)
#endif ~_Segment2D_ptr_list;

/*
 * Pares antipodais de um conjunto de um poligono convexo.
 *
 */
 
list(Segment2D_ptr) antipodal_pairs(list(Point2D_ptr), World *w = 0, int animation_depth = 0, int reference_depth = 0);

Segment2D_ptr diameter_convex_polygon(Polygon_ptr pol, World *w = 0, int animation_depth = 0, int reference_depth = 0);
Segment2D_ptr diameter_star_shaped_polygon(Polygon_ptr pol, World *w = 0, int animation_depth = 0, int reference_depth = 0);
Segment2D_ptr diameter_simple_polygon(Polygon_ptr pol, World *w = 0, int animation_depth = 0, int reference_depth = 0);
Segment2D_ptr diameter_generic_polygon(Polygon_ptr pol, World *w = 0, int animation_depth = 0, int reference_depth = 0);
Segment2D_ptr diameter_set_of_points(list(Point2D_ptr) &points, World *w = 0, int animation_depth = 0, int reference_depth = 0);
Segment2D_ptr diameter_set_of_points_bf(list(Point2D_ptr) &points, World *w = 0, int animation_depth = 0, int reference_depth = 0);

// Modelagem dos algoritmos

class SetOfPoints : public Algorithm {
    bool is_able_to_handle(const SelectedObjects &, const GraphicObjects &);
};

class InputIsConvexPolygon : public Algorithm {
    bool is_able_to_handle(const SelectedObjects &, const GraphicObjects &);
};

class AntipodalPairs : public InputIsConvexPolygon {
    bool animation() { return true; }
    
    ResultObjects binder(const SelectedObjects &, const GraphicObjects &, World *w = 0, int animation_depth = 0, int reference_depth = 0);
};

class DiameterConvexPolygon : public InputIsConvexPolygon {
    bool animation() { return true; }

    ResultObjects binder(const SelectedObjects &, const GraphicObjects &, World *w = 0, int animation_depth = 0, int reference_depth = 0);
};
    
class DiameterStarShapedPolygon : public Algorithm {
    bool animation() { return true; }

    bool is_able_to_handle(const SelectedObjects &, const GraphicObjects &);

    ResultObjects binder(const SelectedObjects &, const GraphicObjects &, World *w = 0, int animation_depth = 0, int reference_depth = 0);
};
    
class DiameterSimplePolygon : public Algorithm {
    bool animation() { return true; }

    bool is_able_to_handle(const SelectedObjects &, const GraphicObjects &);

    ResultObjects binder(const SelectedObjects &, const GraphicObjects &, World *w = 0, int animation_depth = 0, int reference_depth = 0);
};
    
class DiameterGenericPolygon : public Algorithm {
    bool animation() { return true; }

    bool is_able_to_handle(const SelectedObjects &, const GraphicObjects &);

    ResultObjects binder(const SelectedObjects &, const GraphicObjects &, World *w = 0, int animation_depth = 0, int reference_depth = 0);
};

class DiameterSetOfPoints : public SetOfPoints {
    bool animation() { return true; }

    ResultObjects binder(const SelectedObjects &, const GraphicObjects &, World *w = 0, int animation_depth = 0, int reference_depth = 0);
};
    
class DiameterSetOfPointsBF : public SetOfPoints {
    bool animation() { return true; }

    ResultObjects binder(const SelectedObjects &, const GraphicObjects &, World *w = 0, int animation_depth = 0, int reference_depth = 0);
};
    
Algorithm* antipodal_pairs_link();
Algorithm* diameter_convex_polygon_link();
Algorithm* diameter_star_shaped_polygon_link();
Algorithm* diameter_simple_polygon_link();
Algorithm* diameter_generic_polygon_link();
Algorithm* diameter_set_of_points_link();
Algorithm* diameter_set_of_points_bf_link();

#endif ~_DIAMETER_DEFINED;









