/*
    Copyright (C) 1992-1994
                    Welson R. Jacometti

    All rights reserved.

    This program is  free software; you can redistribute unmodified
    versions of it as long as you do not  profit from that, neither
    by selling it nor  by   including it  with other products   for
    sale.  This copyright  notice  must  remain in all  distributed
    copies.

    You are  allowed to modify your  personal copy  of this program
    (as long as you do not distribute any changed versions).

    You may,  of course, distribute  along with it, your own source
    code or libraries of algorithms that run under GeoLab.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

    If you must contact the authors, you may write to:
                     geolab-request@dcc.unicamp.br
*/

/*
 *
 * Ellipse.C - Ellipse & GraphicEllipse methods.
 *
 *
 *    	Data	Hora  	Nome     		Modificacao/Motivo
 *
 *    07.01.92  13:40 	Welson R. Jacometti  	Arquivo criado
 *
 */

#include "Obj.h"
#include "Ellipse.h"
#include "color.h"

GraphicObj* Ellipse::new_GraphicObj() {
    return new GraphicEllipse(this);
}

// Metodos necessarios para "Input Tools"
    
bool Ellipse::point_inside(double x, double y)  {
    Point2D f1=F1(), f2=F2();
    double a;

    if (!zero(f1.X()-f2.X()))
      a = fabs(p2.X()-p1.X());
    else
      a = fabs(p2.Y()-p1.Y());

    return less (f1.distance_to_point(x,y)+f2.distance_to_point(x,y), a);
}
     
void Ellipse::bounding_box(double& x1, double& y1, double& x2, double& y2)  {
    set_handles();
    x1=h1.X();
    y1=h1.Y();
    x2=h2.X();
    y2=h2.Y();
}

list(Obj_ptr) Ellipse::get_segments()  {
    list(Obj_ptr) S;
    Segment2D_ptr s_ptr;
    set_handles();
    s_ptr = new Segment2D(h1,h3);S.push (s_ptr);assert(s_ptr);
    s_ptr = new Segment2D(h1,h4);S.push (s_ptr);assert(s_ptr);
    s_ptr = new Segment2D(h2,h3);S.push (s_ptr);assert(s_ptr);
    s_ptr = new Segment2D(h2,h4);S.push (s_ptr);assert(s_ptr);
    return S;   
}

double calc_C(double a, double b) {
  return sqrt(fabs((a * a) - (b * b)));
}

Point2D Ellipse::F1()  {
    Point2D res;
    double w = fabs(p1.X() - p2.X());
    double h = fabs(p1.Y() - p2.Y()); 
    if (w == 0 || h == 0)
      res.set_at_infinity(0, 0);
    else
      if (w < h) 
        res.setXY((p1.X() + p2.X())/2, ((p1.Y() + p2.Y())/2) + calc_C(w/2, h/2));
      else
        res.setXY(((p1.X() + p2.X())/2) - calc_C(w/2, h/2), (p1.Y() + p2.Y())/2);
    return res;
}

Point2D Ellipse::F2()  {
    Point2D res;
    double w = fabs(p1.X() - p2.X());
    double h = fabs(p1.Y() - p2.Y()); 
    if (w == 0 || h == 0)
      res.set_at_infinity(0, 0);
    else
      if (w < h) 
        res.setXY((p1.X() + p2.X())/2, ((p1.Y() + p2.Y())/2) - calc_C(w/2, h/2));
      else
        res.setXY(((p1.X() + p2.X())/2) + calc_C(w/2, h/2), (p1.Y() + p2.Y())/2);
    return res;
}

void Ellipse::unfill_handles(World *w, bool show_handle)  {
    if (show_handle) {
      set_handles();
      h1.unfill_handles(w);
      h2.unfill_handles(w);
      h3.unfill_handles(w);
      h4.unfill_handles(w);
    }
}

void Ellipse::draw_handles(World *w, GC gc, bool highlight, bool show_handle)  {
    Point2D focus1 = F1();
    Point2D focus2 = F2();
    
    if (show_handle) {
      set_handles();
      h1.draw_handles(w, gc, highlight);
      h2.draw_handles(w, gc, highlight);
      h3.draw_handles(w, gc, highlight);
      h4.draw_handles(w, gc, highlight);
    
      // pequenina GAMBI para evitar distorcoes no desenho em 
      // funcao do sistema de coordenadas
    
      int f1x = w->xcoord(focus1.X());
      int f1y = w->ycoord(focus1.Y());
      int f2x = w->xcoord(focus2.X());
      int f2y = w->ycoord(focus2.Y());
    
      w->DrawLine(gc, w->virtual_xcoord(f1x - 2), w->virtual_ycoord(f1y), 
                      w->virtual_xcoord(f1x + 2), w->virtual_ycoord(f1y));
      w->DrawLine(gc, w->virtual_xcoord(f1x), w->virtual_ycoord(f1y - 2), 
                      w->virtual_xcoord(f1x), w->virtual_ycoord(f1y + 2));
      w->DrawLine(gc, w->virtual_xcoord(f2x - 2), w->virtual_ycoord(f2y), 
                      w->virtual_xcoord(f2x + 2), w->virtual_ycoord(f2y));
      w->DrawLine(gc, w->virtual_xcoord(f2x), w->virtual_ycoord(f2y - 2), 
                      w->virtual_xcoord(f2x), w->virtual_ycoord(f2y + 2));
    }
}

void Ellipse::draw(World *w, GC gc)  {
    w->DrawArc(gc, P1().X(), P1().Y(), P2().X(), P2().Y());
}

void Y_from_equation(double x, double x0, double y0, double a, double b, double &y1, double &y2) {
    double a2 = a * a;
    double b2 = b * b;
    double xx02 = (x - x0) * (x - x0);
    double y02 = y0 * y0;
    
    double c1 = (b2 * (1 - (xx02 / a2))) - y02;
    double c2 = 2 * y0;
    double sqrtd = sqrt((c2 * c2) + (4*c1));
    
    y1 = (c2 - sqrtd) / 2;
    y2 = (c2 + sqrtd) / 2;
}

void X_from_equation(double y, double x0, double y0, double a, double b, double &x1, double &x2) {
    double a2 = a * a;
    double b2 = b * b;
    double yy02 = (y - y0) * (y - y0);
    double x02 = x0 * x0;
    
    double c1 = (a2 * (1 - (yy02 / b2))) - x02;
    double c2 = 2 * x0;
    double sqrtd = sqrt((c2 * c2) + (4*c1));
    
    x1 = (c2 - sqrtd) / 2;
    x2 = (c2 + sqrtd) / 2;
}

double Ellipse::distance_to_point(double x, double y)  {
    Point2D q1, q2, q3, q4;
    double w = fabs(p1.X() - p2.X());
    double h = fabs(p1.Y() - p2.Y());

    q1.setXY(p1.X(), (p1.Y() + p2.Y()) / 2);
    q2.setXY(p2.X(), (p1.Y() + p2.Y()) / 2);
    q3.setXY((p1.X() + p2.X()) / 2, p1.Y()); 
    q4.setXY((p1.X() + p2.X()) / 2, p2.Y());

    if (w != 0 && h != 0) {
      double a = w / 2;		// calculo por aproximacao (divisoes sucessivas)
      double b = h / 2;
      double x0 = q3.X();
      double y0 = q1.Y();
      double dmin = TOO_FAR;
      Point2D p(x, y);		// para maior precisao, consideramos os dois casos
      if (w > h) {
        double x1, x2, y11, y12;
        double x14, x24, x34, cdmin1, cdmin2, cdmin3;
      	if (q1.e_distance_sqr(p) < q2.e_distance_sqr(p))
      	  x1 = q1.X();
      	else
      	  x1 = q2.X();
      	x2 = x0;
      	while (fabs(x2 - x1) > EPSLON) {
      	  x24 = (x1 + x2) / 2;
      	  x14 = (x1 + x24) / 2;
      	  x34 = (x24 + x2) / 2;
      	  Y_from_equation(x14, x0, y0, a, b, y11, y12);
      	  cdmin1 = min(p.e_distance(Point2D(x14, y11)), p.e_distance(Point2D(x14, y12)));
      	  Y_from_equation(x24, x0, y0, a, b, y11, y12);
      	  cdmin2 = min(p.e_distance(Point2D(x24, y11)), p.e_distance(Point2D(x24, y12)));
      	  Y_from_equation(x34, x0, y0, a, b, y11, y12);
      	  cdmin3 = min(p.e_distance(Point2D(x34, y11)), p.e_distance(Point2D(x34, y12)));
      	  if (cdmin2 < dmin)
      	    dmin = cdmin2;
      	  if (cdmin1 < cdmin3)
      	    x2 = x24;
      	  else
      	    x1 = x24;
      	}
      }
      else {
        double y1, y2, x11, x12;
        double y14, y24, y34, cdmin1, cdmin2, cdmin3;
      	if (q3.e_distance_sqr(p) < q4.e_distance_sqr(p))
      	  y1 = q3.Y();
      	else
      	  y1 = q4.Y();
      	y2 = y0;
      	while (fabs(y2 - y1) > EPSLON) {
      	  y24 = (y1 + y2) / 2;
      	  y14 = (y1 + y24) / 2;
      	  y34 = (y24 + y2) / 2;
      	  X_from_equation(y14, x0, y0, a, b, x11, x12);
      	  cdmin1 = min(p.e_distance(Point2D(x11, y14)), p.e_distance(Point2D(x12, y14)));
      	  X_from_equation(y24, x0, y0, a, b, x11, x12);
      	  cdmin2 = min(p.e_distance(Point2D(x11, y24)), p.e_distance(Point2D(x12, y24)));
      	  X_from_equation(y34, x0, y0, a, b, x11, x12);
      	  cdmin3 = min(p.e_distance(Point2D(x11, y34)), p.e_distance(Point2D(x12, y34)));
      	  if (cdmin2 < dmin)
      	    dmin = cdmin2;
      	  if (cdmin1 < cdmin3)
      	    y2 = y24;
      	  else
      	    y1 = y24;
      	}
      }
      return dmin;
    }
    return TOO_FAR;
}

// Methods for information retrieval
  
list(Obj_ptr) Ellipse::get_points()  {
      list(Obj_ptr) Objp;
      Point2D_ptr p1,p2,p3,p4;
      set_handles();
      p1 = new Point2D(h1); assert(p1);
      p2 = new Point2D(h2); assert(p2);
      p3 = new Point2D(h3); assert(p3);
      p4 = new Point2D(h4); assert(p4);
      
      Objp.append(p1);
      Objp.append(p2);
      Objp.append(p3);
      Objp.append(p4);
      
      return Objp;
}

/*
 *
 *
 */

// cast macro

#define ellipse	((Ellipse *)obj)

static GraphicRectangle gr;

void GraphicEllipse::int_start(World *w, double initial_X, double initial_Y, Int_Mode mode) {
    gr.int_start(w, initial_X, initial_Y, mode);
    if (mode == LeftButton) {
      gr.set_all(w, get_pixel_from_value(BLACK), get_pixel_from_value(WHITE), 0, 0, false);
      gr.draw(w, XorMode);
    }
    ellipse->set(*(Rectangle *)gr.obj);
}

void GraphicEllipse::int_change(World *w, double new_X, double new_Y, Int_Mode mode) {
    if (mode == LeftButton)
      gr.draw(w, XorMode);
    gr.int_change(w, new_X, new_Y, mode);
    if (mode == LeftButton)
      gr.draw(w, XorMode);
    ellipse->set(*(Rectangle *)gr.obj); 
}

void GraphicEllipse::int_end(World *w, double final_X, double final_Y, Int_Mode mode) {
    if (mode == LeftButton)
      gr.draw(w, XorMode);
    gr.int_change(w, final_X, final_Y, mode);
    ellipse->set(*(Rectangle *)gr.obj); 
}

void GraphicEllipse::resize_start(World *w, double X, double Y, Int_Mode mode) {
    ((Rectangle *)gr.obj)->set(ellipse->P1(), ellipse->P2());
    gr.resize_start(w, X, Y, mode);
    if (mode == LeftButton) {
      gr.set_all(w, get_pixel_from_value(BLACK), get_pixel_from_value(WHITE), 0, 0, false);
      gr.draw(w, XorMode);
    }
}

void GraphicEllipse::resize_change(World *w, double X, double Y, Int_Mode mode) {
    int_change(w, X, Y, mode);
}

void GraphicEllipse::resize_end(World *w, double X, double Y, Int_Mode mode) {
    int_end(w, X, Y, mode);
}

bool GraphicEllipse::where_draw_string (double &x, double &y) {
    Point2D p1 (((Ellipse *)obj)->F1());
    Point2D p2 (((Ellipse *)obj)->F2());
    x = (p1.X()+p2.X())/2;
    y = (p1.Y()+p2.Y())/2;
    return true;    
}

