/*
    Copyright (C) 1992-1994
                    Welson R. Jacometti

    All rights reserved.

    This program is  free software; you can redistribute unmodified
    versions of it as long as you do not  profit from that, neither
    by selling it nor  by   including it  with other products   for
    sale.  This copyright  notice  must  remain in all  distributed
    copies.

    You are  allowed to modify your  personal copy  of this program
    (as long as you do not distribute any changed versions).

    You may,  of course, distribute  along with it, your own source
    code or libraries of algorithms that run under GeoLab.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

    If you must contact the authors, you may write to:
                     geolab-request@dcc.unicamp.br
*/

/*
 *
 * Ellipse.h - definicao das classes Ellipse e GraphicEllipse.
 *
 *             link flag -lm
 *
 *
 *    	Data	Hora  	Nome     		Modificacao/Motivo
 *
 *    07.01.92  13:40 	Welson R. Jacometti  	Arquivo criado
 *
 */

#ifndef _ELLIPSE_DEFINED
#define _ELLIPSE_DEFINED

#include <iostream.h>
#include <stdio.h>
#include <math.h>
#include "global.h"
#include "World.h"
#include "Obj.h"
#include "Point2D.h"
#include "Rectangle.h"

// forward declaration
class GraphicEllipse;

class Ellipse : public Obj {
  public:
    // construtores
    Ellipse(char *s = "Ellipse", ObjType t = Generic_t, bool v = false) : Obj(s, t, v) { set(0, 0, 0, 0); }
    Ellipse(double x1, double y1, double x2, double y2) : Obj("Ellipse") { set(x1, y1, x2, y2); }
    Ellipse( Point2D &p1,  Point2D &p2) : Obj("Ellipse") { set(p1, p2); }
    Ellipse( Ellipse &e) : Obj(e.what_is(), e.stype, e.volati) { set(e.P1(), e.P2()); }    
    
    //destrutor
    ~Ellipse() {}				
    
    // construtor virtual (simulado) que estabelece o relacionamento
    //
    //                 Ellipse --> GraphicEllipse
    
    GraphicObj* new_GraphicObj();
    
    // funcao para auto-reproducao de um objeto - necessario para possibilitar 
    // que a interface manipule objetos externos ao geolab (alem disso torna
    // mais homogeneo o tratamento de objetos conhecidos e desconhecidos)

    Obj* reproduct()  {
    	return new Ellipse(*this);
    }

    // Metodos necessarios para "Input Tools"
    
    virtual bool point_inside_implemented () {return true;}
    virtual bool point_inside(double x, double y) ;
    virtual bool bounding_box_implemented () {return true;}
    virtual void bounding_box (double& x1, double& y1, double& x2, double& y2) ;
    virtual bool get_segments_implemented () {return true;}
    virtual list(Obj_ptr) get_segments () ;
        
    // query & set functions
    Point2D& P1()  { return p1; }
    Point2D& P2()  { return p2; }
    
    // ellipse focus
    Point2D F1() ;
    Point2D F2() ;
    
    virtual void set(double x1, double y1, double x2, double y2) {
    	p1.setXY(x1, y1);
    	p2.setXY(x2, y2);
    }
    
    virtual void set( Point2D p1,  Point2D p2) {
        set(p1.X(), p1.Y(), p2.X(), p2.Y());
    }

    virtual void set( Rectangle r) {
    	set(r.P1(), r.P2());
    }

    // inherited methods redefinition
    void load(FILE *f) { Obj::load(f); p1.load(f); p2.load(f); }
    void save(FILE *f) { Obj::save(f); p1.save(f); p2.save(f); }

    ObjType type()  { return Ellipse_t; }
    
    void translate(double to_x, double to_y) {
    	p1.translate(to_x, to_y);
    	p2.translate(to_x, to_y);
    }

    void rotate(double orig_x, double orig_y, double angle) {
	p1.rotate(orig_x, orig_y, angle);
	p2.rotate(orig_x, orig_y, angle);
    }

    
    void print()  { cout << "Ellipse bounding box : (" << p1.X() << "," << p1.Y() << ") (" << 
                                                               p2.X() << "," << p2.Y() << ")" << endl; }

    void unfill_handles(World *w, bool show_handle = true) ;
    void draw_handles(World *w, GC gc, bool highlight = false, bool show_handle = true) ;
    void draw(World *w, GC gc) ;

    bool select_test(double x, double y)  {
    	set_handles();
    	return h1.select_test(x, y) || h2.select_test(x, y) || h3.select_test(x, y) || h4.select_test(x, y); 
    }
    
    bool select_test(double x1, double y1, double x2, double y2)  {
    	set_handles();
    	return h1.select_test(x1, y1, x2, y2) && h2.select_test(x1, y1, x2, y2) && h3.select_test(x1, y1, x2, y2) && h4.select_test(x1, y1, x2, y2); 
    }

    double distance_to_point(double x, double y) ;


    // Methods for information retrieval.
    virtual bool get_points_implemented() { return true; }
    virtual list(Obj_ptr) get_points() ;      
        	
  protected:
    void set_handles()  {
        h1.setXY(p1.X(), p1.Y());
        h2.setXY(p2.X(), p2.Y());
        h3.setXY(p1.X(), p2.Y());
        h4.setXY(p2.X(), p1.Y());
    }
    
    Point2D p1, p2;		// bounding box
    Point2D h1, h2, h3, h4;	// handles, instanciados aqui por 
    				// questoes de eficiencia (tempo)
};

/*
 *
 * GraphicEllipse.h
 *
 *
 */

class GraphicEllipse : public GraphicObj {
  public:
    //construtores
    GraphicEllipse() : GraphicObj(new Ellipse()) {}
    GraphicEllipse(Ellipse *e) : GraphicObj(e) {}
    
    //destrutor
    ~GraphicEllipse() {}				
    
    // inherited methods redefinition
    void int_start(World *, double initial_X, double initial_Y, Int_Mode mode = LeftButton);
    void int_change(World *, double new_X, double new_Y, Int_Mode mode = LeftButton);
    void int_end(World *, double final_X, double final_Y, Int_Mode mode = LeftButton);
    
    void resize_start(World *, double X, double Y, Int_Mode mode = LeftButton);
    void resize_change(World *, double X, double Y, Int_Mode mode = LeftButton);
    void resize_end(World *, double X, double Y, Int_Mode mode = LeftButton);

    bool where_draw_string (double &x, double &y);
};

typedef Ellipse* Ellipse_ptr;
typedef GraphicEllipse* GraphicEllipse_ptr;

#endif ~_ELLIPSE_DEFINED;







