/*
    Copyright (C) 1992-1994
                    Welson R. Jacometti

    All rights reserved.

    This program is  free software; you can redistribute unmodified
    versions of it as long as you do not  profit from that, neither
    by selling it nor  by   including it  with other products   for
    sale.  This copyright  notice  must  remain in all  distributed
    copies.

    You are  allowed to modify your  personal copy  of this program
    (as long as you do not distribute any changed versions).

    You may,  of course, distribute  along with it, your own source
    code or libraries of algorithms that run under GeoLab.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

    If you must contact the authors, you may write to:
                     geolab-request@dcc.unicamp.br
*/

/*
 * GraphicDemo.C
 *
 *
 *    	Data	Hora  	Nome     		Modificacao/Motivo
 *
 *    21.10.92  13:40 	Welson R. Jacometti   	Arquivo criado
 *
 */

#include <math.h>
#include <stdio.h>
#include <sys/param.h>
#include <sys/types.h>
#include <xview/xview.h>
#include <xview/cms.h>
#include <xview/panel.h>
#include <xview/textsw.h>
#include <xview/xv_xrect.h>
#include <gdd.h>
#include "GraphicDemo_ui.h"
#include "GraphicDemo.h"
#include "Point2D.h"
#include "XString.h"
#include "geolab_ui.h"

bool movie_events(World *world, Xv_Window, Event *event);

double xt(double cx, double r, double t) {
    return cx + (r * cos(t));
}

double yt(double cy, double r, double t) {
    return cy + (r * sin(t));
}

double xp(double k, double f, double x0) {
    return (k * f) + x0;
}

double yp(double k, double g, double y0) {
    return (k * g) + y0;
}

double rad(double x) {
    return (x * PI) / 180;
}

#define going 0
#define looping 1
#define coming 2
#define finished 3

#define still 0
#define advance 1
#define goback 2

typedef struct MoveStringData {
    int stage;
    int stop;
    double counter, counter_loop;
    double cx, cy, r, tsignal;
    double xi, yi, xm, ym, xf, yf;
    double x0, y0, f, g, x1, y1, f1, g1;
    double t;
    double around;
    GraphicXString_ptr Xstr;
    unsigned long fg, bg;
};

typedef MoveStringData* MoveStringData_ptr;

typedef struct CircData {
    double cx, cy, r;
};

declare(list,MoveStringData_ptr)
declare(list,GraphicXString_ptr)

typedef list(GraphicXString_ptr) *list_ptr;

declare(list,list_ptr)

// Dados Globais
GraphicDemo_popup_graphic_demo_objects *GraphicDemo_popup_graphic_demo;
list(list_ptr) *MOVIE;
list_item MOVIE_SCENE;
double STEP1;
double STEP2;
double STEP3;
unsigned int STEP4;
int SHOW_MODE;
int FADE_MODE;
int SPEED;
int MOVIE_POS;
int MOVIE_DIR;
World *w;

void demo_circular(list(GraphicXString_ptr) &input) {
    // *** INICIO DEMO ***
    
    // determina circulo
    double trad, x1, x2, y1, y2;
    CircData centros[4];
    int val = 0; int sign = 0;
    int stageaux = going;
    int yctl = 0;
                
    w->get_values(x1, y1, x2, y2);
    centros[0].cx = ((x1 + x2) / 4);
    centros[0].cy = ((y1 + y2) / 4) + 30;
    centros[0].r = ((3 * (fabs(x2 - x1) / 2)) / 8) + 10;
    centros[1].cx = centros[0].cx;
    centros[1].cy = ((y1 + y2) / 2) + centros[0].cy - 30;
    centros[1].r = centros[0].r;
    centros[2].cx = ((x1 + x2) / 2) + centros[0].cx - 50;
    centros[2].cy = centros[0].cy;
    centros[2].r = centros[0].r;
    centros[3].cx = centros[2].cx;
    centros[3].cy = centros[1].cy;
    centros[3].r = centros[0].r;
    
    trad = 0;
        
    // armazena posicao inicial dos objetos
    list(MoveStringData_ptr) lm;
    MoveStringData_ptr data;
    GraphicXString_ptr it;
    
    forall(it, input) {
      data = new MoveStringData;

      data->cx = centros[val].cx;
      data->cy = centros[val].cy;
      data->r = centros[val].r;
      data->tsignal = sign;
      data->counter = data->counter_loop = 0;
      data->around = 1;
      data->stage = stageaux;
      data->stop = false;

      stageaux -= 20;
                          
      val = (val + 1) % 4;
      sign = !sign;
      
      data->Xstr = it;
      if (sign)
        data->xi = -500;
      else
        data->xi = 1200;
      data->yi = -100 + yctl; yctl += 100; 
      if (yctl > 1000) yctl = 0;
      data->xf = data->Xstr->obj_P().X();
      data->yf = data->Xstr->obj_P().Y();
      data->Xstr->set_obj_P(Point2D(data->xi, data->yi));
      
      // calcula posicao inicial no circulo
      data->t = rad(trad);
      data->xm = xt(data->cx, data->r, data->t);
      data->ym = yt(data->cy, data->r, data->t);
    
      // calcula equacoes parametricas das retas
      data->x0 = data->xi; data->y0 = data->yi;
      data->f = data->xm - data->xi;
      data->g = data->ym - data->yi;
      data->x1 = data->xf; data->y1 = data->yf;
      data->f1 = data->xm - data->xf;
      data->g1 = data->ym - data->yf;
      
      trad += 15;

      lm.push(data);
    }
    
    // faz os objetos deslocarem-se da posicao inicial ate' a posicao no circulo
    int stop = false;
    
    while (!stop) {
      forall(data, lm) {     
        if (data->stage == going) {
          data->Xstr->draw(w, XorMode);
          data->Xstr->setP(xp(data->counter, data->f, data->x0), yp(data->counter, data->g, data->y0));
          data->Xstr->draw(w, XorMode);
          data->counter += STEP1;
          if (data->counter >= 1) {
            data->stage = looping;
            data->counter = 1;
            data->Xstr->draw(w, XorMode);
            data->Xstr->setP(xp(1, data->f, data->x0), yp(1, data->g, data->y0));
            data->Xstr->draw(w, XorMode);
          }
        }

        if (data->stage == looping) {
          double my_counter = data->counter_loop;
          if (data->tsignal)
            my_counter = -my_counter;
          data->Xstr->draw(w, XorMode);
          data->Xstr->setP(xt(data->cx, data->r, rad(my_counter)+data->t), yt(data->cy, data->r, rad(my_counter)+data->t));
          data->Xstr->draw(w, XorMode);
          data->counter_loop += STEP2;
          if (data->counter_loop >= 360) {
            data->counter_loop = 0;
            if (--data->around == 0)
              data->stage = coming;
          }
        }

        if (data->stage == coming) {
          data->Xstr->draw(w, XorMode);
          data->Xstr->setP(xp(data->counter, data->f1, data->x1), yp(data->counter, data->g1, data->y1));
          data->Xstr->draw(w, XorMode);
          data->counter -= STEP1;
          if (data->counter <= 0) {
            data->counter = 0;
            data->stage = finished;
            data->stop = true;
            data->Xstr->draw(w, XorMode);
            data->Xstr->setP(xp(0, data->f1, data->x1), yp(0, data->g1, data->y1));
            data->Xstr->draw(w, XorMode);
          }
        }

        if (data->stage < going)
          data->stage++;
      }

      stop = true;
      forall(data, lm)
        stop = stop && data->stop;
    }
        
    // ***  FIM  DEMO  ***

    forall(data, lm) {
      data->Xstr->set_obj_P(Point2D(data->xf, data->yf));
      delete data;
    }
}

void demo_direct(list(GraphicXString_ptr) &input) {
    // *** INICIO DEMO ***
    
    // determina circulo
    int stageaux = going;
    int xctl = 0, sign = 0;
                
    // armazena posicao inicial dos objetos
    list(MoveStringData_ptr) lm;
    MoveStringData_ptr data;
    GraphicXString_ptr it;
    
    forall(it, input) {
      data = new MoveStringData;
      data->counter = 0;
      data->stage = stageaux;
      data->stop = false;

      stageaux -= 1;
                          
      sign = !sign;
      
      data->Xstr = it;
      if (sign)
        data->yi = -500;
      else
        data->yi = 1200;
      data->xi = -100 + xctl; xctl += 100; 
      if (xctl > 1000) xctl = 0;
      data->xf = data->Xstr->obj_P().X();
      data->yf = data->Xstr->obj_P().Y();
      data->Xstr->set_obj_P(Point2D(data->xi, data->yi));
      
      // calcula equacoes parametricas das retas
      data->x0 = data->xi; data->y0 = data->yi;
      data->f = data->xf - data->xi;
      data->g = data->yf - data->yi;
      
      lm.push(data);
    }
    
    int stop = false;
    
    while (!stop) {
      forall(data, lm) {     
        if (data->stage == going) {
          data->Xstr->draw(w, XorMode);
          data->Xstr->setP(xp(data->counter, data->f, data->x0), yp(data->counter, data->g, data->y0));
          data->Xstr->draw(w, XorMode);
          data->counter += STEP1;
          if (data->counter >= 1) {
            data->stage = finished;
            data->stop = true;
            data->counter = 1;
            data->Xstr->draw(w, XorMode);
            data->Xstr->setP(xp(1, data->f, data->x0), yp(1, data->g, data->y0));
            data->Xstr->draw(w, XorMode);
          }
        }

        if (data->stage < going)
          data->stage++;
      }

      stop = true;
      forall(data, lm)
        stop = stop && data->stop;
    }
        
    // ***  FIM  DEMO  ***

    forall(data, lm) {
      data->Xstr->set_obj_P(Point2D(data->xf, data->yf));
      delete data;
    }
}

void demo_gray_scale(list(GraphicXString_ptr) &input) {
    Cms ctrl_cms;
    Xv_Singlecolor cms_colors[32];
    unsigned char color_value = 256;
    unsigned long foreground;
    list(MoveStringData_ptr) lm;
    MoveStringData_ptr data;
    GraphicXString_ptr it;
    
    forall(it, input) {
      data = new MoveStringData;
      data->Xstr = it;
      it->get_color(data->fg, data->bg);
      lm.push(data);
    }
    
    for (int index = 0; index < 32; index++) {
      cms_colors[index].red = color_value;
      cms_colors[index].green = color_value;
      cms_colors[index].blue = color_value;
      color_value-=8;
    }
    
    ctrl_cms = (Cms)xv_find(NULL, CMS, 
    			CMS_SIZE, 	32 + CMS_CONTROL_COLORS,
    			CMS_COLORS,	cms_colors,
    			NULL);
    	
    for (index = 0; index < 16; index++) {
      foreground = (unsigned long)xv_get(ctrl_cms, CMS_PIXEL, index + CMS_CONTROL_COLORS);
      forall(data, lm) {
        data->Xstr->set_color(w, foreground, data->bg);
        data->Xstr->draw(w, CopyMode);
        w->Flush();
        usleep(STEP4);
      }
    }

    forall(data, lm) {
      data->Xstr->set_color(w, data->fg, data->bg);
      data->Xstr->draw(w, CopyMode);
      delete data;
    }
}

void demo_cross_strings(list(GraphicXString_ptr) &) {
    cerr << "Cross Strings: not implemented" << endl;
}

void fade_quick(list(GraphicXString_ptr) &input) {
    list(MoveStringData_ptr) lm;
    MoveStringData_ptr data;
    GraphicXString_ptr it;
    double yctl = 0;
    int sign = 0;
            
    forall(it, input) {
      data = new MoveStringData;
      
      data->Xstr = it;
      if (sign)
        data->xf = -500;
      else
        data->xf = 1200;
      data->yf = -100 + yctl; yctl += 10; 
      if (yctl > 1000) yctl = 0;
      sign = !sign;
      
      data->xi = data->Xstr->obj_P().X();
      data->yi = data->Xstr->obj_P().Y();
      
      // calcula equacoes parametricas das retas
      data->x0 = data->xi; data->y0 = data->yi;
      data->f = data->xf - data->xi;
      data->g = data->yf - data->yi;

      lm.push(data);
    }
    
    // fade away

    for (double counter = 0; counter <= 1; counter += STEP3) {
      forall(data, lm) {     
        data->Xstr->draw(w, XorMode);
        data->Xstr->setP(xp(counter, data->f, data->x0), yp(counter, data->g, data->y0));
        data->Xstr->draw(w, XorMode);
      }
    }

    forall(data, lm) {
      data->Xstr->draw(w, XorMode);
      data->Xstr->setP(xp(1, data->f, data->x0), yp(1, data->g, data->y0));
      data->Xstr->draw(w, XorMode);
      data->Xstr->draw(w, XorMode);
    }
    
    forall(data, lm) {
      data->Xstr->set_obj_P(Point2D(data->xi, data->yi));
      delete data;
    }
}

void fade_gray_scale(list(GraphicXString_ptr) &input) {
    Cms ctrl_cms;
    Xv_Singlecolor cms_colors[32];
    unsigned char color_value = 256;
    unsigned long foreground;
    list(MoveStringData_ptr) lm;
    MoveStringData_ptr data;
    GraphicXString_ptr it;
    
    forall(it, input) {
      data = new MoveStringData;
      data->Xstr = it;
      it->get_color(data->fg, data->bg);
      lm.push(data);
    }
    
    for (int index = 0; index < 32; index++) {
      cms_colors[index].red = color_value;
      cms_colors[index].green = color_value;
      cms_colors[index].blue = color_value;
      color_value-=8;
    }
    
    ctrl_cms = (Cms)xv_find(NULL, CMS, 
    			CMS_SIZE, 	32 + CMS_CONTROL_COLORS,
    			CMS_COLORS,	cms_colors,
    			NULL);
    	
    for (index = 0; index < 16; index++) {
      foreground = (unsigned long)xv_get(ctrl_cms, CMS_PIXEL, 15 - index + CMS_CONTROL_COLORS);
      forall(data, lm) {
        data->Xstr->set_color(w, foreground, data->bg);
        data->Xstr->draw(w, CopyMode);
        w->Flush();
        usleep(STEP4);
      }
    }

    forall(data, lm) {
      data->Xstr->set_color(w, 0, 0);
      data->Xstr->draw(w, CopyMode);
      data->Xstr->set_color(w, data->fg, data->bg);
      delete data;
    }
}

bool GraphicDemoReset::is_able_to_handle(const SelectedObjects &, const GraphicObjects &) {
    return true;
}

ResultObjects GraphicDemoReset::binder(const SelectedObjects &, const GraphicObjects &, World *, int, int) {
    ResultObjects dummy;
    list_item lstr_it, pt_it;
    list(GraphicXString_ptr) *lstr;

    forall_list_items(lstr_it, *MOVIE) {
      lstr = (*MOVIE)[lstr_it];
      forall_list_items(pt_it, *lstr) {
        delete (*lstr)[pt_it];
      }
      lstr->clear();
      delete lstr;
    }
        
    MOVIE->clear();
    (void)xv_set(GraphicDemo_popup_graphic_demo->textfield_movie_number_of_frames, PANEL_VALUE, MOVIE->size(), NULL);
      
    return dummy;
}


bool GraphicDemoRecord::is_able_to_handle(const SelectedObjects &SO, const GraphicObjects &GO) {
    list_item it;
    forall(it, SO) {
      if (GO[it]->type() != XString_t) {
        SO.init_iterator();
        return false;
      }
    }
          	
    return true;
}
  
ResultObjects GraphicDemoRecord::binder(const SelectedObjects &SO, const GraphicObjects &GO, World *, int, int) {
    ResultObjects dummy;
    list_item it;
    list(GraphicXString_ptr) *lstr = new list(GraphicXString_ptr);
    
    forall(it, SO)
      lstr->append((GraphicXString *)GO[it]->reproduct());
      
    MOVIE->append(lstr);  
    (void)xv_set(GraphicDemo_popup_graphic_demo->textfield_movie_number_of_frames, PANEL_VALUE, MOVIE->size(), NULL);
    
    return dummy;
}


icon_bits movie_icon[] = {
#include "ICONS/movie.icon"
};

icon_bits *GraphicDemoPlay::icon() {
    return movie_icon;
}

bool GraphicDemoPlay::is_able_to_activate(const SelectedObjects &, const GraphicObjects &, World *) {
    return MOVIE->size() > 0;
}


void GraphicDemoPlay::activator(World *world) {
    w = world;
    MOVIE_SCENE = 0;
    (void)xv_set(GraphicDemo_popup_graphic_demo->popup_graphic_demo, XV_SHOW, TRUE, NULL);
    w->add_event_proc(movie_events);
}


void set_speed(int value) {
    switch(value) {
      case 1:
        STEP1 = 1.0 / 300.0;
        STEP2 = 1;
        STEP3 = 1.0 / 150.0;
        STEP4 = 90000;
        break;
      case 2:
        STEP1 = 1.0 / 200.0;
        STEP2 = 1;
        STEP3 = 1.0 / 100.0;
        STEP4 = 70000;
        break;
      case 3:
        STEP1 = 1.0 / 100.0;
        STEP2 = 2;
        STEP3 = 1.0 / 50.0;
        STEP4 = 50000;
        break;
      case 4:
        STEP1 = 1.0 / 100.0;
        STEP2 = 3;
        STEP3 = 1.0 / 50.0;
        STEP4 = 30000;
        break;
      case 5:
        STEP1 = 1.0 / 80.0;
        STEP2 = 3;
        STEP3 = 1.0 / 40.0;
        STEP4 = 15000;
        break;
      case 6:
        STEP1 = 1.0 / 70.0;
        STEP2 = 4;
        STEP3 = 1.0 / 40.0;
        STEP4 = 10000;
        break;
      case 7:
        STEP1 = 1.0 / 60.0;
        STEP2 = 4;
        STEP3 = 1.0 / 40.0;
        STEP4 = 5000;
        break;
      case 8:
        STEP1 = 1.0 / 50.0;
        STEP2 = 5;
        STEP3 = 1.0 / 30.0;
        STEP4 = 2000;
        break;
      case 9:
        STEP1 = 1.0 / 50.0;
        STEP2 = 6;
        STEP3 = 1.0 / 30.0;
        STEP4 = 1000;
        break;
    }
}

#define SHOW_CIRCULAR		0
#define SHOW_DIRECT		1 
#define SHOW_GRAY_SCALE		2
#define SHOW_CROSS_STRINGS	3

void show() {
    set_speed(SPEED);

    if (MOVIE_SCENE) {
      switch(SHOW_MODE) {
        case SHOW_CIRCULAR:
          demo_circular(*(*MOVIE)[MOVIE_SCENE]);
          break;
        case SHOW_DIRECT:
          demo_direct(*(*MOVIE)[MOVIE_SCENE]);
          break;
        case SHOW_GRAY_SCALE:
          demo_gray_scale(*(*MOVIE)[MOVIE_SCENE]);
          break;
        case SHOW_CROSS_STRINGS:
          demo_cross_strings(*(*MOVIE)[MOVIE_SCENE]);
          break;
      }
    
      switch(MOVIE_DIR) {
        case advance:
          MOVIE_POS = (MOVIE_POS + 1 > MOVIE->size() ? MOVIE->size() : (MOVIE_SCENE = MOVIE->succ(MOVIE_SCENE), MOVIE_POS + 1));
          break;
        case goback:
          MOVIE_POS = (MOVIE_POS - 1 == 0 ? 1 : (MOVIE_SCENE = MOVIE->pred(MOVIE_SCENE), MOVIE_POS - 1));
          break;
      }

      (void)xv_set(GraphicDemo_popup_graphic_demo->textfield_movie_frame_position, PANEL_VALUE, MOVIE_POS, NULL);
    }
}

#define FADE_QUICK		0
#define FADE_GRAY_SCALE		1

void fade() {
    set_speed(SPEED);

    if (MOVIE_SCENE) {
      switch(FADE_MODE) {
        case FADE_QUICK:
          fade_quick(*(*MOVIE)[MOVIE_SCENE]);
          break;
        case FADE_GRAY_SCALE:
          fade_gray_scale(*(*MOVIE)[MOVIE_SCENE]);
          break;
      }

      switch(MOVIE_DIR) {
        case advance:
          MOVIE_POS = (MOVIE_POS + 1 > MOVIE->size() ? MOVIE->size() : (MOVIE_SCENE = MOVIE->succ(MOVIE_SCENE), MOVIE_POS + 1));
          break;
        case goback:
          MOVIE_POS = (MOVIE_POS - 1 == 0 ? 1 : (MOVIE_SCENE = MOVIE->pred(MOVIE_SCENE), MOVIE_POS - 1));
          break;
      }

      (void)xv_set(GraphicDemo_popup_graphic_demo->textfield_movie_frame_position, PANEL_VALUE, MOVIE_POS, NULL);
    }
}
 
bool movie_events(World *world, Xv_Window, Event *event) {
    w = world;

    if (event_is_down(event))
      if (event_is_ascii(event))
      	switch(event_action(event)) {
      	  case 46:
      	    MOVIE_DIR = still;
      	    break;
      	  case 62:
      	    MOVIE_DIR = advance;
      	    break;
      	  case 60:
      	    MOVIE_DIR = goback;
      	    break;
      	}
      else
        switch(event_action(event)) {
          case ACTION_SELECT:
            show();
            break;
        
          case ACTION_ADJUST:
            fade();
            break;
        }
    
    // retorna true APENAS se for necessario CONSUMIR o evento
    return false;
}
       
//
// Notify callback function for `setting_graphic_demo_show'.
//
void
movie_show_mode_handler(Panel_item, int value, Event *) {
    SHOW_MODE = value;
}

//
// Notify callback function for `setting_graphic_demo_fade'.
//
void
movie_fade_mode_handler(Panel_item, int value, Event *) {
    FADE_MODE = value;
}

//
// Notify callback function for `slider_movie_speed'.
//
void
movie_speed_handler(Panel_item, int value, Event *) {
    SPEED = value;
}

//
// Notify callback function for `textfield_movie_frame_position'.
//
Panel_setting
movie_frame_position_handler(Panel_item item, Event *event)
{
	GraphicDemo_popup_graphic_demo_objects	*ip = (GraphicDemo_popup_graphic_demo_objects *) xv_get(item, XV_KEY_DATA, INSTANCE);
	int value = (int) xv_get(item, PANEL_VALUE);

        if (MOVIE->size() == 0) {
          (void)xv_set(item, PANEL_VALUE, 0, NULL);
          MOVIE_POS = 0;
          MOVIE_SCENE = 0;
        }
        else
          if (value == 0) {
            (void)xv_set(item, PANEL_VALUE, 1, NULL);
            MOVIE_POS = 1;
            MOVIE_SCENE = MOVIE->first();
          }
          else 
            if (value > MOVIE->size()) {
              (void)xv_set(item, PANEL_VALUE, MOVIE->size(), NULL);
              MOVIE_POS = MOVIE->size();
              MOVIE_SCENE = MOVIE->last();
            }
            else {
              MOVIE_SCENE = MOVIE->first();
              if (MOVIE_POS != 0)
                for (int i = 1; i < value; i++)
                  MOVIE_SCENE = MOVIE->succ(MOVIE_SCENE);
              MOVIE_POS = value;
            }
        	
	return panel_text_notify(item, event);
}

//
// Notify callback function for `button_movie_show'.
//
void
button_movie_show_handler(Panel_item, Event *) {
    show();
}

//
// Notify callback function for `button_movie_fade'.
//
void
button_movie_fade_handler(Panel_item, Event *) {
    fade();
}

Algorithm* GraphicDemoReset_link() {
    return new GraphicDemoReset();
}

Algorithm* GraphicDemoRecord_link() {
    return new GraphicDemoRecord();
}

extern geolab_geolab_base_window_objects geolab_geolab_base_window;

Algorithm* GraphicDemoPlay_link() {
    MOVIE = new list(list_ptr);
    SHOW_MODE = 0;
    FADE_MODE = 0;
    SPEED = 3;
    MOVIE_POS = 0;
    MOVIE_DIR = still;
    GraphicDemo_popup_graphic_demo = new GraphicDemo_popup_graphic_demo_objects();
    GraphicDemo_popup_graphic_demo->objects_initialize(geolab_geolab_base_window.geolab_base_window);
    (void)xv_set(GraphicDemo_popup_graphic_demo->slider_movie_speed, PANEL_VALUE, SPEED, NULL);
    (void)xv_set(GraphicDemo_popup_graphic_demo->textfield_movie_frame_position, PANEL_VALUE, 0, NULL);
    (void)xv_set(GraphicDemo_popup_graphic_demo->textfield_movie_number_of_frames, PANEL_VALUE, 0, NULL);
    return new GraphicDemoPlay();
}







