/*
    Copyright (C) 1992-1994
                    Welson R. Jacometti

    All rights reserved.

    This program is  free software; you can redistribute unmodified
    versions of it as long as you do not  profit from that, neither
    by selling it nor  by   including it  with other products   for
    sale.  This copyright  notice  must  remain in all  distributed
    copies.

    You are  allowed to modify your  personal copy  of this program
    (as long as you do not distribute any changed versions).

    You may,  of course, distribute  along with it, your own source
    code or libraries of algorithms that run under GeoLab.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

    If you must contact the authors, you may write to:
                     geolab-request@dcc.unicamp.br
*/

/*
 * MiscClasses.C - mapeamento dos algoritmos interessantes de Misc.h (.C) em
 *                 algoritmos externos.
 * 
 *
 *    	Data	Hora  	Nome     		Modificacao/Motivo
 *
 *    24.04.92  09:30 	Welson R. Jacometti  	Arquivo criado
 *
 */

#include "MiscClasses.h"
#include "basic.h"
#include "List.h"
#include "Misc.h"
#include "Ray2D.h"
#include "Polygon.h"
#include "Integer.h"

#ifndef _Point2D_ptr_list
#define _Point2D_ptr_list
declare(list,Point2D_ptr)
#endif ~_Point2D_ptr_list

static list(Point2D_ptr) extract_points(const SelectedObjects &SO, const GraphicObjects &GO) {
    list(Point2D_ptr) res;
    list_item it;
    forall(it, SO) {			 	// is_able_to_handle deve ser
      res.push((Point2D *)GO[it]->obj);		// chamado antes para certificar
    }						// que sao todos pontos
    return res;
}

bool MiscPoints::is_able_to_handle(const SelectedObjects &SO, const GraphicObjects &GO) {
    list_item it;
    
    forall(it, SO) {
      if (GO[it]->type() != Point2D_t) {
      	SO.init_iterator();
      	return false;
      }
    }
    
    return true;
}

ResultObjects LowerY::binder(const SelectedObjects &SO, const GraphicObjects &GO, World *w, int animation_depth, int reference_depth) {
    ResultObjects result;
    list(Point2D_ptr) aux;
    
    if (!SO.empty()) {
      aux = extract_points(SO, GO);
      result.push(aux[it_get_min_Y(aux, w, animation_depth, reference_depth)]->reproduct());
    }
      
    return result;
}
  
ResultObjects LowerCCW::binder(const SelectedObjects &SO, const GraphicObjects &GO, World *w, int animation_depth, int reference_depth) {
    ResultObjects result;
    list(Point2D_ptr) aux;
    Point2D ref;
    
    if (!SO.empty() && SO.size() >= 2) {
      aux = extract_points(SO, GO);
      ref = *aux.tail();
      aux.del_item(aux.last());
      result.push(new Ray2D(ref, *aux[it_lower_angle_ccw(aux, ref, w, animation_depth, reference_depth)]));
    }
      
    return result;
}

bool WindingNumber::is_able_to_handle(const SelectedObjects &SO, const GraphicObjects &GO) {
    if (SO.size() == 2)
      if ((GO[SO[SO.first()]]->type() == Polygon_t && GO[SO[SO.succ(SO.first())]]->type() == Point2D_t) ||
          (GO[SO[SO.first()]]->type() == Point2D_t && GO[SO[SO.succ(SO.first())]]->type() == Polygon_t))
        return true; 
    return false;
}

ResultObjects WindingNumber::binder(const SelectedObjects &SO, const GraphicObjects &GO, World *w, int animation_depth, int reference_depth) {
    ResultObjects result;
    Polygon_ptr pol;
    Point2D ref;
    
    if (!SO.empty()) {
      if (GO[SO[SO.first()]]->type() == Point2D_t) {
        ref = *((Point2D_ptr)GO[SO[SO.first()]]->obj);
        pol = ((Polygon_ptr)GO[SO[SO.succ(SO.first())]]->obj);
      }
      else {
        ref = *((Point2D_ptr)GO[SO[SO.succ(SO.first())]]->obj);
        pol = ((Polygon_ptr)GO[SO[SO.first()]]->obj);
      }
    
      int wn = winding_number(pol->copy_vertices(), ref, w, animation_depth, reference_depth);
      
      Integer *res = new Integer(wn);
      res->obj_volatile(TRUE);
      result.push(res);
    }
    
    return result;
}

Algorithm* lowerY_link() {
    return new LowerY();
}

Algorithm* lowerCCW_link() {
    return new LowerCCW();
}

Algorithm* winding_number_link() {
    return new WindingNumber();
}



