/*
    Copyright (C) 1992-1994
                    Welson R. Jacometti

    All rights reserved.

    This program is  free software; you can redistribute unmodified
    versions of it as long as you do not  profit from that, neither
    by selling it nor  by   including it  with other products   for
    sale.  This copyright  notice  must  remain in all  distributed
    copies.

    You are  allowed to modify your  personal copy  of this program
    (as long as you do not distribute any changed versions).

    You may,  of course, distribute  along with it, your own source
    code or libraries of algorithms that run under GeoLab.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

    If you must contact the authors, you may write to:
                     geolab-request@dcc.unicamp.br
*/

/*
 *
 * Obj.C - Obj & GraphicObj methods.
 *
 *
 *    	Data	Hora  	Nome     		Modificacao/Motivo
 *
 *    19.11.91  13:40 	Welson R. Jacometti  	Arquivo criado
 *
 */

#include "Obj.h"

GraphicObj* Obj::new_GraphicObj() {
    return new GraphicObj(this);
}

void Obj::load(FILE *f) {
    char IS_A[MAXPATHLEN];
    fscanf(f, "%s %d\n", IS_A, &stype);
    is_a = IS_A;
}

void Obj::save(FILE *f) {
    fprintf(f, "%s %d\n", is_a.value(), stype);
}

void GraphicObj::draw(World *w, DrawMode dm) {
    if (saved_dm != dm)
      set_draw_mode(w, dm);
    saved_dm = dm;
    obj->draw(w, gco);
    // obj->draw_handles(w, gch, is_selected, show_handle);    
    draw_handles(w);
    if (label.length()) {
      double x,y;
      if (where_draw_string (x, y))
	w->DrawCenterString (gch, x, y, label.value(), label.length());
    }
}

void GraphicObj::load(FILE *f, World *w) {
    bool sh, is, gl, an;
    unsigned long fore, back;
    int lw, ls, lenlb;
    char *lb;
    
    obj->load(f);
    init();
    fscanf(f, "%d ", &lenlb);
    if (lenlb) {      
	lb = new char[lenlb+1];
	fgets (lb, lenlb+1, f);
	String aux (lb);
	label = aux;
	delete lb;
    }
    fscanf(f, "%d %d %d %d %u %u %d %d\n", &sh, &is, &gl, &an, &fore, &back, &lw, &ls);
    if (is) select();
    if (gl) glue();
    if (an) anchor();
    set_all(w, fore, back, lw, ls, sh);
}

void GraphicObj::save(FILE *f) {
    obj->save(f);
    fprintf(f, "%d ", label.length());
    if (label.length())
      fprintf(f, "%s ", label.value());
    fprintf(f, "%d %d %d %d %u %u %d %d\n", show_handle, is_selected, is_glued, is_anchored, foreground, background, line_width, line_style);
}


void GraphicObj::set_attributes(World *w, Attributes& attr)
{
    if (attr.DELETABLE == 1)
      anchor();
    else
      unanchor();
    
    if (attr.GLUE == 0)
      glue();
    else
      unglue();

    set_handle(!attr.HANDLES);
    
    set_string(attr.LABEL);
    
    set_line_attr (w, attr.LINE_WIDTH, attr.LINE_STYLE);    
    set_color (w, attr.FOREGROUND, attr.BACKGROUND);
}

Attributes GraphicObj::get_attributes()
{
    Attributes attr;
    attr.DELETABLE = (int)anchored();
    attr.GLUE = !glued();
    attr.HANDLES = !handle_is_on();
    get_string(attr.LABEL);
    get_line_attr (attr.LINE_WIDTH, attr.LINE_STYLE);
    get_color (attr.FOREGROUND, attr.BACKGROUND);
    
    // defaults para estranhos no ninho...(nao sao atributos do objeto).
    attr.INCREMENT = false;
    attr.INCREMENT_VALUE = 1;
    return attr;
}


