/*
    Copyright (C) 1992-1994
                    Cesar N. Gon

    All rights reserved.

    This program is  free software; you can redistribute unmodified
    versions of it as long as you do not  profit from that, neither
    by selling it nor  by   including it  with other products   for
    sale.  This copyright  notice  must  remain in all  distributed
    copies.

    You are  allowed to modify your  personal copy  of this program
    (as long as you do not distribute any changed versions).

    You may,  of course, distribute  along with it, your own source
    code or libraries of algorithms that run under GeoLab.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

    If you must contact the authors, you may write to:
                     geolab-request@dcc.unicamp.br
*/

/*
 * PSPM.C - Point Set Pattern Matching Algorithms
 *
 *
 *    	Data	Hora  	Nome     		Modificacao/Motivo
 *
 *    16.09.92  21:12 	Cesar N. Gon 	  	Arquivo criado
 *
 */

#include "PSPM.h"
#include "animation.h"
#include "trapsignal.h"
#include "select.h"
#include "objects.h"
#include "PSPM1_ui.h"
#include "animation_tools.h"

#include <assert.h>
#include <sys/param.h>
#include <sys/types.h>
#include <xview/xview.h>
#include <xview/panel.h>
#include <xview/textsw.h>
#include <xview/xv_xrect.h>
#include <gdd.h>
#include <xview/notice.h>

//
// Variaveis
//

PSPM1_popup_objects	*PSPM1_popup;
extern Attr_attribute	INSTANCE;

#define S_FACTOR 1.5
#define P_FACTOR 1.5

// Sorry, but...
World *button_w; // Necess'ario p/ chamada de exec_binder(w) na
                 // notify callback function for `button_match'.
                 // Iniciado em ::installer(w).


//
// fun,c~oes auxiliares

static void extract_points(const SelectedObjects &s, const GraphicObjects &obj, list(Point2D_ptr)& res) {
    list_item it;

    forall(it, s) {
	if (obj[it]->obj->type() == Point2D_t) {
	    res.push((Point2D *)obj[it]->obj);
	}
      }
}


bool colinear_points (list(Point2D_ptr)& P)
{
    if (P.length() > 2) {
	list_item it1, it2, it3;
	it1 = P.first();
	it2 = P.succ(it1);
	it3 = P.succ(it2);
	while (it3) 
	  if (! P[it3]->colinear(*(P[it1]), *(P[it2])))
	    return false;
	  else
	    it3 = P.succ(it3);
    }
    return true;
}

//
// S e P ordenados
//
list(Point2D_ptr) PSPM_in_one_dimension(const list(Point2D_ptr)& S, const list(Point2D_ptr)& P, World *w, int animation_depth, int reference_depth)
{
    // cerr << "ALGORITMO !!!\n";    
    list(Point2D_ptr) M;
    list_item p0, pi, it;
    int register i;
    int k = P.length();
    int n = S.length();
    p0 = pi = P.first();
    double dp, ds;
    bool find = true;

    Point2D_ptr *s = new Point2D_ptr[n]; assert(s);
    int *m = new int[k]; assert(m);

    for(i=0,it=S.first(); i<n; i++,it=S.succ(it))
      s[i] = S[it];
    for(i=0;i<k;i++) 
      m[i] = i;

    i = 0;

    ANIMATION_CODE(w, animation_depth, reference_depth) {
	show_arrow(w, *(P[p0]), *(s[0]));
	frame_stop_mark(w);
    }

    // cerr << "INICIO...\n";

    if (k<=n) {
      
	while (i < k-1 && find) {
	    i++;
	    // cerr << "\n\ni = " << i << endl;
	    pi = P.succ(pi);
	    m[i] = MAX(m[i], m[i-1] + 1);
	    // cerr << "m[i] = " << m[i] << endl;
	

	    // P[pi]->print(); 	 // s[m[i]]->print();
	    
	    ANIMATION_CODE(w, animation_depth, reference_depth) {
		show_arrow(w,*(P[pi]), *(s[m[i]]));
 		frame_stop_mark(w);
	    }

	    dp = (P[p0])->e_distance_sqr(*(P[pi]));
	    while (comp(dp,(ds = (s[m[0]])->e_distance_sqr(*(s[m[i]])))) == Greater && m[i] < n-k+i) {
		ANIMATION_CODE(w, animation_depth, reference_depth) {
		    show_arrow(w, *(P[pi]), *(s[m[i]]));
		}

		// cerr << "ds = " << ds << endl;
		m[i]++;
	    
		ANIMATION_CODE(w, animation_depth, reference_depth) {
		    show_arrow(w, *(P[pi]), *(s[m[i]]));
		    frame_stop_mark(w);
		}
		
	    }

	    if (comp(dp,ds) != Equal) {
		if (m[i] >= n-k+i)
		  find = false;
		else {
		    ANIMATION_CODE(w, animation_depth, reference_depth) {
			frame_stop_mark(w);
			int register j;
			pi = p0;
			for (j=0;j<=i;j++) {
			    show_arrow(w, *(P[pi]), *(s[m[j]]));
			    pi = P.succ(pi);
			}
			frame_stop_mark(w);
		    }
		  
		    m[0]++;
		    i = 0;
		    pi = p0;
		
		    ANIMATION_CODE(w, animation_depth, reference_depth) {
			show_arrow(w, *(P[p0]), *(s[m[0]]));
			frame_stop_mark(w);
		    }
		}
	    }
	} /* fim do while externo */
    } /* fim do if (k<=n) */

    // cerr << "FIM DO WHILE...\n";

    ANIMATION_CODE(w, animation_depth, reference_depth) {
	frame_stop_mark(w);
	int register j;
	pi = p0;
	for (j=0;j<=i;j++) {
	    show_arrow(w, *(P[pi]), *(s[m[j]]));
	    pi = P.succ(pi);
	}
    }

    if (find) {
	ANIMATION_CODE(w, animation_depth, reference_depth) {
	    register int j;
	    for (j=0;j<6;j++) {
		frame_delay_mark(w);
		pi = p0;
		for (i=0;i<k;i++) {
		    show_arrow(w, *(P[pi]), *(s[m[i]]));
		    pi = P.succ(pi);
		}
	    }
	}

	for(i=0; i<k; i++)
	  M.push(s[m[i]]);
    }

    delete [n] s;
    delete [k] m;
    
    return M;
}

//
// implementa,c~ao dos m'etodos de PSPM1
//
bool PSPM1::is_able_to_activate(const SelectedObjects &SO, const GraphicObjects &GO, World *) {

    list_item it;
    Obj_ptr obj1, obj2, obj3;
    
    if (SO.empty())
      return false;
    it = SO.first();
    obj1 = GO[SO[it]]->obj;
    if (obj1->type() != Point2D_t)
      return false;
    if (SO.length()>1) {
	it = SO.succ(it);
	obj2 = GO[SO[it]]->obj;
	if (obj2->type() != Point2D_t)
	  return false;
	if (SO.length()>2) {
	    it = SO.succ(it);
	    while (it) {
		obj3 = GO[SO[it]]->obj;
		if (obj3->type() != Point2D_t)
		  return false;		
		if (!((Point2D_ptr)obj1)->colinear(*(Point2D_ptr)obj2, *(Point2D_ptr)obj3))
		  return false;
		it = SO.succ(it);
	    }
	}
    }

    return true;
}


icon_bits PSPM1_icon[] = {
#include "ICONS/PSPM1.icon"
};

icon_bits *PSPM1::icon() {return PSPM1_icon;}

void PSPM1::installer(World *w) {

    // cerr << ">installer\n";
    
    // Initialize user interface components.
    PSPM1_popup = new PSPM1_popup_objects;
    assert(PSPM1_popup);
    PSPM1_popup->objects_initialize((Frame)xv_get(w->get_canvas(), XV_OWNER));	
    animation_tools_init(w);
    S = new list(Point2D_ptr);assert(S);
    P = new list(Point2D_ptr);assert(P);

    button_w = w;
    // cerr << "<installer\n";
}

void PSPM1::uninstaller(World *w) {

    // cerr << ">uninstaller\n";
    
    // free user interface components memory
    delete PSPM1_popup;
    delete S;
    delete P;
	
    // cerr << "<uninstaller\n";
}

void PSPM1::activator(World *w) {
    // cerr << ">activator\n";

    Point2D_ptr p;
    GraphicObjects GO = objects_get_graphic_objects();
    SelectedObjects SO =  objects_get_selected_objects();
    extract_points(SO, GO, *S);
    forall (p, *S) 
	anim_show(w, *p, S_FACTOR);
    S->sort((int(*)(Point2D_ptr&, Point2D_ptr&))lexico_Xcomp_ptr);

    (void)xv_set(PSPM1_popup->button_match, PANEL_INACTIVE, FALSE, 0);
    (void)xv_set(PSPM1_popup->popup, WIN_SHOW, TRUE, WIN_FRONT, 0);
    
    w->add_event_proc(select_events);
    
    // cerr << "<activator\n";
}

void PSPM1::deactivator(World *w) {
    // cerr << ">deactivator\n";
    
//  Point2D_ptr p;    
//    forall (p, *S) 
//      anim_show(w, *p, S_FACTOR);
    
    S->clear();
    (void)xv_set(PSPM1_popup->button_match, PANEL_INACTIVE, TRUE, 0);
    (void)xv_set(PSPM1_popup->popup, WIN_SHOW, FALSE, 0);
    objects_refresh(w);
    
    // cerr << "<deactivator\n";
}

ResultObjects PSPM1::binder(const SelectedObjects& SO, const GraphicObjects& GO, World *w, int animation_depth, int reference_depth) {
    // cerr << ">binder\n";
    
    ResultObjects Match;
    list(Point2D_ptr) M;    

    Point2D_ptr p;
    extract_points(SO, GO, *P);
    forall (p, *P) 
      anim_show(w, *p, P_FACTOR);

    if (!colinear_points(*P))
      (void)notice_prompt(PSPM1_popup->popup, NULL,
			  NOTICE_MESSAGE_STRINGS, "Pattern set is not a colinear points set.",
			  "Select another points pattern set.",
			  NULL,
			  NOTICE_BUTTON_YES,
			  "Continue",
			  NULL);
    else if (S->length() < P->length()) 
      (void)notice_prompt(PSPM1_popup->popup, NULL,
			  NOTICE_MESSAGE_STRINGS,
			  "Pattern set is greater than sampling set.",
			  "Select another points pattern set.",
			  NULL,
			  NOTICE_BUTTON_YES,
			  "Continue",
			  NULL);
    else {
	P->sort((int(*)(Point2D_ptr&, Point2D_ptr&))lexico_Xcomp_ptr);
	M = PSPM_in_one_dimension(*S, *P, w, animation_depth, reference_depth);
	if (M.empty()) 
	  (void)notice_prompt(PSPM1_popup->popup, NULL,
			      NOTICE_MESSAGE_STRINGS,
			      "No match was found.",
			      NULL,
			      NOTICE_BUTTON_YES,
			      "Failure",			  
			      NULL);
	else {
	    forall (p, M)
	      Match.push(new Point2D(*p));
	    (void)notice_prompt(PSPM1_popup->popup, NULL,
				NOTICE_MESSAGE_STRINGS,
				"Yes, a match was found.",
				NULL,
				NOTICE_BUTTON_YES,
				"Success",			    
				NULL);
	}

    }
    forall (p, *P) 
      anim_show(w, *p, P_FACTOR);
    P->clear();

    // cerr << "<binder\n";
    return Match;
}


Algorithm* PSPM1_link() {
    Algorithm *it;
    it = new PSPM1();
    assert(it);
    return it;
}

//
// Notify callback function for `button_match'.
//
void
button_match_handler(Panel_item item, Event *event)
{
    // cerr << ">button_math_handler\n";
    PSPM1_popup_objects	*ip = (PSPM1_popup_objects *) xv_get(item, XV_KEY_DATA, INSTANCE);
    EF_exec_binder(button_w);
    // cerr << "<button_math_handler\n";
}

//-------------------------------------------------------------------------
//-------------------------------------------------------------------------
//-------------------------------------------------------------------------
//-------------------------------------------------------------------------

//
// implementa,c~ao dos m'etodos de PSPM2
//
bool PSPM2::is_able_to_activate(const SelectedObjects &SO, const GraphicObjects &GO, World *) {
    list(Point2D_ptr) S1;
    extract_points(SO, GO, S1);    
    return !S1.empty();
}


icon_bits PSPM2_icon[] = {
#include "ICONS/PSPM1.icon"
};

icon_bits *PSPM2::icon() {return PSPM2_icon;}

void PSPM2::installer(World *w) {
    // cerr << ">installer\n";

    // Initialize user interface components.
    PSPM1_popup = new PSPM1_popup_objects;
    assert(PSPM1_popup);
    PSPM1_popup->objects_initialize((Frame)xv_get(w->get_canvas(), XV_OWNER));	   
    animation_tools_init(w);
    S = new list(Point2D_ptr);
    assert(S);
    P = new list(Point2D_ptr);
    assert(P);

    button_w = w;
    // cerr << "<installer\n";

}

void PSPM2::uninstaller(World *w) {
    // cerr << ">uninstaller\n";

    // free user interface components memory
    delete PSPM1_popup;
    delete S;
    delete P;
	    
    // cerr << "<uninstaller\n";
}

void PSPM2::activator(World *w) {
    // cerr << ">activator\n";

    Point2D_ptr p;
    GraphicObjects GO = objects_get_graphic_objects();
    SelectedObjects SO =  objects_get_selected_objects();
    extract_points(SO, GO, *S);
    lenS = S->length();
    
    forall (p, *S) 
	anim_show(w, *p, S_FACTOR);

    // cria vetor Ss
    Ss = new list(Point2D_ptr)[lenS]; assert(Ss);
    
    // inicia vetor Ss com listas ordenadas circularmente
    preprocessing();
    
    (void)xv_set(PSPM1_popup->button_match, PANEL_INACTIVE, FALSE, 0);
    (void)xv_set(PSPM1_popup->popup, WIN_SHOW, TRUE, WIN_FRONT, 0);
    
    w->add_event_proc(select_events);
        
    // cerr << "<activator\n";
}

void PSPM2::deactivator(World *w) {
    // cerr << ">deactivator\n";

    S->clear();
    delete [lenS] Ss;
    (void)xv_set(PSPM1_popup->button_match, PANEL_INACTIVE, TRUE, 0);
    (void)xv_set(PSPM1_popup->popup, WIN_SHOW, FALSE, 0);
    objects_refresh(w);
    
    // cerr << "<deactivator\n";
}

void PSPM2::preprocessing() {
    // cerr << ">preprocessing\n";
    list(Point2D_ptr) S1;
    Point2D_ptr p;
    register unsigned i;

    i = 0;
    S1 = *S;    
    forall (p, *S) {
	set_circular_comp_reference(*p);
	S1.sort((int(*)(Point2D_ptr&, Point2D_ptr&))circular_comp_ptr);
	Ss[i++] = S1;
    }
    
    // cerr << "<preprocessing\n";
}

ResultObjects PSPM2::binder(const SelectedObjects& SO, const GraphicObjects& GO, World *w, int animation_depth, int reference_depth) {
    // cerr << "<binder\n";
    
    ResultObjects Match;
    list(Point2D_ptr) M;    

    Point2D_ptr p;
    extract_points(SO, GO, *P);
    forall (p, *P) 
      anim_show(w, *p, P_FACTOR);

    if ( lenS < P->length()) 
      (void)notice_prompt(PSPM1_popup->popup, NULL,
			  NOTICE_MESSAGE_STRINGS,
			  "Pattern set is greater than sampling set.",
			  "Select another points pattern set.",
			  NULL,
			  NOTICE_BUTTON_YES,
			  "Continue",
			  NULL);
    else {
	set_circular_comp_reference(*(P->head()));	
	P->sort((int(*)(Point2D_ptr&, Point2D_ptr&))circular_comp_ptr);
	M = PSPM_in_two_dimension(Ss, *P, w, animation_depth, reference_depth);
	if (M.empty()) 
	  (void)notice_prompt(PSPM1_popup->popup, NULL,
			      NOTICE_MESSAGE_STRINGS,
			      "No match was found.",
			      NULL,
			      NOTICE_BUTTON_YES,
			      "Failure",			  
			      NULL);
	else {
	    forall (p, M)
	      Match.push(new Point2D(*p));
	    (void)notice_prompt(PSPM1_popup->popup, NULL,
				NOTICE_MESSAGE_STRINGS,
				"Yes, a match was found.",
				NULL,
				NOTICE_BUTTON_YES,
				"Success",			    
				NULL);
	}

    }
    forall (p, *P) 
      anim_show(w, *p, P_FACTOR);
    P->clear();

    // cerr << "<binder\n";
    return Match;
}


Algorithm* PSPM2_link() {
    Algorithm *it;
    it = new PSPM2();
    assert(it);
    return it;
}


double asub(double a1, double a2) {
    return (a1 >= a2 ? a1-a2 : a1-a2+2*PI);
}

#define idx(x)  ((((x)-1) % (n-1)) + 1)

//
// listas de Ss e P ordenadas circularmente
//
list(Point2D_ptr) PSPM_in_two_dimension(const list(Point2D_ptr)* Ss, const list(Point2D_ptr)& P, World *w, int animation_depth, int reference_depth)
{
    // cerr << "<**************** ALGORITMO!!!**************\n";
    
    list(Point2D_ptr) M, S1 = Ss[0];
    bool success = false;
    int register count = 0;
    int register i;
    int k = P.length();
    int n = S1.length();
    Point2D_ptr *s = new Point2D_ptr[n]; assert(s); // array of Point2D_ptr
    int *m = new int[k]; assert(m);
    list_item p0, p1, pi, it;
    double dp01, dp0i, ds01, ds0i, ap01, ap0i, as01, as0i;

    if (k<=n)
      if (k>1) {
	  p0 = P.first();
	  p1 = P.succ(p0);
	  dp01 = (P[p0])->e_distance_sqr(*(P[p1])); 
	  ap01 = (P[p0])->angle(*(P[p1]));
	  // cerr << "!! dp01 = " << dp01 << endl;
	  // cerr << "!! ap01 = " << ap01 << endl;
	  
	  ANIMATION_CODE(w, animation_depth, reference_depth) {	      
	      show_arrow(w, *P[p0], *P[p1]);
	      frame_stop_mark(w);
	  }
	  
	  // cerr << ">while (!success...\n";
	  while (!success && count < n) { // n == lenS
	      // cerr << "!! count = " << count << endl;
	      S1 = Ss[count];
	      	      
	      for(i=0,it=S1.first(); i<n ; i++,it=S1.succ(it)) 
		s[i] = S1[it];		 
	      
	      m[0] = 0;
	      m[1] = 1;

	      ANIMATION_CODE(w, animation_depth, reference_depth) {
		  anim_show(w, *s[m[0]], S_FACTOR*1.5);
	      }

	      
	      do {
		  ds01 = (s[m[0]])->e_distance_sqr(*(s[m[1]]));
		  as01 = (s[m[0]])->angle(*(s[m[1]]));
		  // cerr << "!! tenta achar um match para p0,p1\n";
		  // cerr << "!! ds01 = " << ds01 << endl;
		  // cerr << "!! as01 = " << as01 << endl;
		  ANIMATION_CODE(w, animation_depth, reference_depth) {
		      frame_stop_mark(w);
		      show_arrow(w, *s[m[0]], *s[m[1]]);		      
		  }
		  // cerr << "!! dp01 = " << dp01 << endl;
		  // cerr << "!! ds01 = " << ds01 << endl;
		  if (equal(dp01,ds01,1e-2)) {
		      // cerr << "!! tenta achar um match para p2..pn\n";
		      i = 1;
		      pi = p1;
		      success = true;
		      while (i < k-1 && success) {
			  i++;
			  m[i] = m[i-1]+1;
			  // cerr << "!! i = " << i << endl;
			  pi = P.succ(pi);
			  dp0i = (P[p0])->e_distance_sqr(*(P[pi]));
			  ap0i = (P[p0])->angle(*(P[pi]));
			  ds0i = (s[m[0]])->e_distance_sqr(*(s[idx(m[i])]));
			  as0i = (s[m[0]])->angle(*(s[idx(m[i])]));
			  // cerr << "!! dp0i = " << dp0i << endl;
			  // cerr << "!! ap0i = " << ap0i << endl;
			  // cerr << "!! ds0i = " << ds0i << endl;
			  // cerr << "!! as0i = " << as0i << endl;
			  ANIMATION_CODE(w, animation_depth, reference_depth) {
			      show_arrow(w, *s[m[0]], *s[idx(m[i])]);
			      show_arrow(w, *P[p0], *P[pi]);
			      frame_stop_mark(w);
			  }
			  
			  while (!(equal(dp01, ds01, 1e-2) && equal(asub(ap0i,ap01),asub(as0i,as01),1e-2) && equal(dp0i, ds0i,1e-2)) &&  k-i < n-(m[i]-m[1])) {
			      ANIMATION_CODE(w, animation_depth, reference_depth) {
				  frame_delay_mark(w);
				  show_arrow(w, *s[m[0]], *s[idx(m[i])]);
			      }
			      m[i]++;
			      ANIMATION_CODE(w, animation_depth, reference_depth) {
				  show_arrow(w, *s[m[0]], *s[idx(m[i])]);
				  frame_stop_mark(w);
			      }			      
			      ds0i = (s[m[0]])->e_distance_sqr(*(s[idx(m[i])]));
			      as0i = (s[m[0]])->angle(*(s[idx(m[i])]));
			      // cerr << "!! ds0i = " << ds0i << endl;
			      // cerr << "!! as0i = " << as0i << endl;
			  }
			  
			  success = (equal(dp01, ds01, 1e-2) && equal(asub(ap0i,ap01),asub(as0i,as01),1e-2) && equal(dp0i,ds0i,1e-2));
		      }
		      ANIMATION_CODE(w, animation_depth, reference_depth) {
			  if (!success) {
			      // cerr << ">Apagando setas restantes...\n";
			      frame_delay_mark(w);
			      for (;i>1;i--) 
				show_arrow(w, *s[m[0]], *s[idx(m[i])]);
			      for (;pi!=p1;pi=P.pred(pi))
				show_arrow(w, *P[p0], *P[pi]);
			      // cerr << "<Apagando setas restantes...\n";
			  }
		      }
		  }

		  if (!success) {
		      ANIMATION_CODE(w, animation_depth, reference_depth) {
			  frame_delay_mark(w);
			  show_arrow(w, *s[m[0]], *s[m[1]]);
		      }		      
		      m[1]++;
		  }
	      } while (!success && m[1]<n);

	      ANIMATION_CODE(w, animation_depth, reference_depth) {
		  anim_show(w, *s[m[0]], S_FACTOR*1.5);
	      }	      
	      count++;
	      
	  } /* fim do while */
	  // cerr << "<while (!success...\n";
	  
	  if (success) {
	      ANIMATION_CODE(w, animation_depth, reference_depth) {
		  register int j;
		  for (j=0;j<6;j++) {
		      frame_delay_mark(w);
		      pi = p0;
		      for (i=0;i<k;i++) {
			  show_arrow(w, *(P[pi]), *(s[idx(m[i])]));
			  pi = P.succ(pi);
		      }
		  }
	      }
	      
	      for(i=0; i<k; i++)
		M.push(s[idx(m[i])]);
	  }
      }
    
    delete [n] s;
    delete [k] m;
    
    return M;
}
