/*
    Copyright (C) 1992-1994
                    Cesar N. Gon

    All rights reserved.

    This program is  free software; you can redistribute unmodified
    versions of it as long as you do not  profit from that, neither
    by selling it nor  by   including it  with other products   for
    sale.  This copyright  notice  must  remain in all  distributed
    copies.

    You are  allowed to modify your  personal copy  of this program
    (as long as you do not distribute any changed versions).

    You may,  of course, distribute  along with it, your own source
    code or libraries of algorithms that run under GeoLab.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

    If you must contact the authors, you may write to:
                     geolab-request@dcc.unicamp.br
*/

/*
 * Polygon.C - Polygon & GraphicPolygon methods.
 *
 *
 *    	Data	Hora  	Nome     		Modificacao/Motivo
 *
 *    12.02.92  16:38 	Cesar Nivaldo Gon  	Arquivo criado
 *
 */

#include "Obj.h"
#include "Polygon.h" 

#ifndef _list_item_list
#define _list_item_list
declare(list,list_item)
#endif ~_list_item_list;

Polygon::Polygon(list(Point2D_ptr)& Points, ObjType t, bool c) : Obj("Polygon", t), is_closed(c) {
    Point2D_ptr p_ptr;
    forall (p_ptr, Points)
	P.append ((Point2D_ptr)p_ptr->reproduct());
}


GraphicObj* Polygon::new_GraphicObj() {
    return new GraphicPolygon(this);
}


void Polygon::star_shaped () {
    if (!P.empty()) {
	double x=0,y=0;
	Point2D_ptr p_ptr;	
	forall (p_ptr,P) {
	    x=x+p_ptr->X();
	    y=y+p_ptr->Y();
	}	
	Point2D ref(x/P.length(),y/P.length());
	star_shaped (ref);
    }
    else subtype (StarShaped_t);
}

void Polygon::star_shaped ( Point2D &ref) {
    set_circular_comp_reference (ref);
    P.sort((int(*)(Point2D_ptr&, Point2D_ptr&))circular_comp_ptr);
    subtype(StarShaped_t);
}

void Polygon::convex() {
    remove_duplicated_points(P);
    if (length()>3) {
	SideType side;
	list_item vi, vi1, vi2;
	Point2D_ptr p_ptr, q_ptr = P.head();
	forall (p_ptr,P) 
	    if (lexico_Ycomp(*p_ptr,*q_ptr)==Less) q_ptr=p_ptr;
	star_shaped(*q_ptr);

	vi=P.first();
	do {
	    vi1 = P.succ(vi);
	    vi2 = P.cyclic_succ(vi1);
	    side = P[vi]->which_side(*P[vi1], *P[vi2]);
	    if (side == Right || (side == Colinear && vi2 != vi)) { 
		delete P[vi1];
		P.del_item (vi1);
		
		/*
		 * Faz backtrack somente se nao estiver no ponto de partida.
		 *
		 */
		
		if (P.pred(vi))
		  vi=P.pred(vi);
	    }
	    else
	      vi = P.succ(vi); 
	} while (P.succ(vi));
    }
    subtype(Convex_t);
}


int count_intersections(Segment2D s,list(Point2D_ptr) Q) {
    int count=0;
    Segment2D r;
    list_item it;

    forall_list_items (it,Q) {
	r.set (*Q[it],*Q[Q.cyclic_succ(it)]);
	if (s.intersect (r)) count++;
    }

    return count;
}


int Polygon::how_many_intersections ( Segment2D s)  {
    return count_intersections(s, P);
}


list(Point2D_ptr) Polygon::copy_vertices ()  {
    list(Point2D_ptr) Points;
    Point2D_ptr p_ptr;

    forall (p_ptr, P)
      Points.append ((Point2D_ptr)p_ptr->reproduct());

    return Points;
}

bool Polygon::is_simple()  {
    Segment2D_ptr s;
    list(Segment2D_ptr) S = copy_edges();    
    forall (s, S)
      if (how_many_intersections(*s)>3)
	return false;    
    forall (s, S)
      delete s;
    return true;
}

bool Polygon::is_convex()  {
    if (is_simple()) {
	list_item it;
	bool first = true;
	SideType d, dir;
	forall_list_items (it, P) {
	    d = P[it]->which_side(*(P[P.cyclic_succ(it)]), *(P[P.cyclic_pred(it)]));
	    if (d != Colinear) {
		if (first) {
		    dir = d;
		    first = false;
		}
		else
		  if (d != dir)
		    return false;
	    }
	}
	return true;
    }
    return false;
}      
    

void Polygon::simple() {
    remove_duplicated_points(P);
    // cerr << "> Pontos duplicados removidos...\n";
    if (P.length()>3) {
	list(Point2D_ptr) Q;
	list(list_item) Candidates;
	list_item it;
	Point2D_ptr p_ptr;
	Segment2D s;
	bool found;
	int n;
	
        P.permute();
	Q.push (p_ptr = P.pop());
	Q.push (p_ptr = P.pop());
	Q.push (p_ptr = P.pop());
	while (!P.empty()) {	    
	    p_ptr = P.pop();
	      
	    // cerr << "P: ";
	    // p_ptr->print();
	    
	    found=false;
	    forall_list_items (it,Q) {
		s.set (*Q[it],*Q[Q.cyclic_succ(it)]);
		if (s.point_in(*p_ptr)) {		    
		    found = true;
		    Q.init_iterator();
		    break;
		}
	    }
	    if (found) {
		// cerr << "> Insere sobre segmento. \n";
		Q.init_iterator();
		Q.insert(p_ptr,it,after);
	    }
	    else {
		// cerr << "!found...\n";
		// cerr << "Q: " << endl;
		// forall_list_items (it, Q) {
		//    Q[it]->print();
		// }
		
	    	int new_n;
	    	s.set(*p_ptr, *Q.tail());
		// s.print();
	    	n = count_intersections(s, Q);
		// cerr << "    n = " << n << endl;
		forall_list_items (it,Q) {
		    s.set (*p_ptr, *Q[it]);
		    // s.print();
		    new_n = count_intersections (s,Q);
		    // cerr << "new_n = " << new_n << endl;
                    if (n < 3 && new_n < 3)
  		      Candidates.push(Q.cyclic_pred(it));
		    n = new_n;
		}
		if (Candidates.empty()) {
		  cerr << "******* BUG detect!!! *******\n";
		  cerr << "* simple() at Polygon.C     *\n";
		  cerr << "*****************************\n";
		} else {
		    int count = rand () % Candidates.length();
		    // cerr << "> Aresta sorteada: " << count << "\n"; 
		    forall (it,Candidates)
		      if (!count--) break;
		    Q.insert(p_ptr,it,after);
		    Candidates.init_iterator();
		    Candidates.clear();
		}
	    }
	}
	// cerr << "> Fim. \n";
        P=Q;
	Q.clear();
    }
    subtype (Simple_t);
}


list(Segment2D_ptr) Polygon::copy_edges ()  {
    list(Segment2D_ptr) Segments;
    list_item it;

    forall_list_items (it,P) {
      if (is_closed || P.succ(it)) 
        Segments.append( (Segment2D_ptr)new Segment2D ((Point2D)*(P[it]), (Point2D)*(P[P.cyclic_succ(it)])) );
    }

    return Segments;
}

// inherited methods redefinition

void Polygon::load(FILE *f) {
    int size;
    Point2D_ptr p;
    Obj::load(f);
    fscanf(f, "%d\n", &is_closed);
    fscanf(f, "%d\n", &size);
    for (int i = 0; i < size; i++) {
      p = new Point2D();
      p->load(f);
      (void)push(p);
    }
}

void Polygon::save(FILE *f) {
    list_item it;
    Obj::save(f);
    fprintf(f, "%d\n", is_closed);
    fprintf(f, "%d\n", P.size());
    forall_list_items (it,P) {
      P[it]->save(f);
    }
}

void Polygon::translate(double to_x, double to_y) {
    list_item it;
    forall_list_items (it,P) {
      P[it]->translate(to_x, to_y);
    }
}

void Polygon::rotate(double orig_x, double orig_y, double angle) {
    list_item it;
    forall_list_items (it,P) {
      P[it]->rotate(orig_x, orig_y, angle);
    }
}

void Polygon::unfill_handles(World *w, bool show_handle)  {
    list_item it;
    forall_list_items (it,P) {
      if (show_handle)
        P[it]->unfill_handles(w);
    }
}

void Polygon::draw_handles(World *w, GC gc, bool highlight, bool show_handle)  {
    list_item it;
    forall_list_items (it,P) {
      if (show_handle)
        P[it]->draw_handles(w, gc, highlight);
    }
}

void Polygon::draw(World *w, GC gc)  {
    list_item it;
    forall_list_items (it,P) {
      if (is_closed || P.succ(it))
        w->DrawLine (gc, P[it]->X(), P[it]->Y(),P[P.cyclic_succ(it)]->X(), P[P.cyclic_succ(it)]->Y());
    }
}

bool Polygon::select_test (double x, double y)  {
    Point2D_ptr p_ptr;

    forall (p_ptr,P)
      if (select_handle_test (p_ptr->X(),p_ptr->Y(),x,y)) {
	  P.init_iterator();
	  return true;
      }

    return false;    
}

bool Polygon::select_test (double x1, double y1, double x2, double y2)  {
    Point2D_ptr p_ptr;

    forall (p_ptr,P)
      if (!select_handle_test (p_ptr->X(), p_ptr->Y(), x1, y1, x2, y2)) {
	  P.init_iterator();
	  return false;
      }

    return true;
}

double Polygon::distance_to_point(double x, double y)  {
    list_item it;
    double min = TOO_FAR;
    
    forall_list_items (it,P) {
      if (is_closed || P.succ(it)) {
      	double dist = Segment2D((Point2D)*(P[it]), (Point2D)*(P[P.cyclic_succ(it)])).distance_to_point(x, y);
        if (dist < min)
          min = dist;
      }
    }
    
    return min;
}

Point2D_ptr Polygon::handle_at(double x, double y)  {
    Point2D_ptr p_ptr;

    forall (p_ptr,P)
      if (select_handle_test (p_ptr->X(),p_ptr->Y(),x,y)) {
	  P.init_iterator();
	  return p_ptr;
      }

    return 0;
}


bool Polygon::point_inside(double x, double y) 
{
    Point2D p1(x,y);
    Point2D p2(get_max_X(P));
    Point2D_ptr p_ptr;
    Segment2D_ptr s_ptr;
    bool ok;
    
    p2.translate (5,5);
    Segment2D s(p1,p2);

    do {
	ok=true;
	forall (p_ptr, P)
	    if (s.point_in(*p_ptr)) {
		s.P2().translate (0,EPSLON);
		ok=false;
		break;
	    }
    } while (!ok);
		
    int  intersections = 0;    		

    list(Segment2D_ptr) Segments = copy_edges();
    forall (s_ptr, Segments) {
	if (s_ptr->intersect (s)) intersections++;
    }

    forall (s_ptr, Segments)
	delete s_ptr;
    return (intersections % 2);
}

void Polygon::bounding_box (double& x1, double& y1, double& x2, double& y2) 
{
    x1 = (get_min_X (P)).X();
    y1 = (get_min_Y (P)).Y();
    x2 = (get_max_X (P)).X();
    y2 = (get_max_Y (P)).Y();
}


list(Obj_ptr) Polygon::get_segments () 
{
    Segment2D_ptr s_ptr;
    list(Obj_ptr) Segments;
    list(Segment2D_ptr) edges = copy_edges();    
    forall (s_ptr, edges)
      Segments.push(s_ptr);    
    return Segments;
}


void Polygon::print ()  {
   Point2D *p_ptr;
   
   cout << "Polygon:" << endl;
   forall (p_ptr,P) {
       cout << "\t"; p_ptr->print();
   }
}

void Polygon::clear() {
    Point2D_ptr p_ptr;
    forall (p_ptr,P)
      delete p_ptr;
    P.clear();
}

/*
 *
 */

// cast macro

#define polygon		((Polygon *)obj)
 
static Point2D_ptr handle_to_move;
 
void GraphicPolygon::resize_start(World *, double X, double Y, Int_Mode) {
    handle_to_move = polygon->handle_at(X, Y);
} 

void GraphicPolygon::resize_change(World *, double X, double Y, Int_Mode) {
    if (handle_to_move)
      handle_to_move->setXY(X, Y);
} 

void GraphicPolygon::resize_end(World *, double X, double Y, Int_Mode) {
    if (handle_to_move) {
      handle_to_move->setXY(X, Y);
      polygon->subtype(Generic_t);
    }
} 








