/*
    Copyright (C) 1992-1994
                    Cesar N. Gon

    All rights reserved.

    This program is  free software; you can redistribute unmodified
    versions of it as long as you do not  profit from that, neither
    by selling it nor  by   including it  with other products   for
    sale.  This copyright  notice  must  remain in all  distributed
    copies.

    You are  allowed to modify your  personal copy  of this program
    (as long as you do not distribute any changed versions).

    You may,  of course, distribute  along with it, your own source
    code or libraries of algorithms that run under GeoLab.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

    If you must contact the authors, you may write to:
                     geolab-request@dcc.unicamp.br
*/

/*
 *
 * Polygon.h - definicao das classes Polygon e GraphicPolygon.
 *
 *             link flag -lm
 *
 *
 *    	Data	Hora  	Nome     		Modificacao/Motivo
 *
 *    12.02.92  13:48 	Cesar N. Gon  	         Arquivo criado
 *
 */

#ifndef _POLYGON_DEFINED
#define _POLYGON_DEFINED

#include <iostream.h>
#include <stdio.h>
#include "global.h"
#include "utilities.h"
#include "handle.h"
#include "World.h"
#include "Coords2D.h"
#include "Obj.h"
#include "Point2D.h"
#include "Segment2D.h"
#include "basic.h"
#include "List.h"

#ifndef _Point2D_ptr_list
#define _Point2D_ptr_list
declare(list,Point2D_ptr)
#endif ~_Point2D_ptr_list;

#ifndef _Segment2D_ptr_list
#define _Segment2D_ptr_list
declare(list,Segment2D_ptr)
#endif ~_Segment2D_ptr_list;

// forward declaration
class GraphicPolygon;
 
class Polygon : public Obj {
  public:
    // construtores
    Polygon(char *s = "Polygon", ObjType t = Generic_t, bool v = false) : Obj(s, t, v), is_closed(true) {}
    Polygon(ObjType t, bool c = true) : Obj("Polygon",t), is_closed(c) {}
    Polygon(bool c) : Obj("Polygon"), is_closed (c) {}
    Polygon(list(Point2D_ptr)& Points, ObjType t = Generic_t, bool c = true);
    Polygon( Polygon& Pol) : Obj("Polygon", Pol.stype, Pol.volati), is_closed(Pol.is_closed) {P=Pol.copy_vertices();}
    
    //destrutor
    ~Polygon() {P.clear();}

    // construtor virtual (simulado) que estabelece o relacionamento
    //
    //                  Polygon --> GraphicPolygon
    
    GraphicObj* new_GraphicObj();

    Obj* reproduct()  {
    	return new Polygon(*this); 
    }
    
    // inherited methods redefinition
    void load(FILE *);
    void save(FILE *);
       
    ObjType type()  { return Polygon_t; }
    
    void translate(double to_x, double to_y);

    void rotate(double orig_x, double orig_y, double angle);
    
    void print() ;

    void unfill_handles(World *w, bool show_handle = true) ;
    void draw_handles(World *w, GC gc, bool highlight = false, bool show_handle = true) ;
    void draw(World *w, GC gc) ;
    bool select_test(double x, double y) ;
    bool select_test(double x1, double y1, double x2, double y2) ;
    double distance_to_point(double x, double y) ;
    
    virtual Point2D_ptr handle_at(double x, double y) ;

    // Metodos necessarios para "Input Tools"
    
    virtual bool point_inside_implemented () {return subtype()!=Generic_t;}
    virtual bool point_inside(double x, double y) ;
    virtual bool bounding_box_implemented () {return true;}
    virtual void bounding_box (double& x1, double& y1, double& x2, double& y2) ;
    virtual bool get_segments_implemented () {return true;}
    virtual list(Obj_ptr) get_segments () ;
    
    // query & set functions

    bool closed () {return is_closed; }
    bool closed (bool value) {return is_closed = value; }
    
    // others methods

    int length () {return P.length();}
    bool empty () {return P.empty();}
    list_item first () {return P.first();}
    list_item last  () {return P.last();}
    list_item insert (Point2D_ptr p, list_item it, rel_pos dir = after) {
	return P.insert(p,it,dir);
    }
    Point2D_ptr head () {return P.head();}
    Point2D_ptr tail () {return P.tail();}
    
    list_item push (Point2D_ptr p) {return P.push(p);}
    Point2D_ptr pop  () {return P.pop();}

    list_item succ (list_item it) {
	if (is_closed) return P.cyclic_succ(it);
	else return P.succ(it);
    }
      
    list_item pred (list_item it) {
	if (is_closed) return P.cyclic_pred(it);
	else return P.pred(it);
    }

    void permute() {P.permute();subtype(Generic_t);}

    int how_many_intersections ( Segment2D s) ;

    void star_shaped();

    void convex();
    
    void simple();

    void star_shaped( Point2D& ref);

    bool is_simple() ;
    
    bool is_convex() ;

    list(Point2D_ptr)& vertices () {return P;};

    list(Segment2D_ptr) copy_edges () ;

    list(Point2D_ptr) copy_vertices () ;

    void clear();
    
  protected:
    list(Point2D_ptr) P;
    bool is_closed;
};

/*
 *
 * GraphicPolygon
 *
 *
 */

class GraphicPolygon : public GraphicObj {
  public:
    // construtores
    GraphicPolygon() : GraphicObj(new Polygon()) {}
    GraphicPolygon(Polygon *Pol) : GraphicObj(Pol) {}
    
    //destrutor
    ~GraphicPolygon() {}
    
    // inherited methods redefinition
//    void int_start(World *, double initial_X, double initial_Y, Int_Mode);
//    void int_change(World *, double new_X, double new_Y, Int_Mode);
//    void int_end(World *, double final_X, double final_Y, Int_Mode);
    
    void resize_start(World *, double X, double Y, Int_Mode);
    void resize_change(World *, double X, double Y, Int_Mode);
    void resize_end(World *, double X, double Y, Int_Mode);
};

typedef Polygon* Polygon_ptr;
typedef GraphicPolygon* GraphicPolygon_ptr;

#endif ~_POLYGON_DEFINED;





