/*
    Copyright (C) 1992-1994
                    Welson R. Jacometti

    All rights reserved.

    This program is  free software; you can redistribute unmodified
    versions of it as long as you do not  profit from that, neither
    by selling it nor  by   including it  with other products   for
    sale.  This copyright  notice  must  remain in all  distributed
    copies.

    You are  allowed to modify your  personal copy  of this program
    (as long as you do not distribute any changed versions).

    You may,  of course, distribute  along with it, your own source
    code or libraries of algorithms that run under GeoLab.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

    If you must contact the authors, you may write to:
                     geolab-request@dcc.unicamp.br
*/

/*
 *
 * Real.h
 *
 *    	Data	Hora  	Nome     		Modificacao/Motivo
 *
 *    20.01.92  13:40 	Welson R. Jacometti  	Arquivo criado
 *
 */

#ifndef _REAL_DEFINED
#define _REAL_DEFINED

#include <stdio.h>
#include <iostream.h>
#include "global.h"
#include "World.h"
#include "Obj.h"
#include "ExternalObj_id.h"	// obrigatorio em todo objeto externo

// forward declaration
class GraphicReal;
 
class Real : public Obj {
  public:
    // construtores
    Real(char *s = "Real", ObjType t = Generic_t, bool v = false) : Obj(s, t, v) { set(0); }
    Real(double d) : Obj("Real") { set(d); }
    Real( Real &r) : Obj(r.what_is(), r.stype, r.volati) { set(r.get()); }
        
    //destrutor
    ~Real() {}

    // construtor virtual (simulado) que estabelece o relacionamento
    //
    //                   Real --> GraphicReal
    
    GraphicObj* new_GraphicObj();
    
    // funcao para auto-reproducao de um objeto - necessario para possibilitar 
    // que a interface manipule objetos externos ao geolab (alem disso torna
    // mais homogeneo o tratamento de objetos conhecidos e desconhecidos)

    Obj* reproduct()  {
    	return new Real(*this);
    }
        
    // query & set functions
    void set(double d) { val = d; }
    double get()  { return val; }
    
    // inherited methods redefinition
    ObjType type()  { return Real_t; }
    
    void print()  { cout << "Real: " << val << endl; }

    void draw(World *w, GC gc = 0) ;

  protected:
    double val;
};

/*
 *
 * GraphicReal
 *
 *
 */

class GraphicReal : public GraphicObj {
  public:
    // construtores
    GraphicReal() : GraphicObj(new Real()) {}
    GraphicReal(Real *r) : GraphicObj(r) {}
    
    //destrutor
    ~GraphicReal() {}
    
    // inherited methods redefinition
    void draw(World *, DrawMode);
};

typedef Real* Real_ptr;
typedef GraphicReal* GraphicReal_ptr;

#endif ~_REAL_DEFINED;




