/*
    Copyright (C) 1992-1994
                    Welson R. Jacometti

    All rights reserved.

    This program is  free software; you can redistribute unmodified
    versions of it as long as you do not  profit from that, neither
    by selling it nor  by   including it  with other products   for
    sale.  This copyright  notice  must  remain in all  distributed
    copies.

    You are  allowed to modify your  personal copy  of this program
    (as long as you do not distribute any changed versions).

    You may,  of course, distribute  along with it, your own source
    code or libraries of algorithms that run under GeoLab.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

    If you must contact the authors, you may write to:
                     geolab-request@dcc.unicamp.br
*/

/*
 *
 * Rectangle.h - definicao das classes Rectangle e GraphicRectangle
 *
 *               link flag -lm
 *
 *
 *    	Data	Hora  	Nome     		Modificacao/Motivo
 *
 *    07.01.92  13:40 	Welson R. Jacometti  	Arquivo criado
 *
 */

#ifndef _RECTANGLE_DEFINED
#define _RECTANGLE_DEFINED

#include <iostream.h>
#include <stdio.h>
#include <math.h>
#include "global.h"
#include "World.h"
#include "Obj.h"
#include "Point2D.h"
#include "Segment2D.h"

// forward declaration
class GraphicRectangle;

class Rectangle : public Obj {
  public:
    // construtores
    Rectangle(char *s = "Rectangle", ObjType t = Generic_t, bool v = false) : Obj(s, t, v) { set(0, 0, 0, 0); }
    Rectangle(double x1, double y1, double x2, double y2) : Obj("Rectangle") { set(x1, y1, x2, y2); }
    Rectangle( Point2D &a,  Point2D &b) : Obj("Rectangle") { set(a, b); }
    Rectangle( Point2D &p, double width, double height) : Obj("Rectangle") { set(p, width, height); }
    Rectangle(Rectangle &r) : Obj(r.what_is(), r.stype, r.volati) { set(r.P1(), r.P2()); }
    
    //destrutor
    ~Rectangle() {}				
    
    // construtor virtual (simulado) que estabelece o relacionamento
    //
    //                 Rectangle --> GraphicRectangle
    
    GraphicObj* new_GraphicObj();
    
    // funcao para auto-reproducao de um objeto - necessario para possibilitar 
    // que a interface manipule objetos externos ao geolab (alem disso torna
    // mais homogeneo o tratamento de objetos conhecidos e desconhecidos)

    Obj* reproduct()  {
    	return new Rectangle(*this);
    }
        
    // query & set functions
    Point2D& P1() { return p1; }
    Point2D& P2() { return p2; }
    Point2D& P3() { set_handles(); return p3; }
    Point2D& P4() { set_handles(); return p4; }

    virtual void setP1(double x, double y) { p1.setXY(x, y); }
    virtual void setP2(double x, double y) { p2.setXY(x, y); }

    virtual void set(double x1, double y1, double x2, double y2) {
    	setP1(x1, y1);
    	setP2(x2, y2);
    }

    virtual void set(Point2D a, Point2D b) {
    	p1 = a;
    	p2 = b;
    }

    virtual void set(Point2D p, double width, double height) {
    	p1 = p;
    	p2.setXY(p1.X() + width, p1.Y() + height);
    }
    
    double width() { return fabs(P1().X() - P2().X()); }
    double height() { return fabs(P1().Y() - P2().Y()); }
    
    bool equal(Rectangle &r)  {
    	bool same_x = ::equal(min(r.P1().X(), r.P2().X()), min(P1().X(), P2().X()));
    	bool same_y = ::equal(min(r.P1().Y(), r.P2().Y()), min(P1().Y(), P2().Y()));
    	
        return same_x && same_y && ::equal(r.width(), width()) && ::equal(r.height(), height());
    }

    bool operator == (Rectangle &r)  { return equal(r); }


    // Metodos necessarios para "Input Tools"
    
    virtual bool point_inside_implemented () {return true;}
    virtual bool point_inside(double x, double y) ;
    virtual bool bounding_box_implemented () {return true;}
    virtual void bounding_box (double& x1, double& y1, double& x2, double& y2) ;
    virtual bool get_segments_implemented () {return true;}
    virtual list(Obj_ptr) get_segments () ;
        
    // geometric functions

    bool is_square() { return ::equal(width(), height()); }
        
    // inherited methods redefinition
    void load(FILE *f) { Obj::load(f); p1.load(f); p2.load(f); }
    void save(FILE *f) { Obj::save(f); p1.save(f); p2.save(f); }

    ObjType type()  { return Rectangle_t; }
    
    void translate(double to_x, double to_y) {
        p1.translate(to_x, to_y);
        p2.translate(to_x, to_y);
    }

    void rotate(double orig_x, double orig_y, double angle) {
        p1.rotate(orig_x, orig_y, angle);
	p2.rotate(orig_x, orig_y, angle);
    }
    
    void print()  { cout << "Rectangle vertices : (" << p1.X() << "," << p1.Y() << ") (" << 
                                                             p2.X() << "," << p2.Y() << ") (" <<
                                                             p3.X() << "," << p3.Y() << ") (" << 
                                                             p4.X() << "," << p4.Y() << ")" << endl; }

    void unfill_handles(World *w, bool show_handle = true) ;
    void draw_handles(World *w, GC gc, bool highlight = false, bool show_handle = true) ;
    void draw(World *w, GC gc) ;

    bool select_test(double x, double y) {
    	set_handles();
        return p1.select_test(x, y) || p2.select_test(x, y) || p3.select_test(x, y) || p4.select_test(x, y);
    }
    
    bool select_test(double x1, double y1, double x2, double y2) {
    	set_handles();
        return p1.select_test(x1, y1, x2, y2) && p2.select_test(x1, y1, x2, y2) && p3.select_test(x1, y1, x2, y2) && p4.select_test(x1, y1, x2, y2);
    }

    double distance_to_point(double x, double y)  {
    	set_handles();
    	return min(Segment2D(p1, p3).distance_to_point(x, y), 
    	         min(Segment2D(p3, p2).distance_to_point(x, y), 
    	           min(Segment2D(p2, p4).distance_to_point(x, y), Segment2D(p4, p1).distance_to_point(x, y))));
    }

    // Methods for information retrieval.
    virtual bool get_points_implemented() { return true; }
    virtual list(Obj_ptr) get_points() ;
    	    
  protected:
    void set_handles() {
        p3.setXY(p1.X(), p2.Y());
        p4.setXY(p2.X(), p1.Y());
    }
    
    Point2D p1, p2, p3, p4;	// p3 e p4 sao handles extras, instanciados
    				// aqui por questoes de eficiencia (tempo)
};

/*
 *
 * GraphicRectangle
 *
 *
 */

class GraphicRectangle : public GraphicObj {
  public:
    // construtores
    GraphicRectangle() : GraphicObj(new Rectangle()) {}
    GraphicRectangle(Rectangle *r) : GraphicObj(r) {}
        
    //destrutor
    ~GraphicRectangle() {}				
    
    // inherited methods redefinition
    void int_start(World *, double initial_X, double initial_Y, Int_Mode mode = LeftButton);
    void int_change(World *, double new_X, double new_Y, Int_Mode mode = LeftButton);
    void int_end(World *, double final_X, double final_Y, Int_Mode mode = LeftButton);
    
    void resize_start(World *, double X, double Y, Int_Mode mode = LeftButton);
    void resize_change(World *, double X, double Y, Int_Mode mode = LeftButton);
    void resize_end(World *, double X, double Y, Int_Mode mode = LeftButton);

    bool where_draw_string (double &x, double &y);
};

typedef Rectangle* Rectangle_ptr;
typedef GraphicRectangle* GraphicRectangle_ptr;

#endif ~_RECTANGLE_DEFINED;



