/*
    Copyright (C) 1992-1994
                    Cesar N. Gon

    All rights reserved.

    This program is  free software; you can redistribute unmodified
    versions of it as long as you do not  profit from that, neither
    by selling it nor  by   including it  with other products   for
    sale.  This copyright  notice  must  remain in all  distributed
    copies.

    You are  allowed to modify your  personal copy  of this program
    (as long as you do not distribute any changed versions).

    You may,  of course, distribute  along with it, your own source
    code or libraries of algorithms that run under GeoLab.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

    If you must contact the authors, you may write to:
                     geolab-request@dcc.unicamp.br
*/

/*
 * Segment2D_test.C
 *
 *
 *    	Data	Hora  	Nome     		Modificacao/Motivo
 *
 *    20.01.92  13:40 	Cesar N. Gon 	  	Arquivo criado
 *
 */

#include "Segment2D_test.h"
#include "Boolean.h"

bool point_in_segm(Segment2D_ptr& s,Point2D_ptr& p) {
    return s->point_in(*p);
}

bool intersect_segm(Segment2D_ptr& s1,Segment2D_ptr& s2) {
    return s1->intersect(*s2);
}

Point2D intersection_segm(Segment2D_ptr& s1,Segment2D_ptr& s2) {
    return s1->intersection(*s2);
}

/*
 * implementacao dos metodos
 *
 */
 
bool point_in::is_able_to_handle(const SelectedObjects &SO, const GraphicObjects &GO) {
    list_item it;
    int ns = 0, np = 0;
    
    forall(it, SO) {
      if (GO[it]->type() == Segment2D_t)
        ns++;
      else
        if (GO[it]->type() == Point2D_t)
          np++;
        else {
          SO.init_iterator();
      	  return false;
      	}

      if (ns > 1 || np > 1) {
      	SO.init_iterator();
      	return false;
      }
    }        
    
    if (ns == 1 && np == 1)
      return true;
    else
      return false;
}

ResultObjects point_in::binder(const SelectedObjects& SO, const GraphicObjects& GO, World *, int, int) {
    ResultObjects Result;
    Segment2D_ptr s;
    Point2D_ptr p;
    list_item it = SO.first();
    Obj_ptr obj;
    Boolean* Bool_ptr;

    if (!SO.empty()) {
      obj = GO[SO[it]]->obj;
      if (obj->type() == Segment2D_t)
        s = (Segment2D_ptr)obj;
      else
        p = (Point2D_ptr)obj;
      obj = GO[SO[SO.succ(it)]]->obj;
      if (obj->type() == Segment2D_t)
        s = (Segment2D_ptr)obj;
      else
        p = (Point2D_ptr)obj;
      Bool_ptr = new Boolean (point_in_segm(s,p));
      Bool_ptr->obj_volatile(true);
      Result.push(Bool_ptr);
    }
    
    return Result;
}

bool intersect::is_able_to_handle(const SelectedObjects &SO, const GraphicObjects &GO) {
    list_item it;
    int ns = 0;
    
    forall(it, SO) {
      if (GO[it]->type() != Segment2D_t) {
      	SO.init_iterator();
      	return false;
      }
      else
        ns++;
    }
    
    if (ns == 2)
      return true;
    else
      return false;
}

ResultObjects intersect::binder(const SelectedObjects& SO, const GraphicObjects& GO, World *, int, int) {
    ResultObjects Result;
    Segment2D_ptr s1,s2;
    list_item it = SO.first();
    Boolean* Bool_ptr;

    if (!SO.empty()) {
      s1 = (Segment2D_ptr)GO[SO[it]]->obj;
      s2 = (Segment2D_ptr)GO[SO[SO.succ(it)]]->obj;
      Bool_ptr = new Boolean (intersect_segm(s1,s2));
      Bool_ptr->obj_volatile(true);
      Result.push(Bool_ptr);
    }

    return Result;
}

ResultObjects intersection::binder(const SelectedObjects& SO, const GraphicObjects& GO, World *, int, int) {
    ResultObjects Result;
    Segment2D_ptr s1,s2;
    list_item it = SO.first();
    Point2D_ptr p;
   
    if (!SO.empty()) {
      s1 = (Segment2D_ptr)GO[SO[it]]->obj;
      s2 = (Segment2D_ptr)GO[SO[SO.succ(it)]]->obj;
      if (s1->intersect(*s2)) {
        p = new Point2D (intersection_segm(s1,s2));
        Result.push(p);
      }
    }

    return Result;
}

Algorithm* point_in_link() {
    return new point_in();
}

Algorithm* intersect_link() {
    return new intersect();
}

Algorithm* intersection_link() {
    return new intersection();
}

