/*
    Copyright (C) 1992-1994
                    Welson R. Jacometti

    All rights reserved.

    This program is  free software; you can redistribute unmodified
    versions of it as long as you do not  profit from that, neither
    by selling it nor  by   including it  with other products   for
    sale.  This copyright  notice  must  remain in all  distributed
    copies.

    You are  allowed to modify your  personal copy  of this program
    (as long as you do not distribute any changed versions).

    You may,  of course, distribute  along with it, your own source
    code or libraries of algorithms that run under GeoLab.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

    If you must contact the authors, you may write to:
                     geolab-request@dcc.unicamp.br
*/

/*
 *
 * SpanningTree.C - implementacao dos metodos classes SpanningTree e 
 *                  GraphicSpanningTree.
 *
 *
 *    	Data	Hora  	Nome     		Modificacao/Motivo
 *
 *    18.02.92  12:10 	Welson R. Jacometti  	Arquivo criado
 *
 */

#include "Obj.h"
#include "Point2D.h"
#include "Segment2D.h"
#include "SpanningTree.h"

GraphicObj* SpanningTree::new_GraphicObj() {
    return new GraphicSpanningTree(this);
}

void SpanningTree::translate(double to_x, double to_y) {
    node n;
    
    forall_nodes(n, spanning) {
    	((Point2D *)spanning.inf(n))->translate(to_x, to_y);
    }
}

void SpanningTree::print()  {
    cout << "SpanningTree (uses LEDA printout function for graphs):" << endl;
    spanning.print();
}
	
void SpanningTree::unfill_handles(World *w, bool show_handle)  {
    if (show_handle) {
      node n;
      forall_nodes(n, spanning) {
        ((Point2D *)spanning.inf(n))->unfill_handles(w);
      }
    }
}

void SpanningTree::draw_handles(World *w, GC gc, bool highlight, bool show_handle)  {
    if (show_handle) {
      node n;
      forall_nodes(n, spanning) {
        ((Point2D *)spanning.inf(n))->draw_handles(w, gc, highlight);
      }
    }
}

void SpanningTree::draw(World *w, GC gc)  {
    edge e;
    forall_edges(e, spanning) {
      node i = spanning.source(e);
      node f = spanning.target(e);
      Point2D p1 = *((Point2D_ptr)(spanning.inf(i)));
      Point2D p2 = *((Point2D_ptr)(spanning.inf(f)));
      Segment2D(p1, p2).draw(w, gc);
    }
}

bool SpanningTree::select_test(double x, double y)  {
    bool res = false;
    node n;
    forall_nodes(n, spanning) {
      if (((Point2D *)spanning.inf(n))->select_test(x, y)) {
        res = true;
        spanning.reset();
        break;
      }
    }
    return res;
}

bool SpanningTree::select_test(double x1, double y1, double x2, double y2)  {
    bool res = true;
    node n;
    forall_nodes(n, spanning) {
      if (!((Point2D *)spanning.inf(n))->select_test(x1, y1, x2, y2)) {
        res = false;
        spanning.reset();
        break;
      }
    }
    return res;
}

double SpanningTree::distance_to_point(double x, double y)  {
    double res = TOO_FAR;
    edge e;
    
    forall_edges(e, spanning) {
      double aux;
      node i = spanning.source(e);
      node f = spanning.target(e);
      Point2D p1 = *((Point2D_ptr)(spanning.inf(i)));
      Point2D p2 = *((Point2D_ptr)(spanning.inf(f)));
      aux = Segment2D(p1, p2).distance_to_point(x, y);
      if (aux < res)
        res = aux;
    }
    
    return res;
}

/*
 *
 */
 
// cast macro

#define spanning_tree	((SpanningTree *)obj)

void GraphicSpanningTree::resize_start(World *, double X, double Y, Int_Mode) {
}

void GraphicSpanningTree::resize_change(World *, double X, double Y, Int_Mode) {
}

void GraphicSpanningTree::resize_end(World *, double X, double Y, Int_Mode) {
}







