/*
    Copyright (C) 1992-1994
                    Welson R. Jacometti

    All rights reserved.

    This program is  free software; you can redistribute unmodified
    versions of it as long as you do not  profit from that, neither
    by selling it nor  by   including it  with other products   for
    sale.  This copyright  notice  must  remain in all  distributed
    copies.

    You are  allowed to modify your  personal copy  of this program
    (as long as you do not distribute any changed versions).

    You may,  of course, distribute  along with it, your own source
    code or libraries of algorithms that run under GeoLab.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

    If you must contact the authors, you may write to:
                     geolab-request@dcc.unicamp.br
*/

/*
 *
 * SpanningTree.h - definicao das classes SpanningTree e GraphicSpanningTree.
 *
 *
 *    	Data	Hora  	Nome     		Modificacao/Motivo
 *
 *    18.02.92  12:10 	Welson R. Jacometti  	Arquivo criado
 *
 */

#ifndef _SPANNING_TREE_DEFINED
#define _SPANNING_TREE_DEFINED

#include "global.h"
#include "Obj.h"
#include "basic.h"
#include "Graph.h"
#include "Point2D.h"
#include "Segment2D.h"
#include "ExternalObj_id.h"

#ifndef _Point2D_ptrent_UGRAPH
#define _Point2D_ptrent_UGRAPH
declare2 (UGRAPH,Point2D_ptr,ent)
#endif ~_Point2D_ptrent_UGRAPH;

class SpanningTree : public Obj {
  public:
    // construtores
    SpanningTree(char *s = "SpanningTree", ObjType t = Generic_t, bool v = false) : Obj(s, t, v) {}
    SpanningTree( SpanningTree &e) : Obj(e.what_is(), e.stype, e.volati) { spanning = e.get_graph(); }
    
    // destrutor
    ~SpanningTree() {}
    
    // construtor virtual (simulado) que estabelece o relacionamento
    //
    //               SpanningTree --> GraphicSpanningTree
    
    virtual GraphicObj* new_GraphicObj();

    // funcao para auto-reproducao de um objeto -> necessario quando
    // queremos, por exemplo, copiar um objeto na interface e este
    // objeto e' externo ao GeoLab.

    virtual Obj* reproduct()  {
        return new SpanningTree(*this);
    }

    // query & set methods
    
    graph get_graph()  { return spanning; }
    void set_graph( graph &g) { spanning = g; }
    
    // inherited methods redefinition
    ObjType type()  { return SpanningTree_t; }

    void translate(double to_x, double to_y);
    
    void print() ;
    void unfill_handles(World *w, bool show_handle = true) ;
    void draw_handles(World *w, GC gc, bool highlight = false, bool show_handle = true) ;
    void draw(World *, GC) ;

    bool select_test(double x, double y) ;
    bool select_test(double x1, double y1, double x2, double y2) ;

    double distance_to_point(double x, double y) ;
                
  protected:
    UGRAPH(Point2D_ptr,ent) spanning;
};

/*
 *
 * GraphicSpanningTree
 *
 *
 */

class GraphicSpanningTree : public GraphicObj {
  public:
    // construtores
    GraphicSpanningTree() : GraphicObj(new SpanningTree()) {}
    GraphicSpanningTree(SpanningTree *e) : GraphicObj(e) {}
    
    //destrutor
    ~GraphicSpanningTree() {}
    
    // inherited methods redefinition

    // void load(FILE *, World *) {warning(what_is(),"load");}
    // void save(FILE *) {warning(what_is(),"save");}
    
    void resize_start(World *, double X, double Y, Int_Mode);
    void resize_change(World *, double X, double Y, Int_Mode);
    void resize_end(World *, double X, double Y, Int_Mode);
};

typedef SpanningTree* SpanningTree_ptr;
typedef GraphicSpanningTree* GraphicSpanningTree_ptr;

#endif ~_SpanningTree;









