/*
    Copyright (C) 1992-1994
                    Welson R. Jacometti

    All rights reserved.

    This program is  free software; you can redistribute unmodified
    versions of it as long as you do not  profit from that, neither
    by selling it nor  by   including it  with other products   for
    sale.  This copyright  notice  must  remain in all  distributed
    copies.

    You are  allowed to modify your  personal copy  of this program
    (as long as you do not distribute any changed versions).

    You may,  of course, distribute  along with it, your own source
    code or libraries of algorithms that run under GeoLab.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

    If you must contact the authors, you may write to:
                     geolab-request@dcc.unicamp.br
*/

/*
 *
 * algorithms.C - criacao dinamica do menu de algoritmos e associacao de 
 *                modulos para link dinamico.
 *
 *
 *    	Data	Hora  	Nome     		Modificacao/Motivo
 *
 *    29.01.92  21:25 	Welson R. Jacometti  	Arquivo criado
 *
 *
 *
 * IMPORTANTE:
 *
 *    O disparo de algoritmos externos e' feito unica e exclusivamente em
 *  dois modulos: algorithms.C(.h) e instantiation.C(.h) 
 *    Nestes modulos, antes da referencia ao algoritmo externo, DEVE ser
 *  feita a mudanca da rotina de tratamento de sinais, para que o usuario
 *  seja corretamente informado sobre problemas decorrentes (a) do ambiente
 *  ou (b) de um determinado algoritmo externo.
 *
 */

     
#include "algorithms.h"
#include "basic.h"
#include "Stack.h"
#include "List.h"
#include "utilities.h"
#include "geolex.h"
#include "animation.h"
#include "edit_mode.h"
#include "objects.h"
#include "trapsignal.h"
#include "benchmark.h"
#include "Obj.h"
#include "String.h"
#include "compose.h"

bool DISCARD_INPUT = false;
bool UNSELECT_INPUT = false;
bool OUTPUT_SELECTED = false;
bool DYNAMIC_DISCARD_OUTPUT = true;
bool DUMP_HEAP = false;
Algorithm* LAST_ALGORITHM = 0;

// nao exportavel - usada para agilizar o processo de "enable" algorithms

bool JUST_ENABLED = true;

// Attr_attribute FUNCTION_NAME = xv_unique_key();
// Attr_attribute MODULE_NAME   = xv_unique_key();
Attr_attribute ALG_INSTANCE  = xv_unique_key();

typedef struct menu_str {
    Menu menu;
    char *str;
} Menu_str;

typedef Menu_str* Menu_str_ptr;

typedef struct alg_handle {
    void *mhandle;
    Algorithm *alg;
    Menu_item item;
} AlgHandle;

typedef AlgHandle* AlgHandle_ptr;

extern "C" int yylex();			// linking a C module with C++

#ifndef _Menu_str_ptr_stack
#define _Menu_str_ptr_stack
declare(stack,Menu_str_ptr)
#endif ~_Menu_str_ptr_stack;

#ifndef _AlgHandle_ptr_list
#define _AlgHandle_ptr_list
declare(list,AlgHandle_ptr)
#endif ~_AlgHandle_ptr_list

list(AlgHandle_ptr) ALGHANDLES;

/*
 * Nota:
 *
 *   O algoritmo abaixo procurara' pelo arquivo MENU_FILENAME, definido em
 *   algorithms.h nos seguintes diretorios (e nessa ordem):
 *
 *   1. Diretorio corrente
 *   2. No HOME do usuario (usa getenv)
 *   3. No path especificado pela variavel de ambiente GEOLAB_MENU (se
 *      estiver definida).
 *
 *   No caso de nao encontrar o arquivo em nenhum destes diretorios, entao
 *   o menu e' construido vazio e a execucao prossegue. Se a opcao 3 for
 *   tentada sem sucesso, entao uma mensagem de erro dizendo que a variavel
 *   de ambiente ou nao esta' setada, ou esta' setada indevidamente e'
 *   emitida.
 *
 */

String compose_external_algorithm_name(String function_name) {
    return (String("_") + function_name + String("_link__Fv"));
}

Menu menu_algorithms(caddr_t *ip, Xv_opaque owner) {
    extern Menu_item algorithms_settings_handler(Menu_item, Menu_generate);
    extern Menu_item algorithms_statistics_handler(Menu_item, Menu_generate);
    
    stack(Menu_str_ptr) s;
    int token;
    Menu res, aux;
    Menu_str_ptr mstr;
    void* mhandle;
    Algorithm_ptr alg = 0;
    Menu_item mi;
    AlgHandle_ptr new_entry;
    String strerr, composite; 
    String home, menu_dir;
    char *item, *function, *mname;
    bool title_done = false, no_item = true;
    address f;
                    
    if (!(yyin = fopen(MENU_FILENAME, "r"))) {
	home = getenv("HOME");
	if (!home.value())
	  warning("algorithms.C", "menu_algorithms", "could not find HOME environment variable");
      home = home + String("/") + String(MENU_FILENAME);
	if (!(yyin = fopen(home.value(), "r"))) {
	    menu_dir = getenv(MENU_DIR);
	    if (!menu_dir.value())
	      warning("algorithms.C", "menu_algorithms", "could not find GEOLAB_MENU environment variable");
	    menu_dir = menu_dir + String("/") + String(MENU_FILENAME);
	    if (!(yyin = fopen(menu_dir.value(), "r"))) {
		warning("algorithms.C", "menu_algorithms", "could not find menu file");
		warning("algorithms.C", "menu_algorithms", "please set the environment variable GEOLAB_MENU properly");
		return (Menu)xv_create(XV_NULL, MENU_COMMAND_MENU,
				       XV_KEY_DATA, INSTANCE, ip,
				       MENU_ITEM,
				       XV_KEY_DATA, INSTANCE, ip,
				       MENU_STRING, "Settings...",
				       MENU_GEN_PROC, algorithms_settings_handler,
				       XV_NULL,
				       MENU_ITEM,
				       XV_KEY_DATA, INSTANCE, ip,
				       MENU_STRING, "Statistics...",
				       MENU_GEN_PROC, algorithms_statistics_handler,
				       XV_NULL,
				       MENU_GEN_PIN_WINDOW, owner, "Algorithms", 
				       XV_NULL);
	    }
      }
    }
    
    for(;;) {
	if ((token = yylex()) == _ALGO_ERROR || token == _ALGO_EOF)
	  break;
        
	if (token != _ALGO_TITLE && !title_done)
	  error("algorithms.C", "menu_algorithms", "TITLE missing at the beggining of the menu file");
	
	switch (token) {
	  case _ALGO_TITLE:
	    if (title_done)
	      error("algorithms.C", "menu_algorithms", "TITLE should occur only once, in the beggining of the file");
	    title_done = true;
	    res = (Menu)xv_create(XV_NULL, MENU_COMMAND_MENU, 
				  XV_KEY_DATA, INSTANCE, ip, 
				  MENU_GEN_PIN_WINDOW, owner, strdup(straux), 
				  XV_NULL);
	    break;
	    
	  case _ALGO_MENU:
	    mstr = new Menu_str;
	    mstr->str = strdup(straux);
	    mstr->menu = (Menu)xv_create(XV_NULL, MENU_COMMAND_MENU, 
					 XV_KEY_DATA, INSTANCE, ip,
					 MENU_GEN_PIN_WINDOW, owner, mstr->str, 
					 XV_NULL);
	    s.push(mstr);
	    no_item = true;
	    break;
	    
	  case _ALGO_END:
	    if (s.empty())
	      error("algorithms.C", "menu_algorithms", "END for no MENU");
	    mstr = s.pop();
	    if (!s.empty())
	      aux = s.top()->menu;
	    else
	      aux = res;
	    
	    if (no_item)
	      xv_destroy(mstr->menu);
	    else
	      (void)xv_set(aux, MENU_ITEM,
			   XV_KEY_DATA, INSTANCE, ip,
			   MENU_STRING, mstr->str, 
			   MENU_PULLRIGHT, mstr->menu, 
			   XV_NULL, 
			   XV_NULL);
	    delete mstr;
	    break;
	    
	  case _ALGO_ITEMNAME:
	    item = strdup(straux);
	    break;
	    
	  case _ALGO_ITEMFUNCTION:
	    function = strdup(straux);
	    break;
	    
	  case _ALGO_ITEMMODULE:
	    mname = strdup(straux);
	    
	    composite = compose_external_algorithm_name(function);
	    
	    mhandle = open_dynamic_module(mname);
	    
	    if (f = dlsym(mhandle, composite.value())) {
		alg = ((link_function)f)();
		alg->algorithm_name(item);
		alg->function_name(function);
		alg->module_name(mname);
	    }
	    else {
		alg = 0;
		strerr = String("can't find function ") + composite;
		warning("algorithms.C", "menu_algorithms", strerr.value());
	    }
	    
	    mi = xv_create(XV_NULL, MENUITEM, 
			   XV_KEY_DATA, INSTANCE, ip,
			   XV_KEY_DATA, ALG_INSTANCE, alg,
			   // XV_KEY_DATA, FUNCTION_NAME, function,
			   // XV_KEY_DATA, MODULE_NAME, mname,
			   MENU_STRING, item, 
			   MENU_NOTIFY_PROC, menu_algorithms_dispatcher, 
			   XV_NULL);
	    
	    new_entry = new AlgHandle;
	    new_entry->mhandle = mhandle;
	    new_entry->alg = alg;
	    new_entry->item = mi;
	    ALGHANDLES.push(new_entry);
	    
	    if (!s.empty())
	      aux = s.top()->menu;
	    else
	      aux = res;
	    
	    (void)xv_set(aux, MENU_APPEND_ITEM, mi, XV_NULL);
	    
	    no_item = false;
	    break;
	}
    }

    if (token == _ALGO_EOF && !s.empty())
      error("algorithms.C", "menu_algorithms", "unexpected end of file");
    
    if (token == _ALGO_ERROR)
      error("algorithms.C", "menu_algorithms", "error ocurred in lexical analysis");
      
    (void)xv_set(res, MENU_ITEM,
      	  			XV_KEY_DATA, INSTANCE, ip,
      	  			MENU_STRING, "Settings...",
      	  			MENU_GEN_PROC, algorithms_settings_handler,
      	  			XV_NULL,
      	  		MENU_ITEM,
      	  			XV_KEY_DATA, INSTANCE, ip,
      	  			MENU_STRING, "Statistics...",
      	  			MENU_GEN_PROC, algorithms_statistics_handler,
      	  			XV_NULL,
          	XV_NULL);
          	
    return res;
}

bool dynamic_module_exists(char *mname) {
    FILE *f;
	
    if (f = fopen(mname, "r")) {
      fclose(f);
      return true;
    }

    return false;
}

void* open_dynamic_module(char *mname) {
    String strerr, algorithms_dir = mname;
    void *aux;    

    if (!dynamic_module_exists(algorithms_dir.value())) {
      algorithms_dir = String(getenv(ALGORITHMS_DIR)) + String("/") + String(mname);
      if (!dynamic_module_exists(algorithms_dir.value())) {
        strerr = String("could not open dynamic module file '") + String(mname) + String("' (try setting GEOLAB_ALGORITHMS environment variable properly)");
        error("algorithms.C", "open_dynamic_module", strerr.value());
      }
    }

    if (!(aux = dlopen(algorithms_dir.value(), 1))) {
      strerr = String("fatal error detected by dlopen on '") + String(mname) + String("' :") + String(dlerror());
      error("algorithms.C", "open_dynamic_module", strerr.value());
    }

    return aux;
}

void close_dynamic_modules() {
    AlgHandle_ptr alg_handle;
    
    forall(alg_handle, ALGHANDLES) {
      if (alg_handle->alg)
        delete alg_handle->alg;
      dlclose(alg_handle->mhandle);
      delete alg_handle;
    }
    
    ALGHANDLES.clear();
}

void display_current_algorithm(char *algorithm) {
    extern geolab_popup_algorithms_statistics_objects geolab_popup_algorithms_statistics;
    (void)xv_set(geolab_popup_algorithms_statistics.textfield_algorithms_statistics_current, PANEL_VALUE, algorithm, XV_NULL);
}

void display_ellapsed_time(double r, double v) {
    extern geolab_popup_algorithms_statistics_objects geolab_popup_algorithms_statistics;
    static char real_time[20];
    static char virtual_time[20];

    (void)sprintf(real_time, "%.3f", r);
    (void)sprintf(virtual_time, "%.3f", v);
      
    (void)xv_set(geolab_popup_algorithms_statistics.const_message_algorithms_statistics_real_place_holder, PANEL_LABEL_STRING, real_time, XV_NULL);
    (void)xv_set(geolab_popup_algorithms_statistics.const_message_algorithms_statistics_virtual_place_holder, PANEL_LABEL_STRING, virtual_time, XV_NULL);
}

void dump_heap() {
//    mallocmap();
}

void display_memory_status() {
    extern geolab_popup_algorithms_hackers_dream_objects geolab_popup_algorithms_hackers_dream;
    struct mallinfo minfo;
    static char total[30];
    static char alloc[50];
    static char freem[30];
    
    minfo = mallinfo();

    (void)sprintf(total, "%d", minfo.arena);
    (void)sprintf(alloc, "%d   Overhead: %d", minfo.uordblks, minfo.arena - minfo.uordblks - minfo.fordblks);
    (void)sprintf(freem, "%d", minfo.fordblks);
    
    (void)xv_set(geolab_popup_algorithms_hackers_dream.const_message_algorithms_hackers_dream_total_place_holder, PANEL_LABEL_STRING, total, XV_NULL);
    (void)xv_set(geolab_popup_algorithms_hackers_dream.const_message_algorithms_hackers_dream_allocated_place_holder, PANEL_LABEL_STRING, alloc, XV_NULL);
    (void)xv_set(geolab_popup_algorithms_hackers_dream.const_message_algorithms_hackers_dream_free_place_holder, PANEL_LABEL_STRING, freem, XV_NULL);
}

void dispatcher(World *w, Algorithm *alg) {
    ResultObjects Result;
    display_current_algorithm(alg->algorithm_name());
    start_timer();
    if (animation_on())
      frame_save(w);
    set_external_signal_handler();	// prepare to call external algorithm
    Result = alg->binder(objects_get_selected_objects(), objects_get_graphic_objects(), w, 1, REFERENCE_DEPTH);
    set_internal_signal_handler();	// ok, get back
    if (animation_on())
      frame_restore(w);
    display_ellapsed_time(get_real_ellapsed_time(), get_virtual_ellapsed_time());
    if (DUMP_HEAP)
      dump_heap();
    if (DISCARD_INPUT)
      objects_delete_selected(w);
    else
      if (UNSELECT_INPUT) {
	  objects_unselect(w);
	  objects_repaint(w);
	  enable_all_algorithms();
      }
    objects_install_created_objects(w, OUTPUT_SELECTED, Result, alg->function_name(), alg->module_name());
    display_memory_status();
}

void exec_binder(World *w, Algorithm* alg) {
    LAST_ALGORITHM = alg;
    LAST_PLAY = false;
    if (animation_on())
      notify_stop();
    else 
      dispatcher(w, alg);
}

void menu_algorithms_dispatcher(Menu, Menu_item menu_item) {
    geolab_geolab_base_window_objects *ip = (geolab_geolab_base_window_objects *) xv_get(menu_item, XV_KEY_DATA, INSTANCE);
    Algorithm *alg = (Algorithm_ptr)xv_get(menu_item, XV_KEY_DATA, ALG_INSTANCE);
    if (alg) {    
	switch (alg->type()) {
	  case ALG:
	    exec_binder(ip->world, alg);
	    if (COMPOSING)
	      compose_add_algorithm(alg, DISCARD_INPUT, UNSELECT_INPUT, OUTPUT_SELECTED, alg->function_name(), alg->module_name());    
	    break;
	    
	  case EFM:
	    install_external_function_mode(ip->world,(ExternalFunctionMode *)alg);
	    break;
	  case EOM:
	    install_external_object_mode(ip->world,(ExternalObjectMode *)alg);
	    break;
	  default:
	    warning("algorithms.C", "dispatcher","unknown algorithm type");
	}
    }
    else {
	String strerr = String("no instantiation for last algorithm called");
	warning("algorithms.C", "dispatcher", strerr.value());
    }
    
    // Nota:
    //       REFERENCE_DEPTH e' uma variavel exportada pelo modulo de animacao.
    //       Seu valor representa o nivel maximo a ser animado por um dado
    //       algoritmo em funcao de seu nivel interno (ANIMATION_DEPTH). 
    //       O nivel inicial e' 1 (valor fixado pelo dispatcher), mas pode 
    //       ser alterado se for conveniente nos metodos "binder". 
    //       Internamente a qualquer algoritmo, chamadas a outros algoritmos
    //       que o compoem devem ser feitas incrementando-se o valor de
    //       ANIMATION_DEPTH, o que faz com que as profundidades diferentes
    //       tenham valores coerentes.
}

bool compose_exec_start(World *w, Algorithm *alg) {
    display_current_algorithm(">>> Composition <<<");
    start_timer();
    if (alg->is_able_to_handle(objects_get_selected_objects(), objects_get_graphic_objects())) {
      if (animation_on())
        frame_save(w);
      return true;
    }
    else
      return false;
}

void compose_exec(World *w, Algorithm *alg, bool di, bool ui, bool os, char *fn, char *mn) {
    if (alg) {
      set_external_signal_handler();	// prepare to call external algorithm
      ResultObjects Result = alg->binder(objects_get_selected_objects(), objects_get_graphic_objects(), w, 1, REFERENCE_DEPTH);
      set_internal_signal_handler();	// ok, get back
      if (di)
        objects_delete_selected_no_refresh();
      else
        if (ui)
          objects_unselect_no_refresh();
      objects_install_created_objects_no_refresh(w, os, Result, fn, mn);
    }
    else {
      String strerr = String("composition has a non-instantiated algorithm");
      warning("algorithms.C", "compose_exec", strerr.value());
    }
}

void compose_exec_finish(World *w) {
    if (animation_on())
      frame_save(w);
    if (DUMP_HEAP)
      dump_heap();
    objects_refresh(w);
    display_ellapsed_time(get_real_ellapsed_time(), get_virtual_ellapsed_time());
    display_memory_status();
    select_applyable_algorithms();
    select_applyable_mode(w);
}

bool last_algorithm_is_able_to_handle_current_objects() {
    if (LAST_ALGORITHM)
      return LAST_ALGORITHM->is_able_to_handle(objects_get_selected_objects(), objects_get_graphic_objects());
    else {
      String strerr = String("don't have an instance for the last algorithm");
      warning("algorithms.C", "last_algorithm_is_able_to_handle_current_objects", strerr.value());
    }

    return false;
}

char* last_algorithm_called_name() {
    if (LAST_ALGORITHM)
      return LAST_ALGORITHM->function_name();
    else
      return 0;
}

char* last_algorithm_called_module_name() {
    if (LAST_ALGORITHM)
      return LAST_ALGORITHM->module_name();
    else
      return 0;
}

ResultObjects last_algorithm_called(World *w) {
    ResultObjects res;
    
    if (LAST_ALGORITHM)
      return LAST_ALGORITHM->binder(objects_get_selected_objects(), objects_get_graphic_objects(), w, 1, 0);
    else {
      String strerr = String("no instantiation for last algorithm called");
      warning("algorithms.C", "last_algorithm_called", strerr.value());
    }
    
    return res;
}

Algorithm_ptr last_algorithm() {
    return LAST_ALGORITHM;
}

void install_externally_created_object(GraphicObj_ptr obj, Algorithm_ptr alg) {
    if (!geolab_knows_object(obj->type()))
      obj->set_external(alg->function_name(), alg->module_name());
    obj->unselect();
    objects_insert(obj);
}

void enable_all_algorithms() {
    if (!JUST_ENABLED) {
      AlgHandle_ptr alg_handle;
    
      forall(alg_handle, ALGHANDLES) {
        (void)xv_set(alg_handle->item, MENU_INACTIVE, FALSE, XV_NULL);
      }
      
      JUST_ENABLED = true;
    }
}

void disable_all_algorithms() {
    AlgHandle_ptr alg_handle;
    
    forall(alg_handle, ALGHANDLES) {
      (void)xv_set(alg_handle->item, MENU_INACTIVE, TRUE, XV_NULL);
    }
      
    JUST_ENABLED = false;
}

void disable_algorithms_by_animation() {
    AlgHandle_ptr alg_handle;
    
    set_external_signal_handler();
    forall(alg_handle, ALGHANDLES) {
	if (alg_handle->alg && !alg_handle->alg->animation())
	  (void)xv_set(alg_handle->item, MENU_INACTIVE, TRUE, XV_NULL);
    }
    set_internal_signal_handler();

    JUST_ENABLED = false;
}

void select_applyable_algorithms() {
    AlgHandle_ptr alg_handle;
    
    set_external_signal_handler();
    forall(alg_handle, ALGHANDLES) {
      if (alg_handle->alg && ((animation_on() && alg_handle->alg->animation()) || !animation_on()) && alg_handle->alg->is_able_to_handle(objects_get_selected_objects(), objects_get_graphic_objects()))
        (void)xv_set(alg_handle->item, MENU_INACTIVE, FALSE, XV_NULL);
      else
        (void)xv_set(alg_handle->item, MENU_INACTIVE, TRUE, XV_NULL);
    }
    set_internal_signal_handler();

    JUST_ENABLED = false;
}


     


