/*
    Copyright (C) 1992-1994
                    Cesar N. Gon

    All rights reserved.

    This program is  free software; you can redistribute unmodified
    versions of it as long as you do not  profit from that, neither
    by selling it nor  by   including it  with other products   for
    sale.  This copyright  notice  must  remain in all  distributed
    copies.

    You are  allowed to modify your  personal copy  of this program
    (as long as you do not distribute any changed versions).

    You may,  of course, distribute  along with it, your own source
    code or libraries of algorithms that run under GeoLab.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

    If you must contact the authors, you may write to:
                     geolab-request@dcc.unicamp.br
*/

/*
 * animation_tools.C
 *
 *
 *    	Data	Hora  	Nome     		Modificacao/Motivo
 *
 *    09.12.92  17:06 	Cesar N. Gon 	  	Arquivo criado
 *
 */

#include "animation_tools.h"
#include "handle.h"
#include "objects.h"

static GraphicSegment2D *anim_s;
static GraphicArrow2D *anim_a;
static GraphicLine2D *anim_l;
static GraphicCircle *anim_c;

#define anim_segment	((Segment2D *)anim_s->obj)
#define anim_arrow	((Arrow2D *)anim_a->obj)
#define anim_line	((Line2D *)anim_l->obj)
#define anim_circle	((Circle *)anim_c->obj)



void anim_delay(World* w,unsigned int  msec)
{
    assert(msec>=0);
    if (animation_on()) {
      notify_dispatch();
      w->Flush();
      
      if (msec == 0)
        for(int i = 0; i < 10; i++) {
          notify_dispatch();
          w->Flush();
        }
      else
          usleep((unsigned int)(msec*1000));
      notify_dispatch();
      w->Flush();
    }
    
}


void anim_move_object(World* w, GraphicObj* g_obj,double deltax,
		      double deltay, double speed)
{
    double dx,dy, x = 0, y = 0;

    assert(speed>0);
    
    dx = 4/fabs(w->ratio_x());
    dy = 4/fabs(w->ratio_y());

    if (deltay<0) dy = -1*dy;
    if (deltax<0) dx = -1*dx;

    do
      {
	  x+=dx; y+=dy;
	  
	  if (fabs(x) >= fabs(deltax))
	    {
		dx = deltax -1*(x - dx);
		x = deltax;
	    };
	  if (fabs(y) >= fabs(deltay))
	    {
		dy = deltay -1*(y-dy);
		y = deltay;
	    };
	  
	  g_obj->draw(w,XorMode);
	  g_obj->translate(dx,dy);
	  g_obj->draw(w,XorMode);
	  
	  anim_delay(w,(unsigned int)(560/speed));
	  
      } while ((fabs(x)<fabs(deltax))||(fabs(y)<fabs(deltay)));
}



void animation_tools_init(World *w) {
    static bool FIRST = true;
    
    if (FIRST) {
      anim_s = new GraphicSegment2D(); assert(anim_s);
      anim_a = new GraphicArrow2D(); assert(anim_a);
      anim_l = new GraphicLine2D(); assert(anim_l);
      anim_c = new GraphicCircle(); assert(anim_c);
      FIRST = false;
    }
    anim_s->set_all(w, FOREGROUND, BACKGROUND, 0, 0, false);
    anim_a->set_all(w, FOREGROUND, BACKGROUND, 0, 0, false);
    anim_l->set_all(w, FOREGROUND, BACKGROUND, 0, 0, false);
    anim_c->set_all(w, FOREGROUND, BACKGROUND, 0, 0, false);
}


void anim_show(World *w, GraphicPoint2D& p, double f)
{
    anim_circle->set(((Point2D_ptr)(p.obj))->X(),((Point2D_ptr)(p.obj))->Y(), _HX*f);
    anim_c->draw(w, XorMode);
}


void anim_show(World *w, Point2D& p, double f)
{
    anim_circle->set(p.X(),p.Y(), _HX*f);
    anim_c->draw(w, XorMode);
}


void show_arrow (World *w, Point2D& p1, Point2D& p2)
{
    anim_arrow->set(p1,p2);
    anim_a->draw(w, XorMode);
}

void show_segment (World *w, Point2D& p1, Point2D& p2)
{
    anim_segment->set(p1,p2);
    anim_s->draw(w, XorMode);
}

void draw_blank (World *w, GraphicObj_ptr obj)
{
    unsigned long fg, bg;
    obj->get_color(fg, bg);
    obj->set_color(w, get_pixel_from_value(WHITE), get_pixel_from_value(WHITE));
    obj->draw(w, CopyMode);
    objects_repaint_all_but(obj, w);
    obj->set_color(w, fg, bg);    
}





