/*
    Copyright (C) 1992-1994
                    Welson R. Jacometti

    All rights reserved.

    This program is  free software; you can redistribute unmodified
    versions of it as long as you do not  profit from that, neither
    by selling it nor  by   including it  with other products   for
    sale.  This copyright  notice  must  remain in all  distributed
    copies.

    You are  allowed to modify your  personal copy  of this program
    (as long as you do not distribute any changed versions).

    You may,  of course, distribute  along with it, your own source
    code or libraries of algorithms that run under GeoLab.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

    If you must contact the authors, you may write to:
                     geolab-request@dcc.unicamp.br
*/

/*
 *
 * edit_mode.h - modulo de gerencia dos diversos "modes" de edicao no GeoLab.
 *
 *               link flag -lm
 *
 *
 *    	Data	Hora  	Nome     		Modificacao/Motivo
 *
 *    26.09.91  18:25 	Welson R. Jacometti  	Arquivo criado
 *
 */

#ifndef _EDIT_MODE_DEFINED
#define _EDIT_MODE_DEFINED

#include <xview/xview.h>
#include <xview/panel.h>
#include <xview/svrimage.h>
#include <xview/cursor.h>
#include "global.h"
#include "World.h"
#include "Obj.h"
#include "Algorithm.h"

EXPORT ExternalObjectMode *EXTERNAL_OBJECT_MODE;
EXPORT ExternalFunctionMode *EXTERNAL_FUNCTION_MODE; 

// Mode constants (directly dependent on the interface)

#define POINT			0	// objects
#define SEGMENT			1
#define RAY			2
#define LINE			3
#define RECTANGLE		4
#define ELLIPSE			5
#define EXTRA_OBJECT		6
#define SELECT			7	// operation modes
#define MOVE_COPY		8
#define RESIZE			9
#define ZOOM			10
#define SCROLL			11
#define DYNAMIC			12
#define EXTRA_OPMODE		13

#define SQUARE			100	// extra objects (not directly in
#define CIRCLE			101	// the interface)

#define NOMODE			-99

// suporte para comunicacao com algoritmos externos interativos

extern void EF_exec_binder (World *);
extern void EO_exec_binder (World *);
extern void EF_install_externally_created_object(GraphicObj_ptr);
extern void EO_install_externally_created_object(GraphicObj_ptr);

void install_external_function_mode(World *, ExternalFunctionMode *);
void uninstall_external_function_mode(World *);
void install_external_object_mode(World *, ExternalObjectMode *);
void uninstall_external_object_mode(World *);

void select_applyable_mode(World *);

// miscelanea

void set_cursor(World *w, Xv_Cursor cursor);

EXPORT Xv_Cursor insert_cursor;
EXPORT Xv_Cursor select_cursor;
EXPORT Xv_Cursor move_cursor;
EXPORT Xv_Cursor copy_cursor;
EXPORT Xv_Cursor resizeop_cursor;
EXPORT Xv_Cursor resizecl_cursor;
EXPORT Xv_Cursor zoom_cursor;
EXPORT Xv_Cursor scroll_cursor;
EXPORT Xv_Cursor dynamic_cursor;

void init_mode(Xv_opaque panel_choice, World *w, int value);
void set_select_mode(World *world);
void set_mode(World *world, int value);
void set_mode(int value);
int get_mode();

GraphicObj *new_object();

bool compress_drag_events(World *, Xv_Window, Event *);
bool fixed_movement_events(World *, Xv_Window, Event *);
bool fixed_movement_events_points(World *, Xv_Window, Event *);
bool change_mode_square_events(World *, Xv_Window, Event *);
bool change_mode_circle_events(World *, Xv_Window, Event *);
bool change_cursor_move_copy_events(World *, Xv_Window, Event *);
bool change_cursor_resize_events(World *, Xv_Window, Event *);

#endif ~_EDIT_MODE_DEFINED;













