%{

/*
 * by Welson R. Jacometti
 *
 */

#include "geolex.h"

char straux[MAXITEMLENGTH]; 

void set_string(lex_str) 
char *lex_str;
{ 
    int length;
    for(; *lex_str == ' ' || *lex_str == '\t'; lex_str++);
    length = strlen(++lex_str) - 1;
    strncpy(straux, lex_str, length);
    straux[length] = '\0'; 
}

/* Ana Claudia Biazetti reportou que teve problemas com comentarios muito 
   longos, que nao cabiam em yytext, e aparentemente davam problemas so' 
   no yacc. A solucao encontrada foi redefinir YYLMAX(o tamanho de yytext) 
   de BUFSIZ(definido em stdio.h) para BUFSIZ*4. 
*/

#undef YYLMAX
#define YYLMAX BUFSIZ*4

%}

%start GETTITLE
%start GETMENU
%start GETEND
%start GETFUNCTION
%start GETMODULE

comments        "/*""/"*([^*/]|[^*]"/"|"*"[^/])*"*"*"*/"
string          \"([^"\n]|\\["\n])*\"
white_space     [ \t\n]
blank           [ \t]
other           .

%%

{blank}*TITLE{blank}*           BEGIN GETTITLE;

<GETTITLE>{string}		{
	  			 set_string(yytext);
/*	   			 printf("TITLE = %s\n", straux); */
				 BEGIN 0;
				 return _ALGO_TITLE; 
				}

<GETTITLE>\n                    {
				 BEGIN 0;
				 printf("Error: geolex.l: title string expected (I found a newline)\n");
				 return _ALGO_ERROR; 
				}

<GETTITLE>{other}		{
				 BEGIN 0;
				 printf("Error: geolex.l: title string expected (did you forget quotation ?)\n");
				 return _ALGO_ERROR; 
				}

{blank}*MENU{blank}*            BEGIN GETMENU;

<GETMENU>{string}		{
	  			 set_string(yytext);
/*	   			 printf("SUBMENU = %s\n", straux); */
				 BEGIN 0;
				 return _ALGO_MENU; 
				}

<GETMENU>\n                    {
				 BEGIN 0;
				 printf("Error: geolex.l: menu string expected (I found a newline)\n");
				 return _ALGO_ERROR; 
				}

<GETMENU>{other}		{
				 BEGIN 0;
				 printf("Error: geolex.l: menu string expected (did you forget quotation ?)\n");
				 return _ALGO_ERROR;
				}

{blank}*END{blank}*             BEGIN GETEND;

<GETEND>\n                     {
				 BEGIN 0;
/*				 printf("END OK\n"); */
				 return _ALGO_END; 
				}

<GETEND>{string}		{
				 BEGIN 0;
				 printf("Warning: geolex.l: END followed by a label (discarded)\n");
				 return _ALGO_END; 
				}

<GETEND>{other}	      	        {
				 BEGIN 0;
				 printf("Warning: geolex.l: END followed by something that's not a label (discarded)\n");
				 return _ALGO_END; 
				}

<GETMODULE>{blank}*{string}	{
				 set_string(yytext);
/*				 printf("MODULE = %s\n", straux); */
				 BEGIN 0;
				 return _ALGO_ITEMMODULE; 
				}

<GETMODULE>{other}		{
				 BEGIN 0;
				 printf("Error: geolex.l: Function name must be followed by a module name\n");
				 return _ALGO_ERROR; 
				}

<GETFUNCTION>{blank}*{string}	{
	  			 set_string(yytext);
/*	   			 printf("FUNCTION = %s\n", straux); */
				 BEGIN GETMODULE;
				 return _ALGO_ITEMFUNCTION;  
				}

<GETFUNCTION>{other}		{
				 BEGIN 0;
				 printf("Error: geolex.l: Item name must be followed by a function name\n");
				 return _ALGO_ERROR; 
				}

{string}			{
	  			 set_string(yytext);
/*	   			 printf("ITEM = %s\n", straux); */
				 BEGIN GETFUNCTION;
				 return _ALGO_ITEMNAME;  
				}

{comments}			{
/*				 printf("Comments OK\n"); */
				}

{blank}*			;

{white_space}			;

{other}				{
				 BEGIN 0;
				 printf("%s << Error: geolex.l: this was not expected\n", yytext);
				 return _ALGO_ERROR; 
				}
%%





