/*
    Copyright (C) 1992-1994
		    Juliano L. de Oliveira

    All rights reserved.

    This program is  free software; you can redistribute unmodified
    versions of it as long as you do not  profit from that, neither
    by selling it nor  by   including it  with other products   for
    sale.  This copyright  notice  must  remain in all  distributed
    copies.

    You are  allowed to modify your  personal copy  of this program
    (as long as you do not distribute any changed versions).

    You may,  of course, distribute  along with it, your own source
    code or libraries of algorithms that run under GeoLab.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

    If you must contact the authors, you may write to:
                     geolab-request@dcc.unicamp.br
*/

//
// goodies_file.C 
//

#include <sys/types.h>
#include <stdio.h>
#include <pwd.h>
#include <dirent.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <sys/timeb.h>
#include <xview/xview.h>
#include <xview/panel.h>
#include "goodies_file.h"


extern	goodies_file_window_objects  *goodies_file_window;

int	file_waiting		= 0;
int	file_new_selection	= 0;
int	file_last_op 		= 0;

char	file_global_string[MAXPATHLEN];

struct	itimerval	file_timer;

const	int MAX_NUM_FILES = 1000;

char	*fname[MAX_NUM_FILES];

//
// Get file_type: -1=inexistent file; 0=invalid file; 1=dir; 2=file_ok
//

void
file_window_get_file_type(char *string)
{
	struct	passwd	*pwdp;
	struct 	stat	stbuf;
	FILE 	*fp;
	int	k, 	j,	i;
	char 	c, file[MAXPATHLEN], name_aux[MAXPATHLEN];

	strcpy(file, string);

	k = (strlen(file));
	i = k - 1;
	
	if ((i > 0) && (file[i] == '/'))
		file[i] = '\0';

	if ((k == 0) || (! strcmp(file, ".")) || (! strcmp(file, "./")))
	// remain in the same directory
		strcpy(file, goodies_file_window->file_dir_name);

	if ((! strcmp(file, "..")) || (! strcmp(file, "../")))
	{
	// go up 1 directory level
		strcpy(file, goodies_file_window->file_dir_name);
		j = (strlen(file) - 1);

		while((file[j] != '/') && (j >= 0))
			j--;

		if (j > 0)
			file[j] = '\0';
		else
			strcpy(file, "/");
	}

	if ((file[0] == '.') && (file[1] == '/'))
	{
	// select file in the current directory
		for (j = 0; j <= (k - 2); j++)
			file[j] = file[j + 2];
	}

	if ((k > 2) && (file[0] == '.')
	&&  (file[1] == '.') && (file[2] == '/'))
	{
	// select file in the father directory
		strcpy(name_aux, file);

		for (j = 0; j <= (k - 2); j++)
			name_aux[j] = name_aux[j + 2];

		strcpy(file, goodies_file_window->file_dir_name);
		j = (strlen(file) - 1);

		while((file[j] != '/') && (j >= 0))
			j--;

		if (j > 0)
			file[j] = '\0';
		else
			strcpy(file, "");

		strcat(file, name_aux);
	}

	if (file[0] == '~')
	{
	// get user home directory
		strcpy(name_aux, "");
		j = 1;
		i = 0;
		while ((j < k) && (file[j] != '/'))
			name_aux[i++] = file[j++];

		if (! strcmp(name_aux, ""))
		{
			char *nm = getenv("USER");
			strcpy(name_aux, nm);
		}
		else
			name_aux[i] = '\0';

		pwdp = getpwnam(name_aux);

		if (pwdp)
		{
			if (j == k)
				strcpy(file, pwdp->pw_dir);
			else
			{
				strcpy(name_aux, file);
				sprintf(file,"%s%s",pwdp->pw_dir, name_aux+j);
			}
		}
	}

	if ((file[0] != '/'))
	{
	// file does not have a path
		strcpy(name_aux, file);

		if (strcmp(goodies_file_window->file_dir_name, "/"))
			sprintf(file,"%s/%s", goodies_file_window->file_dir_name, name_aux);
		else
			sprintf(file,"/%s", name_aux);
	}

	k = strlen(file);
	j = 0;

	while((file[k] != '/') && (k >= 0))
		name_aux[j++] = file[k--];

	if (k < 0)
	{
		strcpy(goodies_file_window->file_fd_file_name, file);
		strcpy(goodies_file_window->file_fd_path_name, goodies_file_window->file_dir_name);
	}
	else
	{
		strcpy(goodies_file_window->file_fd_path_name,file);
		goodies_file_window->file_fd_path_name[k] = '\0';
		i = 0;

		while(j >= 0)
			goodies_file_window->file_fd_file_name[i++] = name_aux[--j];
	}

	if (stat(file, &stbuf) != 0)
	{
	//file not exists
		goodies_file_window->file_fd_file_type = -1;
	}
	else
	{
	// file exists
		if ((stbuf.st_mode & S_IFMT) == S_IFDIR)
		{
		// file is a directory
			goodies_file_window->file_fd_file_type = 1; 	
			strcpy(goodies_file_window->file_fd_file_name,"");
			strcpy(goodies_file_window->file_fd_path_name,file);
		}
		else
		{
		// test the magic number...
			goodies_file_window->file_fd_file_type = 2; 	

			fp = fopen(file, "r");

			if (fp)
			{
				for (i = 0; i < 4; i++)
				{
					c = fgetc(fp);

					if ((goodies_file_window->file_magic_number[i] != '*')
					&&  (goodies_file_window->file_magic_number[i] != c))
					// file is not of the given type
						goodies_file_window->file_fd_file_type = 0; 
				}

				fclose(fp);
			}
			else
			// file cannot be opened
				goodies_file_window->file_fd_file_type = 0;
		}
	}
}

//
// file name comparison function to qsort
//
int namecomp(const void *i, const void *j)
{
	int 	*a  = (int *) i;
	int 	*b  = (int *) j;
	int	result = strcmp(fname[*a], fname[*b]);
	return (result);
}

//
// Select files for the file window list
//

void
file_window_select_files()
{
	extern 	int	file_last_op;
	struct 	dirent 	*dir_struct;
	DIR    	*dp;
	int    	i,	dir_count,	file_count;
	char   	file[MAXPATHLEN];

	xv_set(goodies_file_window->file_list, XV_SHOW, FALSE, NULL);

	for (i = 0; i < xv_get(goodies_file_window->file_list,PANEL_LIST_NROWS); i++)
	{
		if (xv_get(goodies_file_window->file_list, PANEL_LIST_SELECTED, i))
			xv_set(goodies_file_window->file_list,PANEL_LIST_SELECT, i, FALSE, NULL);
	}

	file_last_op = 0;

	i = (int) xv_get(goodies_file_window->file_list, PANEL_LIST_NROWS);

	while (i)
	{
		xv_set(goodies_file_window->file_list,
			PANEL_LIST_DELETE, --i, NULL);
	}

	xv_set(goodies_file_window->file_window,
		FRAME_RIGHT_FOOTER, goodies_file_window->file_dir_name, NULL);

	if (! (dp = opendir(goodies_file_window->file_dir_name)))
	{
		xv_set(goodies_file_window->file_window,
			FRAME_LEFT_FOOTER, "Cannot open directory", NULL);

		return;
	}

	int nfiles = 0;

	dir_count = file_count = 0;

	if (strcmp(goodies_file_window->file_dir_name, "/"))
	{
		dir_count++;
		fname[nfiles] = new char[4];
		strcpy(fname[nfiles++], "../");
	}

	while (dir_struct = readdir(dp))
	{
		if (nfiles == MAX_NUM_FILES)
		{
			fname[nfiles] = new char[20];
			strcpy(fname[nfiles++],
				"Too many files...");
			break;
		}

		if (dir_struct->d_ino == 0)	// slot not in use
			continue;

		if (strcmp(dir_struct->d_name, ".")  == 0
		||  strcmp(dir_struct->d_name, "..") == 0)
			continue;

		sprintf(file,"%s/%s",
			goodies_file_window->file_dir_name,
			dir_struct->d_name);

		file_window_get_file_type(file);

		if ((goodies_file_window->file_fd_file_type == 1))
		{
			sprintf(goodies_file_window->file_fd_path_name, "%s/",dir_struct->d_name);
			strcpy(file, goodies_file_window->file_fd_path_name);
			dir_count++;
			fname[nfiles] = new char[strlen(file) + 1];
			strcpy(fname[nfiles++], file);
		}
		else

			if ((goodies_file_window->file_fd_file_type == 2))
			{
				strcpy(file, dir_struct->d_name);
				file_count++;
				fname[nfiles] = new char[strlen(file) + 1];
				strcpy(fname[nfiles++], file);
			}
			else
				continue;

	}
	--nfiles;
	int ind[MAX_NUM_FILES];

	for (i = 0; i <= nfiles; i++)
		ind[i] = i;

	qsort(ind, (nfiles + 1), sizeof(int), namecomp);

	i = 0;

	for (i = 0; i <= nfiles; i++)
	{
		strcpy(file, fname[ind[i]]);

		xv_set(goodies_file_window->file_list,
			PANEL_LIST_INSERT, i,
			PANEL_LIST_STRING, i, file,
			NULL);
	}

	char 	msg1[30];

	if (dir_count == 1)
	{
		if (file_count == 1)
			sprintf(msg1,"%d directory, %d file", dir_count, file_count);
		else
			sprintf(msg1,"%d directory, %d files", dir_count, file_count);
	}
	else
	{
		if (file_count == 1)
			sprintf(msg1,"%d directories, %d file", dir_count, file_count);
		else
			sprintf(msg1,"%d directories, %d files", dir_count, file_count);
	}

	xv_set(goodies_file_window->file_msg1,
		PANEL_LABEL_STRING, msg1, NULL);

	xv_set(goodies_file_window->file_list, XV_SHOW, TRUE, NULL);

	closedir(dp);
}

//
// Call file_function for the chosen file or change to the chosen directory
//
void
file_window_open_file(char *string)
{
	char file[MAXPATHLEN];

	strcpy(file, string);

	file_window_get_file_type(file);

	switch (goodies_file_window->file_fd_file_type)
	{
	case -1:
		if (goodies_file_window->file_new_file)
		{
			strcpy(file, goodies_file_window->file_fd_file_name);
			file_window_get_file_type(goodies_file_window->file_fd_path_name);

			if ((goodies_file_window->file_fd_file_type == 1)
			&&  (strcmp(file," ")))
			{
				(*(goodies_file_window->file_file_func))
					(file,
				 	 goodies_file_window->file_fd_path_name);

				xv_destroy_safe(goodies_file_window->file_window);
				delete goodies_file_window;
				goodies_file_window = NULL;
			}
			else
			{
			xv_set(goodies_file_window->file_window,
				FRAME_LEFT_FOOTER,"Invalid File or Dir", NULL);
			}
		}
		else
		{
		xv_set(goodies_file_window->file_window,
			FRAME_LEFT_FOOTER, "Unknown File or Dir", NULL);
		}

		break;

	case 0:
		xv_set(goodies_file_window->file_window,
			FRAME_LEFT_FOOTER, "Invalid File Type", NULL);
		break;

	case 1:// file is a directory

		strcpy(goodies_file_window->file_dir_name,
			goodies_file_window->file_fd_path_name);

		file_window_select_files();

		xv_set(goodies_file_window->file_textfield,
			PANEL_VALUE, "", NULL);

		break;

	case 2: // call file function for the selected file

		(*(goodies_file_window->file_file_func))
				(goodies_file_window->file_fd_file_name,
				 goodies_file_window->file_fd_path_name);

		xv_destroy_safe(goodies_file_window->file_window);
		delete goodies_file_window;
		goodies_file_window = NULL;

		break;
	}
}


//
// Done callback function for `file_window'.
//
void
file_window_done(Frame frame)
{
	xv_destroy_safe(goodies_file_window->file_window);
	delete goodies_file_window;
	goodies_file_window = NULL;
}


//
// Event callback function for `file_window'.
//
Notify_value
file_window_events(Xv_window win, Event *event, Notify_arg arg, Notify_event_type type)
{
	if (event_id(event) == WIN_RESIZE)
	{

		xv_set(goodies_file_window->file_window,
			FRAME_LEFT_FOOTER, "", NULL);
	
		const Y_FREE_SPACE  		= 130;
		const X_FREE_SPACE  		= 35;
		const INITIAL_WIN_WIDTH 	= 360;
		const INITIAL_LIST_Y	 	= 70;
		const INITIAL_DISPLAY_LENGTH	= 43;
		const LIST_ROW_HEIGHT		= 16;

		int   window_width  = (int) xv_get(win, XV_WIDTH);
		int   window_height = (int) xv_get(win, XV_HEIGHT);

		int   list_width    = (window_width - X_FREE_SPACE);
		
		int   list_rows     = ((window_height - Y_FREE_SPACE)
					/ LIST_ROW_HEIGHT);

		list_rows < 0 ? (list_rows = 0) : list_rows;

		xv_set(goodies_file_window->file_list,
			PANEL_LIST_WIDTH, 	 list_width,
			PANEL_LIST_DISPLAY_ROWS, list_rows,
			NULL);

		int list_height = (int) xv_get(goodies_file_window->file_list, XV_HEIGHT);

		int adjust_factor = ((window_height - list_height -
					Y_FREE_SPACE)/LIST_ROW_HEIGHT);

		adjust_factor < 1 ? (adjust_factor = adjust_factor - 1) :
				     adjust_factor;

		int adjusted_rows = list_rows + adjust_factor;

		adjusted_rows < 0 ? (adjusted_rows = 0) : adjusted_rows;

		xv_set(goodies_file_window->file_list,
			PANEL_LIST_DISPLAY_ROWS, adjusted_rows, NULL);

		int   display_length = (window_width * INITIAL_DISPLAY_LENGTH
					/ INITIAL_WIN_WIDTH);

		xv_set(goodies_file_window->file_textfield,
			PANEL_VALUE_DISPLAY_LENGTH, display_length,
			NULL);

		list_height = (int) xv_get(goodies_file_window->file_list, XV_HEIGHT);

		xv_set(goodies_file_window->file_ok,
			XV_Y, (list_height + INITIAL_LIST_Y), NULL);

		int button_x = (window_width - (2 * X_FREE_SPACE) + 10);

		button_x < 80 ? (button_x = 80) : (button_x);

		xv_set(goodies_file_window->file_cancel,
			XV_X, button_x,
			XV_Y, (list_height + INITIAL_LIST_Y), NULL);

		xv_set(goodies_file_window->file_msg1,
			XV_Y, (list_height + INITIAL_LIST_Y + 40), NULL);
	}

	return notify_next_event_func(win, (Notify_event) event, arg, type);
}

//
// Notify setitimer callback function
//
Notify_value
file_wait_to_open()
{
	if (file_new_selection)
	// there is a newer double click event
	{
		file_new_selection = 0;
	}
	else
		if (file_waiting)
		// no events ocurred since this function was called
		{
			file_waiting = 0;
			file_window_open_file(file_global_string);
		}

	if (goodies_file_window)
	{
		xv_set(goodies_file_window->file_list, 	 XV_SHOW, TRUE, NULL);
		xv_set(goodies_file_window->file_textfield, XV_SHOW, TRUE, NULL);
	}

	return NOTIFY_DONE;
}

//
// Notify callback function for `file_list'.
//
int
file_list_notify(Panel_item item, char *string, Xv_opaque client_data, Panel_list_op op, Event *event)
{
	static  char		previous_string[MAXPATHLEN];
	static	int		previous_millitm = 0;
	static	long		previous_time	 = 0;
	struct	timeb		tp;
	struct	itimerval	old_timer;

	xv_set(goodies_file_window->file_window, FRAME_LEFT_FOOTER, "", NULL);

	if ((op == file_last_op) || ((op != PANEL_LIST_OP_DESELECT) && (op != PANEL_LIST_OP_SELECT)))
		return XV_OK;

	ftime(&tp);

	file_timer.it_value.tv_sec 	  = 0;		// 0 seconds delay
	file_timer.it_value.tv_usec 	  = 900000;	// 900 milliseconds delay
	file_timer.it_interval.tv_sec  = 0;		// no repetition
	file_timer.it_interval.tv_usec = 0;		// no repetition

	if (file_waiting)
	// event arrived while waiting to verify double click
		file_waiting = 0;
	else
	{
		if ((previous_time > 0) && (! strcmp(previous_string, string)))
		{
			long delay_sec = (tp.time - previous_time);

			if (delay_sec == 1)
			{
				int delay_msec = (tp.millitm - previous_millitm);
				if (delay_msec < 0)
				{
					xv_set(goodies_file_window->file_list, XV_SHOW, FALSE, NULL);
					xv_set(goodies_file_window->file_textfield, XV_SHOW, FALSE, NULL);
					file_waiting = 1;
					strcpy(file_global_string, string);
					notify_set_itimer_func(goodies_file_window->file_list,
						file_wait_to_open,ITIMER_REAL,
						&file_timer, &old_timer);

					if (old_timer.it_value.tv_usec == 0)
					// flag to disable previous double click
						file_new_selection = 0;
					else
						file_new_selection = 1;
				}
			}
			else
				if (delay_sec == 0)
				{
					xv_set(goodies_file_window->file_list, XV_SHOW, FALSE, NULL);
					xv_set(goodies_file_window->file_textfield, XV_SHOW, FALSE, NULL);
					file_waiting = 1;
					strcpy(file_global_string, string);
					notify_set_itimer_func(goodies_file_window->file_list,
						file_wait_to_open, ITIMER_REAL,
						&file_timer, &old_timer);

					if (old_timer.it_value.tv_usec == 0)
					// flag to disable previous double click
						file_new_selection = 0;
					else
						file_new_selection = 1;
				}
		}
	}

	file_last_op 	 = op;
	previous_time 	 = tp.time;
	previous_millitm = tp.millitm;
	strcpy(previous_string, string);

	xv_set(goodies_file_window->file_textfield,
		PANEL_READ_ONLY, TRUE,
		PANEL_VALUE, previous_string, NULL);

	xv_set(goodies_file_window->file_panel,
		PANEL_CARET_ITEM, goodies_file_window->file_list, NULL);

	return XV_OK;
}


//
// Event callback function for `file_textfield'.
//
void
file_text_events(Panel_item item, Event *event)
{
	if ((event_id(event) == MS_LEFT)
	||  (event_id(event) == MS_MIDDLE))
	{

		for (int i = 0; i < xv_get(goodies_file_window->file_list,PANEL_LIST_NROWS); i++)
		{
			if (xv_get(goodies_file_window->file_list, PANEL_LIST_SELECTED, i))
				xv_set(goodies_file_window->file_list,PANEL_LIST_SELECT, i, FALSE, NULL);
		}

		file_last_op = 0;
		xv_set(goodies_file_window->file_textfield,
				PANEL_READ_ONLY, FALSE, NULL);
	}

	panel_default_handle_event(item, event);
}

//
// Notify callback function for `file_textfield'.
//
Panel_setting
file_text_notify(Panel_item item, Event *event)
{
	xv_set(goodies_file_window->file_window, FRAME_LEFT_FOOTER, "", NULL);

	char *	value = (char *) xv_get(item, PANEL_VALUE);

	file_timer.it_value.tv_sec 	  = 0;		// 0 seconds delay
	file_timer.it_value.tv_usec 	  = 900000;	// 900 milliseconds delay
	file_timer.it_interval.tv_sec  = 0;		// no repetition
	file_timer.it_interval.tv_usec = 0;		// no repetition

	xv_set(goodies_file_window->file_list, XV_SHOW, FALSE, NULL);
	xv_set(goodies_file_window->file_textfield, XV_SHOW, FALSE, NULL);

	file_waiting   	= 1;
	file_new_selection 	= 0;
	strcpy(file_global_string, value);
	notify_set_itimer_func(goodies_file_window->file_list,file_wait_to_open, ITIMER_REAL, &file_timer, NULL);

	xv_set(goodies_file_window->file_panel, PANEL_CARET_ITEM, item, NULL);

	return panel_text_notify(item, event);
}

//
// Notify callback function for `file_ok'.
//
void
file_ok_notify(Panel_item item, Event *event)
{
	char 	*file;
	int 	nrows 	 = (int) xv_get(goodies_file_window->file_list, PANEL_LIST_NROWS);

	xv_set(goodies_file_window->file_window,    FRAME_LEFT_FOOTER, "", NULL);
	xv_set(goodies_file_window->file_list,      XV_SHOW, FALSE, NULL);
	xv_set(goodies_file_window->file_textfield, XV_SHOW, FALSE, NULL);

	file_timer.it_value.tv_sec 	  = 0;		// 0 seconds delay
	file_timer.it_value.tv_usec 	  = 900000;	// 900 milliseconds delay
	file_timer.it_interval.tv_sec  = 0;		// no repetition
	file_timer.it_interval.tv_usec = 0;		// no repetition

	int 	i 	 = 0;
	int	selected = 0;

	while (i < nrows)
	{
		if (xv_get(goodies_file_window->file_list, PANEL_LIST_SELECTED, i))
		{
			file = (char *) xv_get(goodies_file_window->file_list, PANEL_LIST_STRING, i);
			file_waiting   	= 1;
			file_new_selection 	= 0;
			strcpy(file_global_string, file);
			notify_set_itimer_func(goodies_file_window->file_list,file_wait_to_open, ITIMER_REAL, &file_timer, NULL);
			xv_set(goodies_file_window->file_panel, PANEL_CARET_ITEM, goodies_file_window->file_list, NULL);
			selected  = 1;
			i	  = nrows;
		}

		i++;
	}

	if (! selected)
	{
		file = (char *)	xv_get(goodies_file_window->file_textfield, PANEL_VALUE);
		file_waiting   	= 1;
		file_new_selection 	= 0;
		strcpy(file_global_string, file);
		notify_set_itimer_func(goodies_file_window->file_list,file_wait_to_open, ITIMER_REAL, &file_timer, NULL);
		xv_set(goodies_file_window->file_panel, PANEL_CARET_ITEM, goodies_file_window->file_textfield, NULL);
	}
}

static int PreventDestroy = 0;

//
// Notify callback function for `file_cancel'.
//
void
file_cancel_notify(Panel_item item, Event *event)
{
      if (!PreventDestroy) {
 	xv_destroy_safe(goodies_file_window->file_window);
	delete(goodies_file_window);
	goodies_file_window = NULL;
      }
}

//
// Show and Hide file window (wrj)
//

int goodies_file_window_on_screen() {
  if (goodies_file_window)
    return (int)xv_get(goodies_file_window->file_window, XV_SHOW);
  else
    return 0;
}

void goodies_file_window_show_true() {
  if (goodies_file_window) {
    PreventDestroy = 0;
    (void)xv_set(goodies_file_window->file_window, FRAME_CMD_PUSHPIN_IN, TRUE, XV_SHOW, TRUE, 0);
  }
}

void goodies_file_window_show_false() {
  if (goodies_file_window) {
    PreventDestroy = 1;
    (void)xv_set(goodies_file_window->file_window, FRAME_CMD_PUSHPIN_IN, FALSE, XV_SHOW, FALSE, 0);
  }
}









