/*
    Copyright (C) 1992-1994
		    Juliano L. de Oliveira

    All rights reserved.

    This program is  free software; you can redistribute unmodified
    versions of it as long as you do not  profit from that, neither
    by selling it nor  by   including it  with other products   for
    sale.  This copyright  notice  must  remain in all  distributed
    copies.

    You are  allowed to modify your  personal copy  of this program
    (as long as you do not distribute any changed versions).

    You may,  of course, distribute  along with it, your own source
    code or libraries of algorithms that run under GeoLab.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

    If you must contact the authors, you may write to:
                     geolab-request@dcc.unicamp.br
*/

//
// goodies_file_ui.C
//

#include <stdio.h>
#include <sys/param.h>
#include <xview/xview.h>
#include <xview/panel.h>
#include <xview/notice.h>
#include "goodies_file.h"

// create a global pointer to the file_window object

goodies_file_window_objects *goodies_file_window = NULL;

//
// Create and Initialize the object `file_window'.
// This object will manage the user actions in order to select a file
//
void
goodies_choose_file(Xv_opaque owner, int new_file,char *dir,
	char magic_number[4], char *win_title, char *tex_label,
	char *but1_label, char *but2_label, file_function file_func)
{
	extern void file_window_open_file(char *);

	if (goodies_file_window)
	{
		xv_destroy_safe(goodies_file_window->file_window);
		delete goodies_file_window;
	}

	goodies_file_window = new goodies_file_window_objects;

	goodies_file_window->file_window 	=
		goodies_file_window->file_window_create(owner, win_title);
	goodies_file_window->file_panel 	=
		goodies_file_window->file_panel_create();
	goodies_file_window->file_textfield 	=
		goodies_file_window->file_textfield_create(tex_label);
	goodies_file_window->file_list 		=
		goodies_file_window->file_list_create();
	goodies_file_window->file_ok 		=
		goodies_file_window->file_ok_create(but1_label);
	goodies_file_window->file_cancel 	=
		goodies_file_window->file_cancel_create(but2_label);
	goodies_file_window->file_msg1 	 	=
		goodies_file_window->file_msg1_create();
	goodies_file_window->file_file_func 	= file_func;
	goodies_file_window->file_new_file 	= new_file;

	strcpy(goodies_file_window->file_magic_number, magic_number);
	strcpy(goodies_file_window->file_dir_name,dir);
	strcpy(goodies_file_window->file_fd_file_name,"");
	strcpy(goodies_file_window->file_fd_path_name, dir);
	goodies_file_window->file_fd_file_type = 0;

	file_window_open_file(dir);

	xv_set(goodies_file_window->file_window,XV_SHOW, TRUE, NULL);
}

//
// Create object `file_window' in the specified instance.

//
Xv_opaque
goodies_file_window_objects::file_window_create(Xv_opaque owner, char *title)
{
	extern Notify_value	file_window_events(Xv_window, Event *, Notify_arg, Notify_event_type);

	extern void		file_window_done(Frame);

	Xv_opaque		obj;
	
	obj = xv_create(owner, FRAME_CMD,
		XV_WIDTH, 360,
		XV_HEIGHT, 300,
		XV_LABEL, title,
		FRAME_SHOW_FOOTER, TRUE,
		FRAME_SHOW_RESIZE_CORNER, TRUE,
		FRAME_CMD_PUSHPIN_IN, TRUE,
		FRAME_DONE_PROC,      file_window_done,
		NULL);

	xv_set(obj, WIN_CONSUME_EVENTS, NULL, NULL);

	notify_interpose_event_func(obj,
		(Notify_func) file_window_events, NOTIFY_SAFE);

	xv_set(xv_get(obj, FRAME_CMD_PANEL), WIN_SHOW, FALSE, NULL);

	return obj;
}

//
// Create object `file_panel' in the specified instance.

//
Xv_opaque
goodies_file_window_objects::file_panel_create()
{
	Xv_opaque	obj;
	
	obj = xv_create(goodies_file_window->file_window, PANEL,
		XV_HELP_DATA, "goodies:file_panel",
		XV_X, 0,
		XV_Y, 0,
		XV_WIDTH, WIN_EXTEND_TO_EDGE,
		XV_HEIGHT, WIN_EXTEND_TO_EDGE,
		WIN_BORDER, FALSE,
		NULL);
	return obj;
}

//
// Create object `file_textfield' in the specified instance.

//
Xv_opaque
goodies_file_window_objects::file_textfield_create(char *title)
{
	extern Panel_setting	file_text_notify(Panel_item, Event *);
	extern void		file_text_events(Panel_item, Event *);
	Xv_opaque	obj;
	
	obj = xv_create(goodies_file_window->file_panel, PANEL_TEXT,
		XV_HELP_DATA, "goodies:file_textfield",
		XV_X, 5,
		XV_Y, 12,
		XV_WIDTH, 344,
		XV_HEIGHT, 32,
		PANEL_LABEL_STRING, title,
		PANEL_VALUE_X, 5,
		PANEL_VALUE_Y, 29,
		PANEL_LAYOUT, PANEL_VERTICAL,
		PANEL_VALUE_DISPLAY_LENGTH, 43,
		PANEL_VALUE_STORED_LENGTH, 100,
		PANEL_READ_ONLY, TRUE,
		PANEL_NOTIFY_PROC, file_text_notify,
		PANEL_EVENT_PROC, file_text_events,
		NULL);
	return obj;
}

//
// Create object `file_list' in the specified instance.
//
Xv_opaque
goodies_file_window_objects::file_list_create()
{
	extern int	file_list_notify(Panel_item, char *, Xv_opaque, Panel_list_op, Event *);

	Xv_opaque	obj;
	
	obj = xv_create(goodies_file_window->file_panel, PANEL_LIST,
		XV_HELP_DATA, 	"goodies:file_list",
		XV_X, 		5,
		XV_Y, 		60,
		PANEL_LIST_WIDTH, 	330,
		XV_HEIGHT, 		181,
		PANEL_LAYOUT, 		PANEL_VERTICAL,
		PANEL_LIST_DISPLAY_ROWS, 	8,
		PANEL_LIST_ROW_HEIGHT, 		16,
		PANEL_READ_ONLY, 		TRUE,
		PANEL_CHOOSE_ONE, 		TRUE,
		PANEL_CHOOSE_NONE, 		TRUE,
		PANEL_NOTIFY_PROC, 		file_list_notify,
		NULL);
	return obj;
}

//
// Create object `file_ok' in the specified instance.

//
Xv_opaque
goodies_file_window_objects::file_ok_create(char *title)
{
	extern void		file_ok_notify(Panel_item, Event *);
	Xv_opaque	obj;
	
	obj = xv_create(goodies_file_window->file_panel, PANEL_BUTTON,
		XV_HELP_DATA, "goodies:file_ok",
		XV_X, 10,
		XV_Y, 255,
		PANEL_LABEL_STRING, title,
		PANEL_NOTIFY_PROC, file_ok_notify,
		NULL);
	return obj;
}

//
// Create object `file_cancel' in the specified instance.

//
Xv_opaque
goodies_file_window_objects::file_cancel_create(char *title)
{
	extern void		file_cancel_notify(Panel_item, Event *);
	Xv_opaque	obj;
	
	obj = xv_create(goodies_file_window->file_panel, PANEL_BUTTON,
		XV_HELP_DATA, "goodies:file_cancel",
		XV_X, 210,
		XV_Y, 255,
		PANEL_LABEL_STRING, title,
		PANEL_NOTIFY_PROC, file_cancel_notify,
		NULL);
	return obj;
}

//
// Create object `file_msg1' in the specified instance.
//
Xv_opaque
goodies_file_window_objects::file_msg1_create()
{
	Xv_opaque	obj;
	
	obj = xv_create(goodies_file_window->file_panel, PANEL_MESSAGE,
		XV_HELP_DATA, "goodies:file_msg1",
		XV_X, 10,
		XV_Y, 275,
		PANEL_LABEL_STRING, "1 directory, 0 files",
		NULL);
	return obj;
}


