/*
    Copyright (C) 1992-1994
                    Welson R. Jacometti

    All rights reserved.

    This program is  free software; you can redistribute unmodified
    versions of it as long as you do not  profit from that, neither
    by selling it nor  by   including it  with other products   for
    sale.  This copyright  notice  must  remain in all  distributed
    copies.

    You are  allowed to modify your  personal copy  of this program
    (as long as you do not distribute any changed versions).

    You may,  of course, distribute  along with it, your own source
    code or libraries of algorithms that run under GeoLab.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

    If you must contact the authors, you may write to:
                     geolab-request@dcc.unicamp.br
*/

/*
 *
 * Half_edge.C - implementacao dos metodos e funcoes auxiliares de half_edge.
 *
 *
 *    	Data	Hora  	Nome     		Modificacao/Motivo
 *
 *    25.10.91  17:45 	Welson R. Jacometti  	Arquivo criado
 *
 */

/*******************************************************************************
 *
 * Function prototypes:
 *
 ******************************************************************************/

#include <stdio.h>
#include <assert.h>
#include "half_edge.h"

class duplicate_info
{
  public:
    void *last, *current;
};
typedef duplicate_info *duplicate_info_ptr;


static void set_info(void*& info,void* new_info)
{
    duplicate_info_ptr aux_dupl;
    
    aux_dupl = new duplicate_info;
    aux_dupl->last = info;aux_dupl->current = new_info;
    info = (void*)aux_dupl;
}

static void restore_info(void*& info)
{
    duplicate_info_ptr aux_dupl;
    
    aux_dupl = (duplicate_info_ptr)info;
    info = aux_dupl->last;
    delete aux_dupl;
}


// low-level euler operators: used by interface's euler operators (high-level)

  HalfEdge *addhe(Edge *e, Vertex *v, HalfEdge *where, int sign);
  HalfEdge *delhe(HalfEdge *he);

  void lmev(HalfEdge *he1, HalfEdge *he2, Id vn, float x, float y, float z = 0);
  void lmev(EulerSequence& inv,HalfEdge *he1, HalfEdge *he2, Id vn, float x, float y, float z = 0);
  void lkev(HalfEdge *he1, HalfEdge *he2);

  void lkev(EulerSequence& inv,HalfEdge *he1, HalfEdge *he2);
  Face *lmef(HalfEdge *he1, HalfEdge *he2, Id nf);
  Face *lmef(EulerSequence& inv,HalfEdge *he1, HalfEdge *he2, Id nf);
  void lkef(HalfEdge *h1, HalfEdge *h2);
  void lkef(EulerSequence& inv,HalfEdge *h1, HalfEdge *h2);
  void lmekr(HalfEdge *he1, HalfEdge *he2);
  void lmekr(EulerSequence& inv,HalfEdge *he1, HalfEdge *he2);
  void lkemr(HalfEdge *h1, HalfEdge *h2);
  void lkemr(EulerSequence& inv,HalfEdge *h1, HalfEdge *h2);
  Face *lmfkrh(Loop *l, Id fn);
  Face *lmfkrh(EulerSequence& inv,Loop *l, Id fn);
  void lkfmrh(Face *f1, Face *f2);
  void lkfmrh(EulerSequence& inv,Face *f1, Face *f2);

  void lringmv(Face *f1, Face *f2, Loop *l, int out);
  void lringmv(EulerSequence& inv,Face *f1, Face *f2, Loop *l, int out);

  void laringmv(Face *f1, Face *f2);

  // utilities 
  void merge2(Solid *s1, Solid *s2);
  int match(HalfEdge *h1, HalfEdge *h2);
  int contvv(Vertex *v1, Vertex *v2);
  int contfv(Face *f, Vertex *v);
  int contfv2(Face *f, Vertex *v, int drop);
  int contlv(Loop *l, Vertex *v, int drop);
  int contev(Vertex *v1, Vertex *v2, Vertex *v3);
  int bndrlv(Loop *l, Vertex *v, HalfEdge *&hithe, Vertex *&hitvertex);
  int int2ee(Vertex *v1, Vertex *v2, Vertex *v3, Vertex *v4, int drop, double &t1, double &t2);
  int intrve(Vertex *v1, Vertex *v2, Vertex *v3, double &t);


/*******************************************************************************
 *
 * Error handling functions:
 *
 ******************************************************************************/

static char ErrBuf[100];			// Mensagem de erro


void DefaultWarning( const char *Msg )
{
    cerr << "Warning:" << Msg << endl;
}

void DefaultFatal( const char *Msg, int ExitCode )
{
    cerr << "Fatal error:" << Msg << endl;
    exit( ExitCode );
}


static void (*Warning) (const char *) = DefaultWarning;
static void (*FatalError) (const char *, int) = DefaultFatal;


void SetWarningHandler( void (*Handler) (const char *) )
{
    assert( Handler != NULL );
    Warning = Handler;
}

void SetFatalHandler( void (*Handler) (const char *, int) )
{
    assert( Handler != NULL );
    FatalError = Handler;
}




/*******************************************************************************
 *
 * GWB methods:
 *
 ******************************************************************************/

// mangement methods

bool GWB::killSolid( Id sno )
{
    Solid *s = getSolid( sno );
    if ( s == NULL )
      {
        sprintf( ErrBuf, "ks: solid s%d not found", sno );
	Warning( ErrBuf );
	return false;
      }

    killSolid(s);
    return true;
}

void GWB::killSolid( Solid *s )
{
    Vertex	*v, *nextv;
    Edge	*e, *nexte;
    Loop	*l, *nextl;
    Face	*f, *nextf;

    if ( ! s )
      return;

    v = s -> sverts;
    while (v)
      {
	nextv = v -> nextv;
	delete v;
	v = nextv;
      }

    e = s -> sedges;
    while (e)
      {
	nexte = e -> nexte;
	delete e -> he1;
	delete e -> he2;
	delete e;
	e = nexte;
      }

    f = s -> sfaces;
    while (f)
      {
	nextf = f -> nextf;

	l = f -> floops;
	while (l)
	  {
	    nextl = l -> nextl;
	    delete l;
	    l = nextl;
	  }

	delete f;
	f = nextf;
      }
}

void GWB::killAll()
{
//    while( firsts )
//      killSolid( firsts );

    if ( firsts )
      while( firsts -> nexts )
        {
	  cerr << "killAll:: Deleting solid: " << firsts -> nexts -> solidno
		<< endl;
          killSolid( firsts -> nexts );
        }
    if ( firsts )
        {
	  cerr << "killAll:: Deleting solid: " << firsts -> solidno
		<< endl;
          killSolid( firsts );
        }
    if ( firsts )
      cerr << "killAll:: Firsts is not NULL at end!!!" << endl;
}

// query methods

Id GWB::getNextSolidId() {
    Solid *s = firsts;
    Id maxId = 0;

    while ( s )
      {
        if ( s -> solidno > maxId )
          maxId = s -> solidno;
        s = s -> nexts;
      }
    return maxId+1;
}

Solid *GWB::getSolid ( Id sno ) const {
    Solid *s = firsts;
    while ( s && s -> solidno != sno )
      s = s -> nexts;
    return s;
}


HalfEdge *GWB::getHalfEdge( Id sno, Id fno, Id v1no, Id v2no ) {
    Vertex *v = getVertex( sno, v1no );
    HalfEdge *he;

    if ( v && v -> vedge )
      {
	he = v-> vedge;
	do
	  {
	    if ( he -> nxt -> vtx -> vno == v2no )
	      {
		if ( he -> wloop -> lface -> faceno == fno )
		  return he;
	        he = mate(he);
		if ( he -> wloop -> lface -> faceno == fno )
		  return he;
		he = he -> nxt;
	      }
	    else
	      he = mate(he) -> nxt;
	  }
	while ( he != v -> vedge );
      }
    return NULL;
}


HalfEdge *GWB::getHalfEdge( Id sno, Id fno, Id vno ) {


    Solid *s;
    Face *f;

    if ( s = getSolid( sno ) )
      if ( f = s -> findFace( fno ) )
        return f -> findhe1( vno );
    return NULL;
}


Vertex *GWB::getVertex( Id sno, Id vno ) {
    Solid *s = getSolid( sno );

    if ( s )
      return s -> findVertex( vno );
    return NULL;
/*
    Vertex *v = NULL;
    if ( s )
      {
	v = s -> sverts;
	while ( v && v -> vno != vno )
	  v = v -> nextv;
      }
    return v;
*/
}


// Hi-level Euler's operators

bool GWB::mvfs(Solid *&newsolid, Id sno, Id fno, Id vno,
		  float x, float y, float z) {
    Face     *newface;
    Vertex   *newvertex;
    HalfEdge *newhe;
    Loop     *newloop;

    if ( getSolid( sno ) != NULL )
      {
        sprintf( ErrBuf, "mvfs: solid s%d already exists", sno );
	Warning( ErrBuf );
	return false;
      }

    newsolid = new Solid(this);
    assert(newsolid);
    newface = new Face(newsolid);
    assert(newface);
    newloop = new Loop(newface);
    assert(newloop);
    newvertex = new Vertex(newsolid);
    assert(newvertex);
    newhe = new HalfEdge();
    assert(newhe);

    newsolid->solidno = sno;

    newface->faceno = fno;
    newface->flout = newloop;
    
    newloop->ledg = newhe;
    
    newhe->wloop = newloop;
    newhe->nxt = newhe->prv = newhe;
    newhe->vtx = newvertex;
    newhe->edg = 0;
    
    newvertex->vno = vno;
    newvertex->vcoord[0] = x;
    newvertex->vcoord[1] = y;
    newvertex->vcoord[2] = z;
    newvertex->vcoord[3] = 1.0;

    return true;
}


bool GWB::kvfs( Id sno ) {
    Solid *s = getSolid( sno );
    if ( s == NULL )
      {
        sprintf( ErrBuf, "kvfs: solid s%d not found", sno );
	Warning( ErrBuf );
	return false;
      }

    if ( s -> sedges )
      {
        sprintf( ErrBuf, "kvfs: solid s%d has edges", sno );
	Warning( ErrBuf );
	return false;
      }

    if ( s -> sverts -> nextv ) // solid has more than one vertex?
      {
        sprintf( ErrBuf, "kvfs: solid s%d is not a single vertex", sno );
	Warning( ErrBuf );
	return false;
      }

    delete s -> sverts;
    delete s -> sfaces -> flout -> ledg;
    delete s -> sfaces -> flout;
    delete s -> sfaces;
    delete s;

    return true;
}

bool GWB::mev( Id sno, Id f1, Id f2, Id v1, Id v2, Id v3, Id v4,
               float x, float y, float z ) {

    Solid *s = getSolid( sno );
    if ( s == NULL )
      {
        sprintf( ErrBuf, "mev: solid s%d not found", sno );
	Warning( ErrBuf );
	return false;
      }

    if ( s -> findVertex( v4 ) )
      {
        sprintf( ErrBuf, "[s%d]mev: vertex v%d already exists", sno, v4 );
	Warning( ErrBuf );
	return false;
      }

    return s -> mev( f1, f2, v1, v2, v3, v4, x, y, z );
}

bool GWB::smev( Id sno, Id f1, Id v1, Id v4, float x, float y, float z )
{
    Solid *s = getSolid( sno );
    if ( s == NULL )
      {
        sprintf( ErrBuf, "smev: solid s%d not found", sno );
	Warning( ErrBuf );
	return false;
      }

    if ( s -> findVertex( v4 ) )
      {
        sprintf( ErrBuf, "[s%d]smev: vertex v%d already exists", sno, v4 );
	Warning( ErrBuf );
	return false;
      }

    return s -> smev( f1, v1, v4, x, y, z );
}


bool GWB::kev( Id sno, Id f, Id v1, Id v2 )
{
    Solid *s = getSolid( sno );
    if ( s == NULL )
      {
        sprintf( ErrBuf, "kev: solid s%d not found", sno );
	Warning( ErrBuf );
	return false;
      }
    return s -> kev( f, v1, v2 );
}

bool GWB::mef( Id sno, Id f1, Id v1, Id v2, Id v3, Id v4, Id f2 ) {

    Solid *s = getSolid( sno );
    if ( s == NULL )
      {
        sprintf( ErrBuf, "mef: solid s%d not found", sno );
	Warning( ErrBuf );
	return false;
      }

    if ( s -> findFace( f2 ) )
      {
        sprintf( ErrBuf, "[s%d]mef: face f%d already exists", sno, f2 );
	Warning( ErrBuf );
	return false;
      }

    return s -> mef( f1, v1, v2, v3, v4, f2 );
}

bool GWB::smef( Id sno, Id f1, Id v1, Id v3, Id f2 ) {
    Solid *s = getSolid( sno );
    if ( s == NULL )
      {
        sprintf( ErrBuf, "smef: solid s%d not found", sno );
	Warning( ErrBuf );
	return false;
      }

    if ( s -> findFace( f2 ) )
      {
        sprintf( ErrBuf, "[s%d]smef: face f%d already exists", sno, f2 );
	Warning( ErrBuf );
	return false;
      }

    return s -> smef( v1, v3, f1, f2 );
}

bool GWB::kef( Id sno, Id f, Id v1, Id v2 ) {
    Solid *s = getSolid( sno );
    if ( s == NULL )
      {
        sprintf( ErrBuf, "kef: solid s%d not found", sno );
	Warning( ErrBuf );
	return false;
      }
    return s -> kef( v1, v2, f );
}

bool GWB::mekr( Id sno, Id f, Id v1, Id v2, Id v3, Id v4 ) {

    Solid *s = getSolid( sno );
    if ( s == NULL )
      {
        sprintf( ErrBuf, "mekr: solid s%d not found", sno );
	Warning( ErrBuf );
	return false;
      }
    return s -> mekr( f, v1, v2, v3, v4 );
}

bool GWB::smekr( Id sno, Id f, Id v1, Id v3 ) {

    Solid *s = getSolid( sno );
    if ( s == NULL )
      {
        sprintf( ErrBuf, "smekr: solid s%d not found", sno );
	Warning( ErrBuf );
	return false;
      }
    return s -> smekr( f, v1, v3 );
}

bool GWB::kemr( Id sno, Id f, Id v1, Id v2 ) {

    Solid *s = getSolid( sno );
    if ( s == NULL )
      {
        sprintf( ErrBuf, "kemr: solid s%d not found", sno );
	Warning( ErrBuf );
	return false;
      }
    return s -> kemr( f, v1, v2 );
}



// Load & save ...


void EulerSequence::kill()
{
    list_item n1,n2;
    if (!it.empty())
      {
	  n1 = it.first();
	  n2 = it.succ(n1);
	  
	  while (n2!=0)
	    {
		delete it[n1];
		it.del_item(n1);
		n1 = n2;
		n2 = it.succ(n1);
	    };
	  delete it[n1];
	  it.del_item(n1);
      }
}

void EulerSequence::push(Id f1,Id f2, Id v1, Id v2, Id v3, Id v4, double x, double y, double z,EulerOperator name)
{
    EulerOperatorRepresentation_ptr aux_rep = new EulerOperatorRepresentation;
    assert(aux_rep);
    aux_rep->name = name;
    aux_rep->f1   = f1;
    aux_rep->f2   = f2;
    aux_rep->v1   = v1;
    aux_rep->v2   = v2;
    aux_rep->v3   = v3;
    aux_rep->v4   = v4;
    aux_rep->x    = x;
    aux_rep->y    = y;
    aux_rep->z    = z;
    it.push(aux_rep);

}


void EulerSequence::append(Id f1,Id f2, Id v1, Id v2, Id v3, Id v4, double x, double y, double z,EulerOperator name)
{
    EulerOperatorRepresentation_ptr aux_rep = new EulerOperatorRepresentation;
    assert(aux_rep);
    aux_rep->name = name;
    aux_rep->f1   = f1;
    aux_rep->f2   = f2;
    aux_rep->v1   = v1;
    aux_rep->v2   = v2;
    aux_rep->v3   = v3;
    aux_rep->v4   = v4;
    aux_rep->x    = x;
    aux_rep->y    = y;
    aux_rep->z    = z;
    it.append(aux_rep);
}


void EulerSequence::pop(Id& f1,Id& f2, Id& v1, Id& v2, Id& v3, Id &v4, double& x, double& y, double &z,EulerOperator& name)
{
    EulerOperatorRepresentation_ptr aux_rep;
    aux_rep = it.pop();
    name = aux_rep->name;
    f1   = aux_rep->f1;
    f2   = aux_rep->f2;
    v1   = aux_rep->v1;
    v2   = aux_rep->v2;
    v3   = aux_rep->v3;
    v4   = aux_rep->v4;
    x = aux_rep->x;
    y = aux_rep->y;
    z = aux_rep->z;

    delete aux_rep;

}



void EulerSequence::show()
{
    EulerOperatorRepresentation_ptr el;
    Id f1,f2,v1,v2,v3,v4;
    double x,y,z;
    forall(el,it)
      {
	  f1 = el->f1; f2 = el->f2;
	  v1 = el->v1; v2 = el->v2; v3 = el->v3; v4 = el->v4; 
	  x = el->x; y = el->y; z = el->z;
	  switch(el->name)
	    {
	      case MVFS   :
		{
		    cerr<<" MVFS : "<<endl;
		    cerr<<"   f1 = "<<f1<<endl;
		    cerr<<"   v1 = "<<v1<<endl;
		    cerr<<"   x  = "<<x<<"|y = "<<y<<"|z = "<<z<<endl;
		};break;
	      case KVFS   :
		{
		    cerr<<" KVFS : "<<endl;
		};break;
	      case MEV    :
		{
		    cerr<<" MEV : "<<endl;
		    cerr<<"   f1 = "<<f1<<"|f2 = "<<f2<<endl;
		    cerr<<"   v1 = "<<v1<<"|v2 = "<<v2<<"|v3 = "<<v3<<"|v4 = "<<v4<<endl;
		    cerr<<"   x  = "<<x<<"|y = "<<y<<"|z = "<<z<<endl;
		};break;
	      case SMEV   :
		{
		    cerr<<" SMEV : "<<endl;
		    cerr<<"   f1 = "<<f1<<endl;
		    cerr<<"   v1 = "<<v1<<"|v2 = "<<v2<<endl;
		    cerr<<"   x  = "<<x<<"|y = "<<y<<"|z = "<<z<<endl;
		};break;		
	      case KEV    :
		{
		    cerr<<" KEV : "<<endl;
		    cerr<<"   f1 = "<<f1<<endl;
		    cerr<<"   v1 = "<<v1<<"|v2 = "<<v2<<endl;
		};break;
	      case MEF    :
		{
		    cerr<<" MEF : "<<endl;
		    cerr<<"   f1 = "<<f1<<"|f2 = "<<f2<<endl;
		    cerr<<"   v1 = "<<v1<<"|v2 = "<<v2<<"|v3 = "<<v3<<"|v4 = "<<v4<<endl;
		};break;
	      case SMEF   :
		{
		    cerr<<" SMEF : "<<endl;
		    cerr<<"   f1 = "<<f1<<"|f2 = "<<f2<<endl;
		    cerr<<"   v1 = "<<v1<<"|v2 = "<<v2<<endl;
		};break;		    
	      case KEF    :
		{
		    cerr<<" KEF : "<<endl;
		    cerr<<"   f1 = "<<f1<<endl;
		    cerr<<"   v1 = "<<v1<<"|v2 = "<<v2<<endl;
		};break;				
	      case KEMR   :
		{
		    cerr<<" KEMR : "<<endl;
		    cerr<<"   f1 = "<<f1<<endl;
		    cerr<<"   v1 = "<<v1<<"|v2 = "<<v2<<endl;
		};break;
	      case MEKR   :
		{
		    cerr<<" MEKR : "<<endl;
		    cerr<<"   f1 = "<<f1<<endl;
		    cerr<<"   v1 = "<<v1<<"|v2 = "<<v2<<"|v3 = "<<v3<<"|v4 = "<<v4<<endl;
		};break;
	      case SMEKR  :
		{
		    cerr<<" SMEKR : "<<endl;
		    cerr<<"   f1 = "<<f1<<endl;
		    cerr<<"   v1 = "<<v1<<"|v2 = "<<v2<<endl;
		};break;		
	      case KFMRH  :
		{
		    cerr<<" KFMRH : "<<endl;
		    cerr<<"   f1 = "<<f1<<"|f2 = "<<f2<<endl;
		};break;
	      case MFKRH  :
		{
		    cerr<<" MFKRH : "<<endl;
		    cerr<<"   f1 = "<<f1<<"|f2 = "<<f2<<endl;
		    cerr<<"   v1 = "<<v1<<"|v2 = "<<v2<<endl;
		};break;
	      case RINGMV :
		{
		    cerr<<" RINGMV : "<<endl;
		    cerr<<"   f1 = "<<f1<<"|f2 = "<<f2<<endl;
		    cerr<<"   v1 = "<<v1<<"|v2 = "<<v2<<endl;
		    cerr<<"   lout = "<<v3<<endl;
		};break;
	      default:
		{
		    cerr<<" Not called in half_edge.C - showEulerSequence"<<endl;
		    exit(1);
		};
	    };
      };
}




  
EulerSequence_ptr GWB::invert(Id solid_id)
{
    Solid * solid                 = getSolid(solid_id);
    EulerSequence_ptr aux_list    = new EulerSequence;
    register Edge     * edge      = solid->sedges;
    register Vertex   *pivot , *vertex;
    Face              *face;
    register HalfEdge *he1 , *he2; 
    Id                pid , vid;
    
    
    while (edge!=0)
      {
	  he1 = edge->he1;
	  he2 = edge->he2;
	  
	  if (he1->wloop->lface == he2->wloop->lface)
	    if (he1->wloop == he2->wloop)
	      if (he1->prv == he2)
		lkev(*aux_list,he1,he2);
	      else
		if (he2->prv == he1)
		  lkev(*aux_list,he2,he1);
		else
		  lkemr(*aux_list,he1,he2);
	    else
	      {
		  lmfkrh(*aux_list,he1->wloop,he1->wloop->lface->faceno);
		  lkef(*aux_list,he1,he2);
		  cerr<<" Tested for the First Time ! "<<endl;
	      }
	  else
	    lkef(*aux_list,he1,he2);
	  edge = edge->nexte;
      };

    
    pivot  = solid->sverts;
    vertex = pivot->nextv;
    face   = solid->sfaces;
    pid    = pivot->vno;
    
    while (vertex)
      {
	  vid = vertex->vno;
	  lmekr(*aux_list,face->findhe2(pid,pid),face->findhe2(vid,vid));
	  lkev(*aux_list,face->findhe2(vid,pid),face->findhe2(pid,vid));
	  vertex = vertex->nextv;
      };
    
    aux_list->push(face->faceno,0,pid,0,0,0,pivot->X(),pivot->Y(),pivot->Z(),MVFS);
    
    kvfs(solid->solidno);
    
    return aux_list;
}

void GWB::build(Id solid_id,const EulerSequence& sequence)
{
    register EulerOperatorRepresentation_ptr it;
    Id f1,f2,v1,v2,v3,v4;
    double x,y,z;
    forall(it,sequence.it)
      {
	  f1 = it->f1; f2 = it->f2;
	  v1 = it->v1; v2 = it->v2; v3 = it->v3; v4 = it->v4; 
	  x = it->x; y = it->y; z = it->z;
	  switch(it->name)
	    {
	      case MVFS   : mvfs(solid_id,f1,v1,x,y,z);break;
	      case KVFS   : kvfs(solid_id);break;
	      case MEV    : mev(solid_id,f1,f2,v1,v2,v3,v4,x,y,z);break;
	      case SMEV   : smev(solid_id,f1,v1,v2,x,y,z);break;
	      case KEV    : kev(solid_id,f1,v1,v2);break;
	      case MEF    : mef(solid_id,f1,v1,v2,v3,v4,f2);break;
	      case SMEF   : smef(solid_id,f1,v1,v2,f2);break;
	      case KEF    : kef(solid_id,f1,v1,v2);break;
	      case KEMR   : kemr(solid_id,f1,v1,v2);break;
	      case MEKR   : mekr(solid_id,f1,v1,v2,v3,v4);break;
	      case SMEKR  : smekr(solid_id,f1,v1,v2);break;
	      case KFMRH  : /*kfmrh(solid_id,f1,f2);*/break;
	      case MFKRH  : /*mfkrh(solid_id,f1,v1,v2,f2);*/break;
	      case RINGMV : getFirsts()->ringmv(v1,v2,f1,f2,v3);break;
	      default:
		{
		    cerr<<" Unknown Euler Operator in half_edge.c - GWB::build"<<endl;
		    exit(1);
		};
	    };
      };
}



// geometric support

Solid *GWB::circle(float cx, float cy, float rad, float h, int n) {
    Solid *s;

    mvfs(s, getNextSolidId(), 1, 1, cx+rad, cy, h);
    s->arc(1, 1, cx, cy, rad, h, 0.0, ((n-1)*360.0/n), n-1);
    s->smef(n, 1, 1, 2);
    return s;
}


Solid *GWB::torus(float r1, float r2, int nf1, int nf2) {
    Solid *s;

    s = circle(0.0, r1, r2, 0.0, nf2);
    s->rsweep(nf1);
    return s;
}


void GWB::duplicate(GWB& gwb)
{
    Solid   *s1=0, *s2=0;

    s1 = firsts;

    while (s1 != 0)
      {
	  s2          = s1->duplicate();
	  s2->solidno = s1->solidno;
	  s2->sgwb    = &gwb;

	  s2->nexts   = gwb.firsts;
	  s2->prevs   = 0;
	  if (s2->nexts) s2->nexts->prevs = s2; 
	  gwb.firsts = s2;

	  s1 = s1->nexts;
      }

}

// Solid methods

Solid::Solid(GWB *gwb) {
    sgwb = gwb;
    sfaces = NULL;
    sedges = NULL;
    sverts = NULL;
    nexts = sgwb -> firsts;
    prevs = NULL;
    if (nexts)
      nexts -> prevs = this;
    sgwb -> firsts = this;
}

Solid::~Solid() {
    if (prevs)
      prevs->nexts = nexts;
    if (nexts)
      nexts->prevs = prevs;
    if (this == sgwb -> firsts)
      if (prevs)
        sgwb -> firsts = prevs;
      else
        sgwb -> firsts = nexts;
}

void Solid::addface(Face *face) {
    face->nextf = sfaces;
    face->prevf = 0;
    if (sfaces)
      sfaces->prevf = face;
    sfaces = face;
    face->fsolid = this;
}

void Solid::delface(Face *face) {
    if (face->prevf)
      face->prevf->nextf = face->nextf;
    if (face->nextf)
      face->nextf->prevf = face->prevf;
    if (face == sfaces)
      sfaces = face->nextf;
}

void Solid::addedge(Edge *edge) {
    edge->nexte = sedges;
    edge->preve = 0;
    if (sedges)
      sedges->preve = edge;
    sedges = edge;
}

void Solid::deledge(Edge *edge) {
    if (edge->preve)
      edge->preve->nexte = edge->nexte;
    if (edge->nexte)
      edge->nexte->preve = edge->preve;
    if (edge == sedges)
      sedges = edge->nexte;
}        

void Solid::addvertex(Vertex *vertex) {
    vertex->nextv = sverts;
    vertex->prevv = 0;
    if (sverts)
      sverts->prevv = vertex;
    sverts = vertex;
}       

void Solid::delvertex(Vertex *vertex) {
    if (vertex->prevv)
      vertex->prevv->nextv = vertex->nextv;
    if (vertex->nextv)
      vertex->nextv->prevv = vertex->prevv;
    if (vertex == sverts)
      sverts = vertex->nextv;  
}

Face *Solid::findFace(Id fno) {
    Face *f = sfaces;

    while(f) {
      if (f->faceno == fno)
        return f;
      f = f->nextf;
    }
    return 0;
}

Vertex *Solid::findVertex(Id vno) {
    Vertex *v = sverts;

    while ( v && v -> vno != vno )
      v = v -> nextv;
    return v;
}


#ifdef DEBUG
void Solid::checkv() {
    Vertex *v = sverts;
    HalfEdge *he;
    int maxtry;
    
    while (v) {
      he = v->vedge;
      maxtry = 50;
      if (he)
        do {
          assert(he->edg);
          assert(he->vtx == v);
          assert(maxtry-- > 0);
        } while ((he = mate(he)->nxt) != v->vedge);
      v = v->nextv;
    }
}
#endif DEBUG

void Solid::getmaxnames(Id &maxf, Id &maxv) {
    Vertex *v;
    Face *f;
    
    for (v = sverts, maxv = 0; v; v = v->nextv)
      if (v->vno > maxv)
        maxv = v->vno;
      
    for (f = sfaces, maxf = 0; f; f = f->nextf)
      if (f->faceno > maxf)
        maxf = f->faceno;
}

void Solid::arc(Id f, Id v, float cx, float cy, float rad, float h, float phi1, float phi2, int n) {
    float x, y, angle, inc;
    Id prev, maxv, dummy;
    int i;

    angle = phi1 * PI / 180.0;
    inc = (phi2 - phi1) * PI / (180.0 * n);
    prev = v;  
    getmaxnames(dummy, maxv);
    for(i = 0; i < n; i++) {
      angle += inc;  
      x = cx + cos(angle) * rad;
      y = cy + sin(angle) * rad;
      smev(f, prev, ++maxv, x, y, h);
      prev = maxv;
    }   
}

void Solid::solidls(int sw) {
    Face *f;
    Loop *l;
    Vertex *v;
    Edge *e;
    HalfEdge *he;

    cout << "Solid " << solidno << ":" << endl
         << "Faces:" << endl;
    f = sfaces;
    while (f) {
      f->facels(sw);
      f = f->nextf;
    }

    if (sw > 1) {
      cout << "Vertices:";
      v = sverts;
      while(v) {
      	cout << endl << v->vno << ": " << v->vcoord;
        if (sw > 2) {
          he = v->vedge;
          if (he) {
            do {
              e = he->edg;
              cout << ( (he == e->he1) ? " +" : " -" ) << "<"
		   << e->he1->vtx->vno << " "
                   << e->he2->vtx->vno << ">";
            } while ((he = mate(he)->nxt) != v->vedge);
          }
        }
        v = v->nextv;
      }
      cout << endl
           << "Edges: <v1 v2 f1 f2>";
      f = sfaces;
      while (f) {
        l = f->floops;
        while (l) {
          he = l->ledg;
          do {
            if (e = he->edg) {
              if (he == e->he1) {
              	cout << endl;
                cout << "<" << e->he1->vtx->vno << " "
                     << e->he2->vtx->vno << " " 
                     << e->he1->wloop->lface->faceno << " "
                     << e->he2->wloop->lface->faceno << ">";
              }
            }
          } while ((he = he->nxt) != l->ledg);
          l = l->nextl;
        }
        f = f->nextf;
      }
    }
    cout << endl;
}

void Solid::listsolid() {
    Face *f;
    Loop *l;
    HalfEdge *he;

    f = sfaces;
    while (f) {
      cout << "Face: " << f->faceno << endl;
      l = f->floops;
      while (l) {
        cout << "loop: ";
        he = l->ledg;
        do {
          cout << "  " << he->vtx->vno << ":(" <<  he->vtx->vcoord[0] << " "
               << he->vtx->vcoord[1] << " " << he->vtx->vcoord[2] << ")";
        } while ((he = he->nxt) != l->ledg);
        cout << endl;
        l = l->nextl;
      }
      f = f->nextf;
    }
}

bool Solid::mev( Id f1, Id f2, Id v1, Id v2, Id v3, Id v4,
		    float x, float y, float z )
{
    Face     *face1, *face2;
    HalfEdge *he12, *he13;
    
    if (!(face1 = findFace(f1))) {
      sprintf( ErrBuf, "[s%d]mev: Face f%d not found", solidno, f1 );
      Warning( ErrBuf );
      return false;
    }

    if (!(face2 = findFace(f2))) {
      sprintf( ErrBuf, "[s%d]mev: Face f%d not found", solidno, f2 );
      Warning( ErrBuf );
      return false;
    }

    if (!(he12 = face1->findhe2(v1,v2))) {
      sprintf( ErrBuf, "[s%d,f%d]mev: Edge e%d,%d not found",
		       solidno, f1, v1, v2 );
      Warning( ErrBuf );
      return false;
    }

    if (!(he13 = face2->findhe2(v1,v3))) {
      sprintf( ErrBuf, "[s%d,f%d]mev: Edge e%d,%d not found",
		       solidno, f2, v1, v3 );
      Warning( ErrBuf );
      return false;
    }

    lmev(he12, he13, v4, x, y, z);

    return true;
}

bool Solid::smev(Id fn, Id vn1, Id vn2, float x, float y, float z) {
    Face     *oldface;
    HalfEdge *he;
    
    if (!(oldface = findFace(fn))) {
      sprintf( ErrBuf, "[s%d]smev: Face f%d not found", solidno, fn );
      Warning( ErrBuf );
      return false;
    }

    if (!(he = oldface->findhe1(vn1))) {
      sprintf( ErrBuf, "[s%d,f%d]smev: Vertex v%d not found",
		       solidno, fn, vn1 );
      Warning( ErrBuf );
      return false;
    }

    lmev(he, he, vn2, x, y, z);

    return true;
}

bool Solid::kev(Id fnr, Id v1, Id v2) {
    Face *f;
    register HalfEdge *he1, *he2;

    /* find needed objects, check existence */
    if (!(f = findFace(fnr))) {
      sprintf( ErrBuf, "[s%d]kev: Face f%d not found", solidno, fnr );
      Warning( ErrBuf );
      return false;
    }
    if (!(he1 = f->findhe2(v1, v2))) {
      sprintf( ErrBuf, "[s%d]kev: Edge e%d,%d not found", solidno, v1, v2 );
      Warning( ErrBuf );
      return false;
    }

    he2 = mate(he1);

    /* check that edge occurs in only one loop */
    if (he1->wloop != he2->wloop) {
      sprintf( ErrBuf, "[s%d]kev: Edge e%d,%d occurs in more than one loop",
			solidno, v1, v2 );
      Warning( ErrBuf );
      return false;
    }

    /* check that v2 is an endpoint */
    if (he1->nxt != he2) {
      sprintf( ErrBuf, "[s%d]kev: Vertex v%d is not an endpoint", solidno, v2 );
      Warning( ErrBuf );
      return false;
    }


    /* call lkev() to do the real work */
    lkev(he2, he1);

    return true;
}


bool Solid::mef( Id f1, Id v1, Id v2, Id v3, Id v4, Id f2 )
{
    Face     *f;
    HalfEdge *he12, *he34;
    
    if (!(f = findFace(f1))) {
      sprintf( ErrBuf, "[s%d]mef: Face f%d not found", solidno, f1 );
      Warning( ErrBuf );
      return false;
    }

    if (!(he12 = f->findhe2(v1,v2))) {
      sprintf( ErrBuf, "[s%d,f%d]mef: Edge e%d,%d not found",
		       solidno, f1, v1, v2 );
      Warning( ErrBuf );
      return false;
    }

    if (!(he34 = f->findhe2(v3,v4))) {
      sprintf( ErrBuf, "[s%d,f%d]mef: Edge e%d,%d not found",
		       solidno, f2, v3, v4 );
      Warning( ErrBuf );
      return false;
    }

    (void) lmef(he12, he34, f2);
    return true;
}

bool Solid::smef(Id vert1, Id vert2, Id fac1, Id fac2) {
    Face *oldface;
    register HalfEdge *he1, *he2;

    if (!(oldface = findFace(fac1))) {
      sprintf( ErrBuf, "[s%d]smef: Face f%d not found", solidno, fac1 );
      Warning( ErrBuf );
      return false;
    }

    /* get vert1 */
    if (!(he1 = oldface->findhe1(vert1))) {
      sprintf( ErrBuf, "[s%d,f%d]smef: Vertex v%d not found",
			solidno, fac1, vert1 );
      Warning( ErrBuf );
      return false;
    }

    /* get the next occurance of vert2 */
    he2 = he1;
    if (he2->vtx->vno != vert2)
      do {
        he2 = he2->nxt;
        if (he2 == he1) {
          sprintf( ErrBuf, "[s%d,f%d]smef: Vertex v%d not found",
			   solidno, fac1, vert2 );
          Warning( ErrBuf );
          return false;
        }
      } while (he2->vtx->vno != vert2);

    (void) lmef(he1, he2, fac2);

    return true;
}

bool Solid::kef(Id v1, Id v2, Id fno) {
    Face *f;
    HalfEdge *he;

    /* find the objects, check that they exist */
    if (!(f = findFace(fno))) {
      sprintf( ErrBuf, "[s%d]kef: Face f%d not found", solidno, fno );
      Warning( ErrBuf );
      return false;
    }

    if (!(he = f->findhe2(v1, v2)))
      if (!(he = f->findhe2(v2, v1))) {
        sprintf( ErrBuf, "[s%d,f%d]kef: Edge e%d,%d not found",
			 solidno, fno, v1, v2 );
        Warning( ErrBuf );
        return false;
      }

    /* call lkef() to do the real job */
    lkef(he, mate(he));

    return true;
}

bool Solid::kemr( Id fno, Id v1, Id v2 )
{
    Face *f;
    HalfEdge *he;

    /* find the objects, check that they exist */
    if (!(f = findFace(fno))) {
      sprintf( ErrBuf, "[s%d]kemr: Face f%d not found", solidno, fno );
      Warning( ErrBuf );
      return false;
    }

    if (!(he = f->findhe2(v1, v2)))
//      if (!(he = f->findhe2(v2, v1)))
      {
        sprintf( ErrBuf, "[s%d,f%d]kemr: Edge e%d,%d not found",
			 solidno, fno, v1, v2 );
        Warning( ErrBuf );
        return false;
      }
    
    
    /* call lkemr() to do the real job */
    lkemr(he, mate(he));

    return true;
}

bool Solid::mekr( Id fno, Id v1, Id v2, Id v3, Id v4 )
{
    Face     *f;
    HalfEdge *he12, *he34;
    
    if (!(f = findFace(fno))) {
      sprintf( ErrBuf, "[s%d]mekr: Face f%d not found", solidno, fno );
      Warning( ErrBuf );
      return false;
    }

    if (!(he12 = f->findhe2(v1,v2))) {
      sprintf( ErrBuf, "[s%d,f%d]mekr: Edge e%d,%d not found",
		       solidno, fno, v1, v2 );
      Warning( ErrBuf );
      return false;
    }

    if (!(he34 = f->findhe2(v3, v4))) {
      sprintf( ErrBuf, "[s%d,f%d]mekr: Edge e%d,%d not found",
		       solidno, fno, v3, v4 );
      Warning( ErrBuf );
      return false;
    }

    lmekr(he12, he34);
    return true;
}

bool Solid::smekr( Id fno, Id v1, Id v3 )
{
    Face     *f;
    HalfEdge *he12, *he34;
    
    if (!(f = findFace(fno))) {
      sprintf( ErrBuf, "[s%d]smekr: Face f%d not found", solidno, fno );
      Warning( ErrBuf );
      return false;
    }

    if (!(he12 = f->findhe1(v1))) {
      sprintf( ErrBuf, "[s%d,f%d]smekr: Vertex v%d not found",
		       solidno, fno, v1 );
      Warning( ErrBuf );
      return false;
    }

    if (!(he34 = f->findhe1(v3))) {
      sprintf( ErrBuf, "[s%d,f%d]smekr: Vertex v%d not found",
		       solidno, fno, v3 );
      Warning( ErrBuf );
      return false;
    }

    lmekr(he12, he34);
    return false;
}


void Solid::revert() {
    Face *f;
    Loop *l;
    register HalfEdge *he, *henxt;
    Vertex *prev, *sav;

    f = sfaces;
    while (f) {
      l = f->floops;
      while (l) {
        he = l->ledg;
        do {
          /* revert the direction */
          henxt = he->nxt;
          he->nxt = he->prv;
          he->prv = henxt;
          he = henxt;
        } while (he != l->ledg);
        prev = he->prv->vtx;
        do {
          sav = he->vtx;
          he->vtx = prev;
          he->vtx->vedge = he;
          prev = sav;
          he = he->nxt;
        } while(he != l->ledg);
        l = l->nextl;
      }
      f = f->nextf;
    }
}
   
void Solid::rsweep(int nfaces) {
    HalfEdge *first, *cfirst, *last, *scan;
    TransfMatrix m;
    SpaceCoords v;
   
    HalfEdge *h;
    Face *headf, *tailf;
    int closed_figure;
    Id maxv, maxf;

    getmaxnames(maxf, maxv);

    if (sfaces->nextf) {		/* does the solid have > 1 faces ? */
#ifdef DEBUG
      cout << "closed figure ..." << endl;
#endif
      solidls(2);

      /* assume it's a lamina */
      closed_figure = 1;
      h = sfaces->floops->ledg;
      lmev(h, mate(h)->nxt, ++maxv, h->vtx->vcoord[0], h->vtx->vcoord[1], h->vtx->vcoord[2]); 
      lkef(h->prv, mate(h->prv));
      headf = sfaces;
      
#ifdef DEBUG
      cout << "... opened" << endl;
#endif
      solidls(1);
    }
    else    
      closed_figure = 0;		/* no, it's a wire */


    first = sfaces->floops->ledg;
    while (first->edg != first->nxt->edg) 
      first = first->nxt;
      
    last = first->nxt;
    while(last->edg != last->nxt->edg) 
      last = last->nxt;
    
    cfirst = first;
    m.ident();
    m.rotat((360.0 / nfaces), 0.0, 0.0);
    
    while (--nfaces) {
      v = multmatrix(cfirst->nxt->vtx->vcoord, m);
      lmev(cfirst->nxt, cfirst->nxt, ++maxv, v[0], v[1], v[2]);
      scan = cfirst->nxt;
      while (scan != last->nxt) {
      	v = multmatrix(scan->prv->vtx->vcoord, m);
        lmev(scan->prv, scan->prv, ++maxv, v[0], v[1], v[2]);
        (void) lmef(scan->prv->prv, scan->nxt, ++maxf);
        scan = mate(scan->nxt->nxt);
      }

      last = scan;
      cfirst = mate(cfirst->nxt->nxt);
    }

    tailf = lmef(cfirst->nxt, mate(first), ++maxf);
    while (cfirst != scan) {
      lmef(cfirst, cfirst->nxt->nxt->nxt, ++maxf);
      cfirst = mate(cfirst->prv)->prv;
    }

    if (closed_figure == 1) {
      lkfmrh(headf, tailf);
#ifdef DEBUG
      cout << "lkfmrh done" << endl;
#endif
      headf->loopglue();
#ifdef DEBUG
      cout << "loopglue done" << endl;
#endif
    }
}

void Solid::ringmv(Id v1, Id v2, Id f1, Id f2, int out) {
    Face *ff1, *ff2;
    HalfEdge *dum;

    ff1 = findFace(f1);
    ff2 = findFace(f2);
    
    if (v2) 
      dum = ff1->findhe2(v1, v2);
    else   
      dum = ff1->findhe1(v1);

    if (! dum) {
      sprintf( ErrBuf, "[s%d,f%d]ringmv: Edge e%d,%d not found",
			solidno, f1, v1, v2 );
      FatalError( ErrBuf, 1 );
      exit(1);
    }
    
    lringmv(ff1, ff2, dum->wloop, out);
}

void Solid::aringmv(int f1, int f2) {
    laringmv(findFace(f1), findFace(f2));
}


double Solid::svolume() {
    Face *f;
    Loop *l;
    HalfEdge *he1, *he2;
    SpaceCoords c;
    double res;

    res = 0.0;
    f = sfaces;
    while (f) {
      l = f->floops;
      while (l) {
        he1 = l->ledg;
        he2 = he1->nxt;
        do {
          c = cross(he2->vtx->vcoord, he1->vtx->vcoord);
          res += (he2->nxt->vtx->vcoord * c) / 6.0;
        } while ((he2 = he2->nxt) != he1);
        l = l->nextl;
      }
      f = f->nextf;
    }
    return res;
}




Solid* Solid::duplicate()
{
    Solid    *solid = 0;
    Face     *f = 0,*aux_f = 0;
    Edge     *e = 0,*aux_e = 0;
    Vertex   *v = 0,*aux_v = 0;
    Loop     *l = 0,*aux_l = 0;
    HalfEdge *h = 0,*aux_h = 0,*aux2 = 0;

    solid          = new Solid(); assert(solid);
    solid->solidno = solidno;


    // A seguinte seguencia de codigo cria o esqueleto
    //   das listas do novo solido e apota os nos
    //   analogos de ambas enter si (Vertex <-> Vertex , ... , Face <-> Face).



    //  Cria a lista de faces...
    f     = sfaces;
    if (f != 0)
      {
	  aux_f = solid->sfaces = new Face();
          aux_f->prevf  = 0;
	  aux_f->info = (void*)f;
	  set_info(f->info,(void*)aux_f);
	  
	  while (f->nextf != 0)
	    {
		aux_f->nextf = new Face();
		aux_f->nextf->prevf = aux_f;
		aux_f = aux_f->nextf;
		f     = f->nextf;
		aux_f->info = (void*)f;
		set_info(f->info,(void*)aux_f);
	    }
	  aux_f->nextf = 0;
      }

    // Cria as listas de loops.
    f = sfaces;
    aux_f = solid->sfaces;
    while ( f!= 0)
      {
	  l = f->floops;
	  if (l != 0)
	    {
		aux_l = aux_f->floops = new Loop();
		aux_l->prevl  = 0;
		aux_l->info = (void*)l;
		set_info(l->info,(void*)aux_l);
		
		while (l->nextl != 0)
		  {
		      aux_l->nextl = new Loop();
		      aux_l->nextl->prevl = aux_l;
		      aux_l = aux_l->nextl;
		      l     = l->nextl;
		      aux_l->info = (void*)l;
		      set_info(l->info,(void*)aux_l);
		  }
		aux_l->nextl = 0;
	    }
	  f = f->nextf;
	  aux_f = aux_f->nextf;
      }

    // Cria as listas de half-edges.
    f = sfaces;
    aux_f = solid->sfaces;
    while ( f != 0)
      {
	  l = f->floops;
	  aux_l = aux_f->floops;
	  while (l != 0)
	    {
		h = l->ledg;
		if (h != 0)
		  {
		      aux_h = aux_l->ledg = new HalfEdge();
		      aux_h->info = (void*)h;
		      set_info(h->info,(void*)aux_h);

		      while (h->nxt != l->ledg)
			{
			    aux_h->nxt = new HalfEdge();
			    aux_h->nxt->prv = aux_h;
			    aux_h = aux_h->nxt;
			    h     = h->nxt;
			    aux_h->info = (void*)h;
			    set_info(h->info,(void*)aux_h);
			}
		      aux_l->ledg->prv = aux_h;
		      aux_h->nxt = aux_l->ledg;
		  }
		l = l->nextl;
		aux_l = aux_l->nextl;
	    }
	  aux_f = aux_f->nextf;
	  f = f->nextf;
      }

    
    // Cria a lista de Edges.
    e     = sedges;
    if (e != 0)
      {
	  aux_e = solid->sedges = new Edge();
          aux_e->preve  = 0;
	  aux_e->info = (void*)e;
	  set_info(e->info,(void*)aux_e);
	  
	  while (e->nexte != 0)
	    {
		aux_e->nexte = new Edge();
		aux_e->nexte->preve = aux_e;
		aux_e = aux_e->nexte;
		e     = e->nexte;
		aux_e->info = (void*)e;
		set_info(e->info,(void*)aux_e);
	    }
	  aux_e->nexte = 0;
      }


    // Cria a lista de vertices
    v     = sverts;
    if (v != 0)
      {
	  aux_v = solid->sverts = new Vertex();
          aux_v->prevv  = 0;
	  aux_v->info = (void*)v;
	  //v->info = (void*)aux_v;
	  set_info(v->info,(void*)aux_v);
	  
	  while (v->nextv != 0)
	    {
		aux_v->nextv = new Vertex();
		aux_v->nextv->prevv = aux_v;
		aux_v = aux_v->nextv;
		v     = v->nextv;
		aux_v->info = (void*)v;
		set_info(v->info,(void*)aux_v);
		
	    }
	  aux_v->nextv = 0;
      }


    // A seguinte sequencia de codigo aproveita do fato de os nos
    // analogos nas duas estruturas estarem ligados para
    // fazer a interconexao dos pointers internos da half-edge
    // recem-criada.

    // Interconexao das faces, arestas e loops.
    aux_f = solid->sfaces;
    while (aux_f!= 0)
      {
	  aux_l = aux_f->floops;
	  while (aux_l != 0)
	    {
		aux_h = aux_l->ledg;
		do 
		  {
		      h = (HalfEdge*)aux_h->info;
		      aux_h->edg = (Edge*)((duplicate_info_ptr)h->edg->info)->current;
		      aux_h->vtx = (Vertex*)((duplicate_info_ptr)h->vtx->info)->current;
		      aux_h->wloop = (Loop*)((duplicate_info_ptr)h->wloop->info)->current;
		      aux_h = aux_h->nxt;
		  } while (aux_h != aux_l->ledg);
		l = (Loop*)aux_l->info;
		aux_l->lface = (Face*)(((duplicate_info_ptr)l->lface->info)->current);
		aux_l = aux_l->nextl;
	    }

	  f = (Face*)aux_f->info;
	  aux_f->faceno = f->faceno;
	  aux_f->fsolid = solid;
	  aux_f->flout  = (Loop*)(((duplicate_info_ptr)f->flout->info)->current);
	  aux_f->feq    = f->feq;
	  aux_f = aux_f->nextf;
      }

    // Interconexao das Edges.
    aux_e = solid->sedges;
    while (aux_e!= 0)
      {
	  e = (Edge*)aux_e->info;
	  aux_e->he1  = (HalfEdge*)(((duplicate_info_ptr)e->he1->info)->current);
	  aux_e->he2  = (HalfEdge*)(((duplicate_info_ptr)e->he2->info)->current);
	  aux_e = aux_e->nexte;
      }

    // Interconexao dos vertices.
    aux_v = solid->sverts;
    while (aux_v!= 0)
      {
	  v = (Vertex*)aux_v->info;
	  aux_v->vedge   = (HalfEdge*)(((duplicate_info_ptr)v->vedge->info)->current);
	  aux_v->vcoord  = v->vcoord;
	  aux_v->vno = v->vno;
	  aux_v = aux_v->nextv;
      }




    // A seguinte sequencia de codigo remove as estruturas auliliares
    // armazenadas juntamente com cada no da half-edge original
    // e restabelece os valores antigos dos campos info desses nos.
      
    // Para  faces, loops e half-edges
    f = sfaces;
    while (f!= 0)
      {
	  l = f->floops;
	  while (l != 0)
	    {
		h = l->ledg;
		do 
		  {
		      restore_info(h->info);
		      h = h->nxt;
		  } while (h != l->ledg);
		restore_info(l->info);
		l = l->nextl;
	    }
	  restore_info(f->info);
	  f = f->nextf;
      }

    // Para Edges.
    e = sedges;
    while (e!= 0)
      {
	  restore_info(e->info);
	  e = e->nexte;
      }

    // Para Vertices.
    v = sverts;
    while (v!= 0)
      {
	  restore_info(v->info);
	  v = v->nextv;
      }

    
    return solid;
}




// Face methods

void Face::addloop(Loop *this_loop) {
    this_loop->nextl = floops;
    this_loop->prevl = 0;
    if (floops)
      floops->prevl = this_loop;
    floops = this_loop;
    this_loop->lface = this;
}

void Face::delloop(Loop *this_loop) {
    if (this_loop->prevl)
      this_loop->prevl->nextl = this_loop->nextl;
    if (this_loop->nextl)
      this_loop->nextl->prevl = this_loop->prevl;
    if (this_loop == floops)
      floops = this_loop->nextl;
}

HalfEdge *Face::findhe1(Id vn) {
    Loop *l = floops;
    HalfEdge *h;
    
    while (l) {
      h = l->ledg;
      do {
        if (h->vtx->vno == vn)
          return h;
      } while ((h = h->nxt) != l->ledg);
      l = l->nextl;
    }
    return 0;
}

HalfEdge *Face::findhe2(Id vn1, Id vn2) {
    Loop *l = floops;
    HalfEdge *h;
	        
    while (l) {
      h = l->ledg;
      do {
        if (h->vtx->vno == vn1 && h->nxt->vtx->vno == vn2)
          return h;
      } while ((h = h->nxt) != l->ledg);
      l = l->nextl;
    }
    return 0;
}

void Face::facels(int sw) {
    Loop *l;
    Vertex *v;
    Edge *e;
    HalfEdge *he;
    register int i;
    int breakline;

    breakline = 8;
    if (sw > 2) 
      breakline = 4;

    cout << "Face no. " << faceno << ":      ";

    l = floops;
    while (l) {
      he = l->ledg;
      i = 0;
      do {
        v = he->vtx;
        e = he->edg;
        if (sw > 2 && e) 
          if (he == e->he1)
            cout << v->vno << " +(" << e->he1->vtx->vno << " " << e->he2->vtx->vno << ")-> ";
          else    
            cout << v->vno << " -(" << e->he1->vtx->vno << " " << e->he2->vtx->vno << ")-> ";
        else 
          cout << v->vno << "-> ";
          
        if ((++i) % breakline == 0)
          cout << endl << "        ";
      }
      
      while ((he = he->nxt) != l->ledg);
      if (l == flout)
        cout << "<outer>" << endl;
      else
        cout << "<inner>" << endl;

      l = l->nextl;
      if (l) 
        cout << "        ";
    }

    cout << "Equation: " << feq << endl;
}

void Face::loopglue() {
    HalfEdge *h1, *h2, *h1next;

#ifdef DEBUG
    Solid *s;
    s = fsolid;
#endif

    h1 = floops->ledg;
    h2 = floops->nextl->ledg;

    while (!match(h1, h2)) 
      h2 = h2->nxt;
      
    lmekr(h1, h2);
    
#ifdef DEBUG
    cout << "lmekr: "; 
    s->solidls(2);
#endif
    
    lkev(h1->prv, h2->prv);
    
#ifdef DEBUG
    cout << "lkev: "; 
    s->solidls(2);
#endif
    
    while (h1->nxt != h2) {
      h1next = h1->nxt;
      (void) lmef(h1->nxt, h1->prv, -1);
      
#ifdef DEBUG
      cout << "lmef: "; 
      s->solidls(2);
#endif
      
      lkev(h1->nxt, mate(h1->nxt));
      
#ifdef DEBUG
      cout << "lkev: "; 
      s->solidls(2);
#endif
      
      lkef(mate(h1), h1);
      
#ifdef DEBUG
      cout << "lkef: "; 
      s->solidls(2);
#endif
      
      h1 = h1next;
    }
    
    lkef(mate(h1), h1);
    
#ifdef DEBUG
    cout << "lkef: "; 
    s->solidls(2);
#endif
}

// Loop methods

bool Loop::faceeq(SpaceCoords eq) {
    HalfEdge *he;
    double a, b, c, norm;
    double xi, yi, zi, xj, yj, zj, xc, yc, zc;
    int length;

    a = b = c = xc = yc = zc = 0.0;
    length = 0;
    he = ledg;
    do {
      xi = he->vtx->vcoord[0];
      yi = he->vtx->vcoord[1];
      zi = he->vtx->vcoord[2];
      xj = he->nxt->vtx->vcoord[0];
      yj = he->nxt->vtx->vcoord[1];
      zj = he->nxt->vtx->vcoord[2];

      a += (yi - yj) * (zi + zj);
      b += (zi - zj) * (xi + xj);
      c += (xi - xj) * (yi + yj);

      xc += xi;
      yc += yi;
      zc += zi;
      length++;
    }

    while ((he = he->nxt) != ledg);

    if ((norm = sqrt(a*a + b*b + c*c)) != 0.0) {
      eq[0] = a / norm;
      eq[1] = b / norm;
      eq[2] = c / norm;
      eq[3] = (eq[0]*xc + eq[1]*yc + eq[2]*zc) / (-length);
      return true;
    }
    else {
      sprintf( ErrBuf, "faceeq: Null face f%d", lface->faceno );
      Warning( ErrBuf );
//      cerr << "faceeq: null face " << lface->faceno << endl;
      return false;
    }
}

double Loop::loopar() {
    HalfEdge *he;
    Vertex *v1, *v2, *v3;
    SpaceCoords aa, bb, cc, dd, normal;

    dd[0] = dd[1] = dd[2] = 0.0;
    faceeq(normal);
    he = ledg;
    v1 = he->vtx;
    he = he->nxt;
    do {
      v2 = he->vtx;
      v3 = he->nxt->vtx;

      aa[0] = v2->vcoord[0] - v1->vcoord[0];
      aa[1] = v2->vcoord[1] - v1->vcoord[1];
      aa[2] = v2->vcoord[2] - v1->vcoord[2];

      bb[0] = v3->vcoord[0] - v1->vcoord[0];
      bb[1] = v3->vcoord[1] - v1->vcoord[1];
      bb[2] = v3->vcoord[2] - v1->vcoord[2];

      cc = cross(aa, bb);

      dd[0] += cc[0];
      dd[1] += cc[1];
      dd[2] += cc[2];
    } while ((he = he->nxt) != ledg);

    return 0.5 * (normal * dd);
}

// Vertex methods

void Vertex::listneighbors() {
    HalfEdge *adj;
    
    adj = vedge;
    if (adj)
      do {
      	cout << adj->nxt->vtx->vno << " ";
       } while ((adj = mate(adj)->nxt) != vedge);
    else
      cout << "Vertex: " << vno << " (no adjacent vertices)";
    cout << endl;
}

// euler operators (those appliable to a particular class are within them)

// extended euler operators



// This extended Euler operator eliminates the edges that incide on v and
// the vertex v itself. It is required that vertex v has at least one incident
// edge.
  
Face_ptr kill_rays(Vertex_ptr v)
{
    HalfEdge_ptr he1,he2;
    Face_ptr face = 0;
    bool remains = true;


    if (v->vedge == 0)
      {
	  cerr<<" Error : half_edge.C - kill_rays : No inicident edge on vertex v "<<endl;
	  return 0;
      }
    
    while (remains)
      {
	  he1 = v->vedge;
	  he2 = mate(he1);
	  
	  if (he1->wloop->lface == he2->wloop->lface)
	    if (he1->wloop == he2->wloop)
	      if (he1->prv == he2)
		{
		    face = he1->wloop->lface;
		    lkev(he1,he2);
		    remains = false;
		}
	      else
		if (he2->prv == he1)
		  {
		      face = he1->wloop->lface;
		      lkemr(he1,he2);
		  }
		else
		  {
		      face = he1->wloop->lface;
		      lkemr(he1,he2);
		  }
	    else
	      {
		  lmfkrh(he1->wloop,he1->wloop->lface->faceno);
		  face = he1->wloop->lface;
		  lkef(he1,he2);
		  cerr<<" Tested for the First Time ! 2 "<<endl;
	      }
	  else
	    {
		face = he1->wloop->lface;
		lkef(he1,he2);
	    }
      }

    return face;
}

  

HalfEdge *addhe(Edge *e, Vertex *v, HalfEdge *where, int sign) {
    HalfEdge *he;
    
    if (!where->edg) 
      he = where;
    else {
      he = new HalfEdge();
      assert(he);
      where->prv->nxt = he;
      he->prv = where->prv;
      where->prv = he;
      he->nxt = where;
    }

    he->edg = e;
    he->vtx = v;
    he->wloop = where->wloop;

    if (sign == PLUS)
      e->he1 = he;
    else
      e->he2 = he;

    return he;
}

HalfEdge *delhe(HalfEdge *he) {
    if (!he->edg) {
      delete he;
      return 0;
    }
    else 
      if (he->nxt == he) {
      	he->edg = 0;
      	return he;
      }
      else {
        he->prv->nxt = he->nxt;
        he->nxt->prv = he->prv;
        delete he;
        return he->prv;
      }
}

void lmev(HalfEdge *he1, HalfEdge *he2, Id vn, float x, float y, float z) {
    HalfEdge *he;
    Vertex *newvertex;
    Edge *newedge;

#ifdef DEBUG
    Solid *s;
    s = he1->wloop->lface->fsolid;
#endif
    
    newedge = new Edge(he1->wloop->lface->fsolid);
    assert(newedge);
    newvertex = new Vertex(he1->wloop->lface->fsolid);
    assert(newvertex);

    newvertex->vcoord[0] = x;
    newvertex->vcoord[1] = y;
    newvertex->vcoord[2] = z;
    newvertex->vcoord[3] = 1.0;
    newvertex->vno = vn;

    he = he1;
    while (he != he2) {
      he->vtx = newvertex;
      he = mate(he)->nxt;
    }

    addhe(newedge, he2->vtx, he1, MINUS);
    addhe(newedge, newvertex, he2, PLUS);

    newvertex->vedge = he2->prv;
    he2->vtx->vedge = he2;

#ifdef DEBUG
    s->checkv();
#endif
}



void lmev(EulerSequence& inv,HalfEdge *he1, HalfEdge *he2, Id vn, float x, float y, float z )
{
   inv.push(he1->wloop->lface->faceno,0,he1->vtx->vno,vn,0,0,0,0,0,KEV);
   lmev(he1,he2,vn,x,y,z);
}






void lkev(HalfEdge *he1, HalfEdge *he2) {
    HalfEdge     *he;

#ifdef DEBUG
    Solid *s;
    s = he1->wloop->lface->fsolid;
#endif

    he = he2->nxt;
    while (he != he1) {
      he->vtx = he2->vtx;
      he = mate(he)->nxt;
    }

    he1->wloop->ledg = delhe(he1);
    he2->wloop->ledg = delhe(he2);
    he2->vtx->vedge = he2->nxt;
    
    if (!he2->vtx->vedge->edg) 
      he2->vtx->vedge = 0;

    he1->wloop->lface->fsolid->deledge(he1->edg); 
    he1->wloop->lface->fsolid->delvertex(he1->vtx);
    delete he1->edg;
    delete he1->vtx;

#ifdef DEBUG
    s->checkv();
#endif
}


void lkev(EulerSequence& inv,HalfEdge *he1, HalfEdge *he2)
{
    Id v1,v2,v3,v4;
    Id f1,f2;
    double x,y,z;

    f1 = he2->wloop->lface->faceno;
    f2 = he1->wloop->lface->faceno;

    v1 = he2->vtx->vno;
    v4 = he1->vtx->vno;

    x = he1->vtx->vcoord[Space_XCoord];
    y = he1->vtx->vcoord[Space_YCoord];
    z = he1->vtx->vcoord[Space_ZCoord];

    if ((he2->nxt == he1)&&(he1->nxt == he2))
	v2 = v3 = v1;
    else if ((he2->nxt == he1)&&(he1->nxt != he2))
      {
	  v2 = v3 = he1->nxt->nxt->vtx->vno;
      }
    else if ((he2->nxt != he1)&&(he1->nxt == he2))
      {
	  v2 = v3 = he2->nxt->nxt->vtx->vno;
      }
    else
      {
	  v2 = he2->nxt->nxt->vtx->vno;
          v3 = he1->nxt->nxt->vtx->vno;
      };
    inv.push(f1,f2,v1,v2,v3,v4,x,y,z,MEV);
    lkev(he1,he2);
}



Face *lmef(HalfEdge *he1, HalfEdge *he2, Id nf) {
    Face *newface;
    Loop *newloop;
    Edge *newedge;
    HalfEdge *he, *nhe1, *nhe2, *temp;

    newface = new Face(he1->wloop->lface->fsolid);
    assert(newface);
    newloop = new Loop(newface);
    assert(newloop);
    newedge = new Edge(he1->wloop->lface->fsolid);
    assert(newedge);
    
    newface->faceno = nf;
    newface->flout = newloop;

    he = he1;
    while (he != he2) {
      he->wloop = newloop;
      he = he->nxt;
    }

    nhe1 = addhe(newedge, he2->vtx, he1, MINUS);
    nhe2 = addhe(newedge, he1->vtx, he2, PLUS);

    nhe1->prv->nxt = nhe2;
    nhe2->prv->nxt = nhe1;
    temp = nhe1->prv;
    nhe1->prv = nhe2->prv;
    nhe2->prv = temp;

    newloop->ledg = nhe1;
    he2->wloop->ledg = nhe2;

    return newface;
}



Face *lmef(EulerSequence& inv,HalfEdge *he1, HalfEdge *he2, Id nf)
{
    inv.push(he2->wloop->lface->faceno,0,he1->vtx->vno,he2->vtx->vno,0,0,0,0,0,KEF);
    return (lmef(he1,he2,nf));
}

void lkef(HalfEdge *h1, HalfEdge *h2) {
    Face *f1, *f2;
    Loop *l, *l1, *l2;
    register HalfEdge *he;

    Solid *s;
    s = h1->wloop->lface->fsolid;
    
    l1 = h1->wloop;
    f1 = l1->lface;
    l2 = h2->wloop;
    f2 = l2->lface;

    while (l = f2->floops) {
      f2->delloop(l); 
      f1->addloop(l);
    }

    he = l2->ledg;
    do {
        he->wloop = l1;
    } while ((he = he->nxt) != l2->ledg);

    h1->prv->nxt = h2;
    h2->prv->nxt = h1;
    he = h2->prv;
    h2->prv = h1->prv;
    h1->prv = he;

    delhe(h2);
    delhe(h1);

    h2->vtx->vedge = h1->nxt;
    if (!h2->vtx->vedge->edg) 
      h2->vtx->vedge = 0;
    h1->vtx->vedge = h2->nxt;
    if (!h1->vtx->vedge->edg) 
      h1->vtx->vedge = 0;

    l1->ledg = h1->nxt;

    f1->fsolid->delface(f2); 
    f1->delloop(l2); 
    f1->fsolid->deledge(h2->edg); 
    delete f2;
    delete l2;
    delete h2->edg;

#ifdef DEBUG
    s->checkv();
#endif

    
}



void lkef(EulerSequence& inv,HalfEdge *h1, HalfEdge *h2)
{
    Face  *f1, *f2;
    Loop  *l , *l1, *l2;


    l1 = h1->wloop;
    f1 = l1->lface;
    l2 = h2->wloop;
    f2 = l2->lface;

    for(l = f2->floops; l != 0 ;l = l->nextl)
      {
	  if (l != l2)
	    {
		inv.push(f1->faceno,f2->faceno,l->ledg->vtx->vno,l->ledg->nxt->vtx->vno,(l == f1->flout ? 1 : 0), 0, 0,0,0,RINGMV);
	    }
      };

    inv.push(f1->faceno,f2->faceno,h2->nxt->vtx->vno,h2->nxt->nxt->vtx->vno,h1->nxt->vtx->vno,h1->nxt->nxt->vtx->vno,0,0,0,MEF);


    lkef(h1,h2);
    
}



void lmekr(HalfEdge *he1, HalfEdge *he2) {
    Edge *n_ed;
     
    register HalfEdge *n_he1, *n_he2;
    Face *f;
    Loop *l1, *l2;

    l1 = he1->wloop;
    l2 = he2->wloop;
    f = l1->lface;

    n_he1 = l2->ledg;
    do {
      n_he1->wloop = l1;
    } while ((n_he1 = n_he1->nxt) != l2->ledg);

    n_ed = new Edge(f->fsolid);
    assert(n_ed);

    n_he1 = addhe(n_ed, he1->vtx, he1, PLUS);
    n_he2 = addhe(n_ed, he2->vtx, he2, MINUS);

    n_he1->nxt = he2;
    n_he2->nxt = he1;
    he2->prv = n_he1;
    he1->prv = n_he2;

    if (f->flout == l2) 
      f->flout = l1;
 
    l2->lface->delloop(l2); delete l2;
}

void lmekr(EulerSequence& inv,HalfEdge *he1, HalfEdge *he2)
{
    inv.push(he1->wloop->lface->faceno,0,he2->vtx->vno,he1->vtx->vno,0,0,0,0,0,KEMR);
    lmekr(he1,he2);
}


void lkemr(HalfEdge *h1, HalfEdge *h2) {
    register HalfEdge *h3, *h4;
    Loop *nl;
    Loop *ol;
    Edge *killedge;

#ifdef DEBUG
    Solid *s;
    s = h1->wloop->lface->fsolid;
#endif

    ol = h1->wloop;
    nl = new Loop(ol->lface);
    assert(nl);
    killedge = h1->edg;

    h3 = h1->nxt;
    h1->nxt = h2->nxt;
    h2->nxt->prv = h1;
    h2->nxt = h3;
    h3->prv = h2;

    h4 = h2;
    do {
      h4->wloop = nl;
    } while ((h4 = h4->nxt) != h2);

    ol->ledg = h3 = delhe(h1);
    nl->ledg = h4 = delhe(h2);

    h3->vtx->vedge = (h3->edg) ? h3 : 0;
    h4->vtx->vedge = (h4->edg) ? h4 : 0;
    h3->nxt->vtx->vedge = (h3->nxt->edg) ? h3->nxt : 0;
    h4->nxt->vtx->vedge = (h4->nxt->edg) ? h4->nxt : 0;

    ol->lface->fsolid->deledge(killedge); delete killedge;  

#ifdef DEBUG
    s->checkv();
#endif
}


void lkemr(EulerSequence& inv,HalfEdge *h1, HalfEdge *h2)
{
    inv.push(h1->wloop->lface->faceno,0,h1->nxt->vtx->vno,h1->nxt->nxt->vtx->vno,h2->nxt->vtx->vno,h2->nxt->nxt->vtx->vno,0,0,0,MEKR);
    lkemr(h1,h2);
}



Face *lmfkrh(Loop *l, Id fn) {
    Face *nf;

    nf = new Face(l->lface->fsolid);
    assert(nf);
    nf->faceno = fn;
    nf->flout = l;
    l->lface->delloop(l); 
    nf->addloop(l);
    return nf;
}


Face *lmfkrh(EulerSequence& inv,Loop *l, Id fn)
{
    inv.push(l->lface->faceno,fn,0,0,0,0,0,0,0,KFMRH);
    return(lmfkrh(l,fn));
}

void lkfmrh(Face *f1, Face *f2) {
    Loop *l;

    while (l = f2->floops) {
      f2->delloop(l);
      f1->addloop(l);
    }
    f1->fsolid->delface(f2); delete f2;
}


void lkfmrh(EulerSequence& inv,Face *f1, Face *f2)
{
    inv.push(f1->faceno,f2->faceno,f2->flout->ledg->vtx->vno,f2->flout->ledg->nxt->vtx->vno,0,0,0,0,0,MFKRH);
    lkfmrh(f1,f2);
}


void lringmv(Face *f1, Face *f2, Loop *l, int out) {
    f1->delloop(l);
    f2->addloop(l);
    if (out) 
      f2->flout = l;
}


void lringmv(EulerSequence& inv,Face *f1, Face *f2, Loop *l, int out)
{
    inv.push(f2->faceno,f1->faceno,l->ledg->vtx->vno,l->ledg->nxt->vtx->vno, (l == f1->flout) ? 1 : 0 , 0 , 0 , 0 , 0,RINGMV);
    lringmv(f1,f2,l,out);
}


void laringmv(Face *f1, Face *f2) {
    SpaceCoords tmp;
    int drop, in;
    Loop *ring;
    Loop *l;

    (void)f1->flout->faceeq(f1->feq);
    drop = f1->feq.drop();
    l = f1->floops;
    while (l) {
      if ((ring = l) != f1->flout) {
        l = l->nextl;
        in = contfv2(f1, ring->ledg->vtx, drop);
        if (!in) {
          f1->delloop(ring);
          f2->addloop(ring);
        }
      }
      else 
        l = l->nextl;
    }
}


// utilities

void merge2(Solid *s1, Solid *s2) {
    Face *f;
    Edge *e;
    Vertex *v;

    while (f = s2->sfaces) {
      s2->delface(f);
      s1->addface(f);
    }

    while (e = s2->sedges) {
      s2->deledge(e);
      s1->addedge(e);
    }

    while (v = s2->sverts) {
      s2->delvertex(v);
      s1->addvertex(v);
    }

    delete s2;
}

int match(HalfEdge *h1, HalfEdge *h2) {
    return contvv(h1->vtx, h2->vtx);
}

int contvv(Vertex *v1, Vertex *v2) {
    double v[3];
    double diff;

    v[0] = v1->vcoord[0] - v2->vcoord[0];
    v[1] = v1->vcoord[1] - v2->vcoord[1];
    v[2] = v1->vcoord[2] - v2->vcoord[2];

    diff = v[0] * v[0] + v[1] * v[1] + v[2] * v[2];
    return diff < EPS;
}

int contfv(Face *f, Vertex *v) {
    int drop;

    (void)f->flout->faceeq(f->feq);
    drop = f->feq.drop();
    return contfv2(f, v, drop);
}

int contfv2(Face *f, Vertex *v, int drop) {
    Loop *l;
    int cont;

    cont = contlv(f->flout, v, drop);

    if (cont == 0)
      return 0;
    if (cont > 1)
      return cont;

    l = f->floops;
    while (l) {
      if ((l != f->flout) && l->ledg->nxt->nxt != l->ledg) {
      	cont = contlv(l, v, drop);
        if (cont == 1) 
          return 0;
        if (cont > 1) 
          return cont;
      }
      l = l->nextl;
    }
    return 1;
}

int contlv(Loop *l, Vertex *v, int drop) {
    HalfEdge *he1, *he2;
    Vertex aux;
    double t1, t2;
    int intersections, intr, c1, c2;

    HalfEdge *hithe;
    Vertex *hitvertex;

#ifdef DEBUG
    cout << "contlv: vertex " << v->vno << " vs. face " << l->lface->faceno << endl;
#endif

    if ((intr = bndrlv(l, v, hithe, hitvertex)) > 0)
      return intr;

    he2 = l->ledg;

label:
    aux.vcoord[0] = (he2->vtx->vcoord[0] + he2->nxt->vtx->vcoord[0]) / 2.0;
    aux.vcoord[1] = (he2->vtx->vcoord[1] + he2->nxt->vtx->vcoord[1]) / 2.0;
    aux.vcoord[2] = (he2->vtx->vcoord[2] + he2->nxt->vtx->vcoord[2]) / 2.0;
    he1 = l->ledg;
    intersections = 0;
    do {
      intr = int2ee(v, &aux, he1->vtx, he1->nxt->vtx, drop, t1, t2);
      if (intr == 1) {
        c1 = comp(t2, 0.0, EPS);
        c2 = comp(t2, 1.0, EPS);
        if (c1 == Equal || c2 == Equal) {
          he2 = he2->nxt;
          if (he2 == l->ledg) 
            return false;
          goto label;
        }
        if (c1 == Greater && c2 == Less)
          if (t1 >= 0.0) 
            intersections++;
      }
    } while ((he1 = he1->nxt) != l->ledg);
    
#ifdef DEBUG
    cout << "contlv: intersections = " << intersections << endl;
#endif

    intersections = intersections % 2;
    return intersections;
}

int contev(Vertex *v1, Vertex *v2, Vertex *v3) {
    double  t;

    if (intrve(v1, v2, v3, t))
      if (t >= (-EPS) && t <= (1.0+EPS))
        return 1;
    return 0;
}

int bndrlv(Loop *l, Vertex *v, HalfEdge *&hithe, Vertex *&hitvertex) {
    HalfEdge *he;

    he = l->ledg;
    do {
      if (contvv(he->vtx, v)) {
        hitvertex = he->vtx;
        hithe = 0;
        return 3;
      }
    } while ((he = he->nxt) != l->ledg);
    he = l->ledg;
    do {
      if (contev(he->vtx, he->nxt->vtx, v)) {
        hitvertex = 0;
        hithe = he;
        return 2;
      }
    } while ((he = he->nxt) != l->ledg);

    return 0;
}


int int2ee(Vertex *v1, Vertex *v2, Vertex *v3, Vertex *v4, int drop, double &t1, double &t2) {
    double D1, D2, D, a1, a2, b1, b2, c1, c2;

    switch (drop) {
      case Space_XCoord:
        a1 = v2->vcoord[1] - v1->vcoord[1];
        a2 = v2->vcoord[2] - v1->vcoord[2];
        b1 = v3->vcoord[1] - v4->vcoord[1];
        b2 = v3->vcoord[2] - v4->vcoord[2];
        break;

      case Space_YCoord:
        a1 = v2->vcoord[0] - v1->vcoord[0];
        a2 = v2->vcoord[2] - v1->vcoord[2];
        b1 = v3->vcoord[0] - v4->vcoord[0];
        b2 = v3->vcoord[2] - v4->vcoord[2];
        break;

      case Space_ZCoord:
        a1 = v2->vcoord[0] - v1->vcoord[0];
        a2 = v2->vcoord[1] - v1->vcoord[1];
        b1 = v3->vcoord[0] - v4->vcoord[0];
        b2 = v3->vcoord[1] - v4->vcoord[1];
        break;
    }

    D = a1*b2 - a2*b1;
    if (zero(D, EPS))
      return 0;
    else {
      switch (drop) {
        case Space_XCoord:
          c1 = v1->vcoord[1] - v3->vcoord[1];
          c2 = v1->vcoord[2] - v3->vcoord[2];
          break;
    
        case Space_YCoord:
          c1 = v1->vcoord[0] - v3->vcoord[0];
          c2 = v1->vcoord[2] - v3->vcoord[2];
          break;

        case Space_ZCoord:
          c1 = v1->vcoord[0] - v3->vcoord[0];
          c2 = v1->vcoord[1] - v3->vcoord[1];
          break;
      }

      D1 = c2*b1 - c1*b2;
      D2 = a2*c1 - a1*c2;
      t1 = D1/D;
      t2 = D2/D;
  
      return 1;
    }
}

int intrve(Vertex *v1, Vertex *v2, Vertex *v3, double &t) {
    Vertex testv;
    SpaceCoords r1, r2;
    float r1r1, tprime;

    r1[0] = v2->vcoord[0] - v1->vcoord[0];
    r1[1] = v2->vcoord[1] - v1->vcoord[1];
    r1[2] = v2->vcoord[2] - v1->vcoord[2];

    r1r1 = r1 * r1;
    if (r1r1 < EPS * EPS) {
      t = 0.0;
      return contvv(v1, v3);
    }
    else {
      r2[0] = v3->vcoord[0] - v1->vcoord[0];
      r2[1] = v3->vcoord[1] - v1->vcoord[1];
      r2[2] = v3->vcoord[2] - v1->vcoord[2];
      
      tprime = (r1 * r2) / r1r1;
      testv.vcoord[0] = v1->vcoord[0] + tprime * r1[0];
      testv.vcoord[1] = v1->vcoord[1] + tprime * r1[1];
      testv.vcoord[2] = v1->vcoord[2] + tprime * r1[2];
      t = tprime;
      return contvv(&testv, v3);
    }
}









