/*
    Copyright (C) 1992-1994
                    Cesar N. Gon

    All rights reserved.

    This program is  free software; you can redistribute unmodified
    versions of it as long as you do not  profit from that, neither
    by selling it nor  by   including it  with other products   for
    sale.  This copyright  notice  must  remain in all  distributed
    copies.

    You are  allowed to modify your  personal copy  of this program
    (as long as you do not distribute any changed versions).

    You may,  of course, distribute  along with it, your own source
    code or libraries of algorithms that run under GeoLab.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

    If you must contact the authors, you may write to:
                     geolab-request@dcc.unicamp.br
*/

/*
 *
 * input_tools.C - 
 *
 *             link flag -lm
 *
 *
 *    	Data	Hora  	Nome     		Modificacao/Motivo
 *
 *    13.02.92  08:38 	Cesar N. Gon  	         Arquivo criado
 *
 */

#include "input_tools.h"
#include "Polygon.h"
#include "Circle.h"
#include "Ellipse.h"
#include "utilities.h"
#include "objects.h"
#include "basic.h"
#include "List.h"
#include "benchmark.h"
#include "Interval_set.h"
#include "tools.h"
#include "geolab_ui.h"

#include <sys/time.h>
#include <xview/xview.h>
#include <xview/frame.h>
#include <xview/panel.h>
#include <xview/notice.h>


#ifndef _list_item_list
#define _list_item_list
declare(list,list_item)
#endif ~_list_item_list;

#ifndef _GraphicObj_ptr_list
#define _GraphicObj_ptr_list
declare(list,GraphicObj_ptr)
#endif ~_GraphicObj_ptr_list;

#ifndef _Point2D_ptr_list
#define _Point2D_ptr_list
declare(list,Point2D_ptr_list)
#endif ~_Point2D_ptr_list;

#ifndef _Segment2D_ptr_interval_set
#define _Segment2D_ptr_interval_set
declare(interval_set,Segment2D_ptr)
#endif ~_Segment2D_ptr_interval_set

     
/*
 * Variaveis exportaveis (via extern) que podem ser ajustadas via janela
 * de properties. 
 *
 */

int POLYGON_TYPE                 = 1; // Simple_t
int POLYGON_SELECTED             = 0;
int POLYGON_DISCARD_INPUT        = 0;
int POLYGON_UNSELECT_INPUT       = 0;
int COVER_SELECTED               = 0;
int COVER_DISCARD_INPUT          = 0;
int COVER_SCALE                  = 1;
int COVER_POINTS                 = 50;
int COVER_UNSELECT_INPUT         = 0;
int COVER_TIMEOUT                = 0;
int RANDOM_POINTS_SELECTED       = 0;
int RANDOM_POINTS_DISCARD_INPUT  = 0;
int RANDOM_POINTS_SCALE          = 1;
int RANDOM_POINTS_NUMBER         = 50;
int RANDOM_POINTS_TIMEOUT        = 1;
int RANDOM_POINTS_ALL_REGION     = 1;
int RANDOM_POINTS_UNSELECT_INPUT = 0;

#define TIMEOUT_PRECISION 100


// prototypes de funcoes auxiliares

void Calc_a_b (Ellipse_ptr E, double& a, double& b);
int  test_for_timeout (World *w, int timeout_value);
int option_timeout(World *w);


void random_points_schedule (World *w)
{
    list(Obj_ptr) Points;
    Point2D_ptr p_ptr;
    Obj_ptr obj;
    register int i;
    double x1, y1, x2, y2;
    int stop = 0; /* 0. cont; 1. stop -1. get_input */

    start_timer();
    if (RANDOM_POINTS_ALL_REGION) {
	w->get_values (x1,y1,x2,y2);
	regularize_coords (x1,y1,x2,y2);
	for (i=0;i<RANDOM_POINTS_NUMBER;i++) {
	    p_ptr = new Point2D(random_point_in_box (x1, y1, x2, y2));
	    assert(p_ptr);
	    Points.push(p_ptr);
	    if (RANDOM_POINTS_TIMEOUT && !(i % TIMEOUT_PRECISION)) {
		stop = test_for_timeout (w, RANDOM_POINTS_TIMEOUT);
		if (stop) break;
	    }
	 }
    }
    else {
	list(list_item)* SelectedObjects=&objects_get_selected_objects();
	list(GraphicObj_ptr)* GraphicObjects=&objects_get_graphic_objects();
	list_item it;
	register int out;
	register int in;
	
	forall (it,*SelectedObjects) {
	    obj=(*GraphicObjects)[it]->obj;
	    if (obj->point_inside_implemented() && obj->bounding_box_implemented()) {
		obj->bounding_box (x1, y1, x2, y2);
		regularize_coords (x1, y1, x2, y2);
		for (in=0,out=0,i=0;in<RANDOM_POINTS_NUMBER;i++) {
		    p_ptr = new Point2D(random_point_in_box (x1, y1, x2, y2));
		    assert(p_ptr);
		    if (obj->point_inside (p_ptr->X(),p_ptr->Y())) {
			Points.push(p_ptr);
			in++;
		    }
		    else {
			out++;
			delete p_ptr;
		    }
		    if (RANDOM_POINTS_TIMEOUT && !(i % TIMEOUT_PRECISION)) {
			stop = test_for_timeout (w, RANDOM_POINTS_TIMEOUT);
			if (stop) break;
		    }
		}
		if (stop) break;
	    }
	}
    }
    if (stop != 1) {
	if (RANDOM_POINTS_DISCARD_INPUT) objects_delete_selected(w);
	else if (RANDOM_POINTS_UNSELECT_INPUT) objects_unselect(w);
	// install points
	objects_install_created_objects (w,RANDOM_POINTS_SELECTED,Points);
    }
    else
      forall (obj, Points)
	delete obj;
}

void random_cover_schedule (World *w)
{
    list(Obj_ptr) Points;
    Point2D_ptr p_ptr;
    register int i;
    interval_set(Segment2D_ptr) S;
    list(Obj_ptr) Segments;
    Segment2D_ptr s_ptr;
    is_item is_it;
    list_item it;
    double x, t, rv;
    int stop = 0;

    start_timer();
    list(list_item)*SelectedObjects=&objects_get_selected_objects();
    list(GraphicObj_ptr)*GraphicObjects=&objects_get_graphic_objects();
    Obj_ptr obj, obj1;

    forall (it,*SelectedObjects) {
	obj=(*GraphicObjects)[it]->obj;
	if (obj->get_segments_implemented()) {
	    Segments = obj->get_segments();
	    switch (obj->type()) {
	      case Circle_t: {		
		  Point2D c (((Circle_ptr)obj)->center());
		  double r = ((Circle_ptr)obj)->radius();
		  for (i=0;i<COVER_POINTS;i++) {
		      t = random_value_in_interval (0, 2*PI);
		      p_ptr = new Point2D; assert(p_ptr);
		      p_ptr->setXY(c.X()+r*cos(t), c.Y()+r*sin(t));
		      Points.push(p_ptr);
		      if (COVER_TIMEOUT && !(i % TIMEOUT_PRECISION)) {
			  stop = test_for_timeout (w, COVER_TIMEOUT);
			  if (stop) break;
		      }
		  }
	      }
		break;
	      case Ellipse_t: {
		  Ellipse_ptr E = (Ellipse_ptr)obj;
		  double a,b;
		  Calc_a_b (E,a,b);
		  Point2D c (((*E).P1().X()+(*E).P2().X())/2,((*E).P1().Y()+(*E).P2().Y())/2);
		  for (i=0;i<COVER_POINTS;i++) {
		      t = random_value_in_interval (0, 2*PI);
		      p_ptr = new Point2D; assert(p_ptr);
		      p_ptr->setXY(c.X()+a*cos(t), c.Y()+b*sin(t));
		      Points.push(p_ptr);
		      if (COVER_TIMEOUT && !(i % TIMEOUT_PRECISION)) {
			  stop = test_for_timeout (w, COVER_TIMEOUT);
			  if (stop) break;
		      }
		  }
	      }
		break;
	      default:
		x=0;
		forall (obj1, Segments) {
		    s_ptr=(Segment2D_ptr)obj1;
		    S.insert (real(x),real(x=x+s_ptr->P1().e_distance (s_ptr->P2())),s_ptr);
		}
		for (i=0;i<COVER_POINTS;i++) {
		    rv = random_value_in_interval (0,(double)x);		
		    is_it = (S.intersection(real(rv),real(rv))).head();
		    s_ptr = S.inf(is_it);
		    t = (rv - S.left(is_it))/s_ptr->P1().e_distance (s_ptr->P2());
		    p_ptr = new Point2D; assert(p_ptr);
		    p_ptr->setXY((s_ptr->P1()).X()+((s_ptr->P2()).X()-(s_ptr->P1()).X())*t, (s_ptr->P1()).Y()+((s_ptr->P2()).Y()-(s_ptr->P1()).Y())*t);
		    Points.push(p_ptr);
		    if (COVER_TIMEOUT && !(i % TIMEOUT_PRECISION)) {
			stop = test_for_timeout (w, COVER_TIMEOUT);
			if (stop) break;
		    }
		}
	    }
	    S.clear();
	    forall (obj, Segments)
		delete obj;
	    Segments.clear();
	}
	if (stop) break;
    } /* for */
    if (stop != 1) {
	if (COVER_DISCARD_INPUT) objects_delete_selected(w);
	else if (COVER_UNSELECT_INPUT) objects_unselect(w);
	// install points
	objects_install_created_objects (w,COVER_SELECTED,Points);
    }
    else
      forall (obj, Points)
	delete obj;
}

void random_polygon_schedule (World *w)
{
    list(list_item)* SelectedObjects=&objects_get_selected_objects();
    list(GraphicObj_ptr)* GraphicObjects=&objects_get_graphic_objects();
    list(Point2D_ptr) Points;
    list_item it;
    Obj_ptr obj;

    start_timer();
    // extrai todos pontos selecionados
    forall (it,*SelectedObjects) {
	obj=(*GraphicObjects)[it]->obj;
	if (obj->type()==Point2D_t)
	  Points.push((Point2D_ptr)obj);
    }
    
    if (POLYGON_DISCARD_INPUT) objects_delete_selected(w);
    else if (POLYGON_UNSELECT_INPUT) objects_unselect(w);

    if (!Points.empty()) {
    
	Polygon_ptr P = new Polygon(Points); assert(P);
    
	randomize();
    
	switch (POLYGON_TYPE) {
	  case 0 :
	    P->permute();
	    break;
	  case 1 :
	    P->simple();
	    break;
	  case 2 :
	    P->permute();
	    P->star_shaped(*(P->head()));
	    break;
	  case 3 :
	    P->convex();
	    break;
	  default: error ("input_tools.C","random_polygon_schedule","unknown value of POLYGON_TYPE");
	}
	
	// install the polygon created
	  objects_install_created_objects (w,POLYGON_SELECTED,P);
    }
}

void print_selected_objects ()
{
    list(list_item)* SelectedObjects=&objects_get_selected_objects();
    list(GraphicObj_ptr)* GraphicObjects=&objects_get_graphic_objects();
    list_item it;
    Obj_ptr obj;

    cout << "\n Selected Objects:\n";
    forall (it, *SelectedObjects) {
	obj=(*GraphicObjects)[it]->obj;
	obj->print();
    }    
}
			   

void print_objects ()
{
    list(GraphicObj_ptr)* GraphicObjects=&objects_get_graphic_objects();
    GraphicObj_ptr g_obj;

    cout << "\n Objects:\n";
    forall (g_obj, *GraphicObjects)
      g_obj->obj->print();    
}



/* inicializa o gerador de numeros aleatorios em funcao do tempo */
void randomize ()
{
    struct timeval now_tp;
    struct timezone now_tzp;
    gettimeofday (&now_tp,&now_tzp);
    srand ((int)now_tp.tv_sec);
}


Point2D random_point_in_box (double xmin, double ymin, double xmax, double ymax)
{
    Point2D rp;
    rp.setXY (xmin + rand() % (long)(xmax-xmin), ymin + rand() % (long)(ymax-ymin));
    return rp;
}


double random_value_in_interval (double a, double b)
{
    double t = (double)(rand() % 10000)/10000;
    return a+t*(b-a);
}


void put_warning_left (World *w, char *msg)
{
    xv_set((Frame)xv_get(w->get_canvas(), XV_OWNER), FRAME_LEFT_FOOTER,msg, XV_NULL);
}


void Calc_a_b (Ellipse_ptr E, double& a, double& b)
{
    double x1 = E->P1().X(), y1 = E->P1().Y(), x2 = E->P2().X(), y2 = E->P2().Y();
    regularize_coords (x1,y1,x2,y2);
    a = (x2-x1)/2;
    b = (y2-y1)/2;
}


int test_for_timeout (World *w, int timeout_value)
{
    int ans;
    double x = get_real_ellapsed_time();
    if (x>timeout_value /*60*/) {
	 ans = option_timeout(w);
	 if (!ans) start_timer();
	 return ans;
    }
    else
      return 0;
}


int option_timeout(World *w)
{
    Frame frame = (Frame)xv_get(w->get_canvas(), XV_OWNER);
    geolab_geolab_base_window_objects  *igp = (geolab_geolab_base_window_objects *) xv_get(frame, XV_KEY_DATA, INSTANCE);
    
    int op = (int)notice_prompt(igp->geolab_base_window, NULL, 
			NOTICE_MESSAGE_STRINGS,    "Timeout occurred. "
				                   "Do you want to abort ?",
    						   NULL,
                        NOTICE_BUTTON,		   "Continue",             0,
			NOTICE_BUTTON,		   "Abort, Ignore Points", 1,
		        NOTICE_BUTTON,             "Abort, Get Points",   -1,
			NULL);
    return op;
}
