/*
    Copyright (C) 1992-1994
                    Welson R. Jacometti

    All rights reserved.

    This program is  free software; you can redistribute unmodified
    versions of it as long as you do not  profit from that, neither
    by selling it nor  by   including it  with other products   for
    sale.  This copyright  notice  must  remain in all  distributed
    copies.

    You are  allowed to modify your  personal copy  of this program
    (as long as you do not distribute any changed versions).

    You may,  of course, distribute  along with it, your own source
    code or libraries of algorithms that run under GeoLab.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

    If you must contact the authors, you may write to:
                     geolab-request@dcc.unicamp.br
*/

/*
 * intersections.C - algoritmos para problemas de intersecao. Sao considerados
 *                   elementos de entrada sendo segmentos, raios e retas.
 * 
 *
 *    	Data	Hora  	Nome     		Modificacao/Motivo
 *
 *    24.04.92  09:30 	Welson R. Jacometti  	Arquivo criado
 *
 */

#include "intersections.h"

/*
 **** variaveis para animacao ***/

#include "Integer.h"
#include "Circle.h"

static GraphicCircle *anim_c; 
static GraphicSegment2D *anim_s;
static GraphicRay2D *anim_r;
static GraphicLine2D *anim_l;
static GraphicInteger *anim_i;

#define anim_circle	((Circle *)anim_c->obj)
#define anim_segment	((Segment2D *)anim_s->obj)
#define anim_ray	((Ray2D *)anim_r->obj)
#define anim_line	((Line2D *)anim_l->obj)
#define anim_integer	((Integer *)anim_i->obj)

void intersections_animation_init(World *w) {
    static bool FIRST = true;
        
    if (FIRST) {
      anim_c = new GraphicCircle();
      anim_s = new GraphicSegment2D();
      anim_r = new GraphicRay2D();
      anim_l = new GraphicLine2D();
      anim_i = new GraphicInteger();
      FIRST = false;
    }
    anim_c->set_all(w, FOREGROUND, BACKGROUND, 0, 0, false);
    anim_s->set_all(w, FOREGROUND, BACKGROUND, 3, 0, false);
    anim_r->set_all(w, FOREGROUND, BACKGROUND, 3, 0, false);
    anim_l->set_all(w, FOREGROUND, BACKGROUND, 3, 0, false);
}
			
/*
 ********************************/
			 
// brute force

/*
 * Segments
 *
 */
 
int all_intersections_counting_bf(list(Segment2D_ptr) &segments, World *w, int animation_depth, int reference_depth) {
    list_item it_s, it_intersect_s;
    int res = 0;			// O(n^2) algorithm

    ANIMATION_BEGIN(w, animation_depth, reference_depth)
      intersections_animation_init(w);      
    ANIMATION_END(w, animation_depth, reference_depth)
    
    if (segments.size() > 1)
      forall_list_items(it_s, segments) {
      	
      	ANIMATION_BEGIN(w, animation_depth, reference_depth)
      	  *anim_segment = *(segments[it_s]);
      	  anim_s->draw(w, XorMode);
      	ANIMATION_END(w, animation_depth, reference_depth)
      	
      	it_intersect_s = segments.succ(it_s);
      	while (it_intersect_s) {
      		
      	  ANIMATION_BEGIN(w, animation_depth, reference_depth)
       	    *anim_segment = *(segments[it_intersect_s]);
      	    anim_s->draw(w, XorMode);
      	    frame_stop_mark(w);
      	    anim_s->draw(w, XorMode);
      	  ANIMATION_END(w, animation_depth, reference_depth)
      	  
          if (segments[it_s]->intersect(*segments[it_intersect_s])) {
            res++;
            
            ANIMATION_BEGIN(w, animation_depth, reference_depth)
              anim_integer->set(res);
              anim_i->draw(w, XorMode);
            ANIMATION_END(w, animation_depth, reference_depth)
            
          }
       	  it_intersect_s = segments.succ(it_intersect_s);
        }

      	ANIMATION_BEGIN(w, animation_depth, reference_depth)
      	  *anim_segment = *(segments[it_s]);
      	  anim_s->draw(w, XorMode);
      	ANIMATION_END(w, animation_depth, reference_depth)
      }
    
    return res;
}

list(Point2D_ptr) all_intersections_report_bf(list(Segment2D_ptr) &segments, World *w, int animation_depth, int reference_depth) {
    list_item it_s, it_intersect_s;
    list(Point2D_ptr) res;		// O(n^2) algorithm

    ANIMATION_BEGIN(w, animation_depth, reference_depth)
      intersections_animation_init(w);      
    ANIMATION_END(w, animation_depth, reference_depth)

    if (segments.size() > 1)
      forall_list_items(it_s, segments) {

      	ANIMATION_BEGIN(w, animation_depth, reference_depth)
      	  *anim_segment = *(segments[it_s]);
      	  anim_s->draw(w, XorMode);
      	ANIMATION_END(w, animation_depth, reference_depth)
      	
      	it_intersect_s = segments.succ(it_s);
      	while (it_intersect_s) {

      	  ANIMATION_BEGIN(w, animation_depth, reference_depth)
       	    *anim_segment = *(segments[it_intersect_s]);
      	    anim_s->draw(w, XorMode);
      	    frame_stop_mark(w);
      	    anim_s->draw(w, XorMode);
      	  ANIMATION_END(w, animation_depth, reference_depth)

          if (segments[it_s]->intersect(*segments[it_intersect_s])) {
            res.push(new Point2D(segments[it_s]->intersection(*segments[it_intersect_s])));
            
            ANIMATION_BEGIN(w, animation_depth, reference_depth)
              anim_circle->set(res.head()->X(), res.head()->Y(), 5);
              anim_c->draw(w, XorMode);
            ANIMATION_END(w, animation_depth, reference_depth)
            
          }
       	  it_intersect_s = segments.succ(it_intersect_s);
        }

      	ANIMATION_BEGIN(w, animation_depth, reference_depth)
      	  *anim_segment = *(segments[it_s]);
      	  anim_s->draw(w, XorMode);
      	ANIMATION_END(w, animation_depth, reference_depth)

      }
    
    return res;
}

/*
 * Rays
 *
 */
 
int all_intersections_rays_counting_bf(list(Ray2D_ptr) &rays, World *w, int animation_depth, int reference_depth) {
    list_item it_r, it_intersect_r;
    int res = 0;			// O(n^2) algorithm

    ANIMATION_BEGIN(w, animation_depth, reference_depth)
      intersections_animation_init(w);      
    ANIMATION_END(w, animation_depth, reference_depth)

    if (rays.size() > 1)
      forall_list_items(it_r, rays) {

      	ANIMATION_BEGIN(w, animation_depth, reference_depth)
      	  *anim_ray = *(rays[it_r]);
      	  anim_r->draw(w, XorMode);
      	ANIMATION_END(w, animation_depth, reference_depth)

      	it_intersect_r = rays.succ(it_r);
      	while (it_intersect_r) {

      	  ANIMATION_BEGIN(w, animation_depth, reference_depth)
       	    *anim_ray = *(rays[it_intersect_r]);
      	    anim_r->draw(w, XorMode);
      	    frame_stop_mark(w);
      	    anim_r->draw(w, XorMode);
      	  ANIMATION_END(w, animation_depth, reference_depth)
      	  
          if (rays[it_r]->intersect(*rays[it_intersect_r])) {
            res++;

            ANIMATION_BEGIN(w, animation_depth, reference_depth)
              anim_integer->set(res);
              anim_i->draw(w, XorMode);
            ANIMATION_END(w, animation_depth, reference_depth)
            
          }
       	  it_intersect_r = rays.succ(it_intersect_r);
        }

      	ANIMATION_BEGIN(w, animation_depth, reference_depth)
      	  *anim_ray = *(rays[it_r]);
      	  anim_r->draw(w, XorMode);
      	ANIMATION_END(w, animation_depth, reference_depth)

      }
    
    return res;
}

list(Point2D_ptr) all_intersections_rays_report_bf(list(Ray2D_ptr) &rays, World *w, int animation_depth, int reference_depth) {
    list_item it_r, it_intersect_r;
    list(Point2D_ptr) res;		// O(n^2) algorithm

    ANIMATION_BEGIN(w, animation_depth, reference_depth)
      intersections_animation_init(w);      
    ANIMATION_END(w, animation_depth, reference_depth)

    if (rays.size() > 1)
      forall_list_items(it_r, rays) {

      	ANIMATION_BEGIN(w, animation_depth, reference_depth)
      	  *anim_ray = *(rays[it_r]);
      	  anim_r->draw(w, XorMode);
      	ANIMATION_END(w, animation_depth, reference_depth)

      	it_intersect_r = rays.succ(it_r);
      	while (it_intersect_r) {

      	  ANIMATION_BEGIN(w, animation_depth, reference_depth)
       	    *anim_ray = *(rays[it_intersect_r]);
      	    anim_r->draw(w, XorMode);
      	    frame_stop_mark(w);
      	    anim_r->draw(w, XorMode);
      	  ANIMATION_END(w, animation_depth, reference_depth)

          if (rays[it_r]->intersect(*rays[it_intersect_r])) {
            res.push(new Point2D(rays[it_r]->intersection(*rays[it_intersect_r])));

            ANIMATION_BEGIN(w, animation_depth, reference_depth)
              anim_circle->set(res.head()->X(), res.head()->Y(), 5);
              anim_c->draw(w, XorMode);
            ANIMATION_END(w, animation_depth, reference_depth)
            
          }
       	  it_intersect_r = rays.succ(it_intersect_r);
        }

      	ANIMATION_BEGIN(w, animation_depth, reference_depth)
      	  *anim_ray = *(rays[it_r]);
      	  anim_r->draw(w, XorMode);
      	ANIMATION_END(w, animation_depth, reference_depth)

      }
    
    return res;
}

/*
 * Lines
 *
 */
 
int all_intersections_lines_counting_bf(list(Line2D_ptr) &lines, World *w, int animation_depth, int reference_depth) {
    list_item it_l, it_intersect_l;
    int res = 0;			// O(n^2) algorithm

    ANIMATION_BEGIN(w, animation_depth, reference_depth)
      intersections_animation_init(w);      
    ANIMATION_END(w, animation_depth, reference_depth)

    if (lines.size() > 1)
      forall_list_items(it_l, lines) {

      	ANIMATION_BEGIN(w, animation_depth, reference_depth)
      	  *anim_line = *(lines[it_l]);
      	  anim_l->draw(w, XorMode);
      	ANIMATION_END(w, animation_depth, reference_depth)

      	it_intersect_l = lines.succ(it_l);
      	while (it_intersect_l) {

      	  ANIMATION_BEGIN(w, animation_depth, reference_depth)
       	    *anim_line = *(lines[it_intersect_l]);
      	    anim_l->draw(w, XorMode);
      	    frame_stop_mark(w);
      	    anim_l->draw(w, XorMode);
      	  ANIMATION_END(w, animation_depth, reference_depth)
      	  
          if (lines[it_l]->intersect(*lines[it_intersect_l])) {
            res++;

            ANIMATION_BEGIN(w, animation_depth, reference_depth)
              anim_integer->set(res);
              anim_i->draw(w, XorMode);
            ANIMATION_END(w, animation_depth, reference_depth)

          }
       	  it_intersect_l = lines.succ(it_intersect_l);
        }

      	ANIMATION_BEGIN(w, animation_depth, reference_depth)
      	  *anim_line = *(lines[it_l]);
      	  anim_l->draw(w, XorMode);
      	ANIMATION_END(w, animation_depth, reference_depth)

      }
    
    return res;
}

list(Point2D_ptr) all_intersections_lines_report_bf(list(Line2D_ptr) &lines, World *w, int animation_depth, int reference_depth) {
    list_item it_l, it_intersect_l;
    list(Point2D_ptr) res;		// O(n^2) algorithm

    ANIMATION_BEGIN(w, animation_depth, reference_depth)
      intersections_animation_init(w);      
    ANIMATION_END(w, animation_depth, reference_depth)

    if (lines.size() > 1)
      forall_list_items(it_l, lines) {

      	ANIMATION_BEGIN(w, animation_depth, reference_depth)
      	  *anim_line = *(lines[it_l]);
      	  anim_l->draw(w, XorMode);
      	ANIMATION_END(w, animation_depth, reference_depth)

      	it_intersect_l = lines.succ(it_l);
      	while (it_intersect_l) {

      	  ANIMATION_BEGIN(w, animation_depth, reference_depth)
       	    *anim_line = *(lines[it_intersect_l]);
      	    anim_l->draw(w, XorMode);
      	    frame_stop_mark(w);
      	    anim_l->draw(w, XorMode);
      	  ANIMATION_END(w, animation_depth, reference_depth)
      	  
          if (lines[it_l]->intersect(*lines[it_intersect_l])) {
            res.push(new Point2D(lines[it_l]->intersection(*lines[it_intersect_l])));

            ANIMATION_BEGIN(w, animation_depth, reference_depth)
              anim_circle->set(res.head()->X(), res.head()->Y(), 5);
              anim_c->draw(w, XorMode);
            ANIMATION_END(w, animation_depth, reference_depth)
            
          }
       	  it_intersect_l = lines.succ(it_intersect_l);
        }

      	ANIMATION_BEGIN(w, animation_depth, reference_depth)
      	  *anim_line = *(lines[it_l]);
      	  anim_l->draw(w, XorMode);
      	ANIMATION_END(w, animation_depth, reference_depth)

      }
    
    return res;
}






