/*
    Copyright (C) 1992-1994
                    Welson R. Jacometti

    All rights reserved.

    This program is  free software; you can redistribute unmodified
    versions of it as long as you do not  profit from that, neither
    by selling it nor  by   including it  with other products   for
    sale.  This copyright  notice  must  remain in all  distributed
    copies.

    You are  allowed to modify your  personal copy  of this program
    (as long as you do not distribute any changed versions).

    You may,  of course, distribute  along with it, your own source
    code or libraries of algorithms that run under GeoLab.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

    If you must contact the authors, you may write to:
                     geolab-request@dcc.unicamp.br
*/

/*
 *
 * move_copy.C - modulo de tratamento de eventos no modo de copia ou de
 *               movimentacao de objetos.
 *
 *
 *    	Data	Hora  	Nome     		Modificacao/Motivo
 *
 *    12.12.91  08:45 	Welson R. Jacometti  	Arquivo criado
 *
 */

#include "move_copy.h"
#include "Obj.h"
#include "color.h"
#include "objects.h"

#define NO_IF_GLUED true

bool move_copy_events(World *world, Xv_Window, Event *event) {
    static bool idle = true;
    static GraphicObj *current_obj = 0;
    static double xi = 0, yi = 0;
    double X = world->virtual_xcoord(event_x(event));
    double Y = world->virtual_ycoord(event_y(event));
    double dx, dy;

    dx = X - xi;
    dy = Y - yi;

    switch (event_action(event)) {
      case ACTION_SELECT:
      case ACTION_ADJUST:
        if (event_is_down(event)) {
          if (idle) {
            if (objects_something_selected_at(X, Y, NO_IF_GLUED)) {
              if (event_action(event) == ACTION_ADJUST)
                (void)objects_duplicate_selected(NO_IF_GLUED);
              objects_translate_start(world, NO_IF_GLUED);
              idle = false;
            }
            else
              if (current_obj = objects_object_selected_at(X, Y, NO_IF_GLUED)) {
              	if (event_action(event) == ACTION_ADJUST) {
                  GraphicObj *dupl = current_obj->reproduct();
              	  if (dupl)
              	    objects_insert(dupl);
              	}
              	unsigned long fg, bg;
              	current_obj->get_color(fg, bg);
              	current_obj->set_color(world, get_pixel_from_value(WHITE), get_pixel_from_value(WHITE));
              	current_obj->draw(world, CopyMode);
              	objects_repaint_all_but(current_obj, world);
              	current_obj->set_color(world, fg, bg);
              	current_obj->draw(world, XorMode);
              	idle = false;
              }
          }
        }
        else 
          if (!idle) {
            idle = true;
            if (current_obj) {
              current_obj->draw(world, XorMode);
              current_obj->translate(dx, dy);
              current_obj->draw(world, CopyMode);
              current_obj = 0;
            }
            else
              objects_translate_end(world, dx, dy, NO_IF_GLUED);
          }
        break;
        
      case LOC_DRAG:
        if (!idle)
          if (current_obj) {
            current_obj->draw(world, XorMode);
            current_obj->translate(dx, dy);
            current_obj->draw(world, XorMode);
          }
          else
            objects_translate_change(world, dx, dy, NO_IF_GLUED);
        break;
    }
    
    xi = X;
    yi = Y;
        
    return false;
}
