/*
    Copyright (C) 1992-1994
                    Cesar N. Gon

    All rights reserved.

    This program is  free software; you can redistribute unmodified
    versions of it as long as you do not  profit from that, neither
    by selling it nor  by   including it  with other products   for
    sale.  This copyright  notice  must  remain in all  distributed
    copies.

    You are  allowed to modify your  personal copy  of this program
    (as long as you do not distribute any changed versions).

    You may,  of course, distribute  along with it, your own source
    code or libraries of algorithms that run under GeoLab.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

    If you must contact the authors, you may write to:
                     geolab-request@dcc.unicamp.br
*/

/*
 * rotate.h - Rota,c~ao de objetos
 *
 *
 *    	Data	Hora  	Nome     		Modificacao/Motivo
 *
 *    21.05.93  12:00 	Cesar N. Gon 	  	Arquivo criado
 *
 */

#ifndef _ROTATE_DEFINED
#define _ROTATE_DEFINED

#include "External.h"		// obrigatorio a todo modulo externo
#include "Algorithm.h"
#include "Point2D.h"
#include "Ray2D.h"
#include "line_settings.h"


#define MIN_E_DISTANCE_SQR 9

icon_bits rotate_icon[] = {
#include "ICONS/rotate.icon"
};


class rotate : public ExternalFunctionMode {
  public:

    ResultObjects binder(const SelectedObjects &, const GraphicObjects &, World *, int, int);
    
    bool is_able_to_activate(const SelectedObjects &, const GraphicObjects &, World *) { return true; }
    
    void activator(World *w) {
	ray.set_all(w, FOREGROUND, BACKGROUND, 0, 0, false);
	ray0.set_all(w, FOREGROUND, BACKGROUND, 0, get_line_style_from_index(DASHED), false);
    }
    
    void deactivator(World *w) {
	if (start) {
	    ray0.draw(w, XorMode);
	    ray.draw(w, XorMode);
	}
    }

    bool click_select_down(World *w, double x, double y) {
	start = false;
	selectdown = true;
	ref.setXY(x,y);
	return true;
    }

    bool click_select_up(World *w, double x, double y) {
	if (start) {
	    ray0.draw(w, XorMode);
	    ray.draw(w, XorMode);
	    dir1 = dir;
	    dir.setXY(x,y);
	    EF_exec_binder(w);
	    start = false;
	}
	selectdown = false; 
	return true;
    }

    bool loc_drag(World *w, double x, double y) {
        if (selectdown) {
	    dir1 = dir;
	    dir.setXY(x,y);
	    if (!start) {
		if (ref.e_distance_sqr(dir) > MIN_E_DISTANCE_SQR) {
		    dir0 = dir1 = dir;
		    ((Ray2D *)ray0.obj)->set(ref,dir);
		    ((Ray2D *)ray.obj)->set(ref,dir);
		    ray0.draw(w, XorMode);
		    setangle();
		    ray.draw(w, XorMode);
		    start = true;
		}
	    }
	    else {
		ray.draw(w, XorMode);
		((Ray2D *)ray.obj)->set(ref,dir);
		setangle();
		ray.draw(w, XorMode);
		EF_exec_binder(w);
	    }
	}
	return true;
    }

    icon_bits *icon() {return rotate_icon;}

  private:
    bool start;
    bool selectdown;
    Point2D ref, dir0, dir1, dir;
    GraphicRay2D ray0, ray;

    void setangle() {
	static char sang[14];
	double angle = (ref.angle(dir)-ref.angle(dir0))*180/PI;
	sprintf(sang,"%.1f degrees",angle >= 0 ? angle : 360+angle);
    	ray.set_string(sang);
    }
    
};

Algorithm* rotate_link();

#endif ~_ROTATE_DEFINED;




