/*
    Copyright (C) 1992-1994
                    Welson R. Jacometti

    All rights reserved.

    This program is  free software; you can redistribute unmodified
    versions of it as long as you do not  profit from that, neither
    by selling it nor  by   including it  with other products   for
    sale.  This copyright  notice  must  remain in all  distributed
    copies.

    You are  allowed to modify your  personal copy  of this program
    (as long as you do not distribute any changed versions).

    You may,  of course, distribute  along with it, your own source
    code or libraries of algorithms that run under GeoLab.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

    If you must contact the authors, you may write to:
                     geolab-request@dcc.unicamp.br
*/

/*
 * three_points_circle - Algoritmo que produz um circulo dados tres pontos.
 *
 *
 * Welson R. Jacometti
 *
 */
 
#include "three_points_circle.h"

Obj* circle_by_three_points( Point2D &a,  Point2D &b,  Point2D &c) {
    if (a.colinear(b, c)) {
      Line2D *l = new Line2D();
      l->set(a, b);
      return l;
    }
    else {
      Point2D cc = circumcentre(a, b, c);
      Circle *circle = new Circle();
      circle->set(cc, cc.e_distance(a));
      return circle;
    }
}

bool three_points_circle::is_able_to_handle(const SelectedObjects &SO, const GraphicObjects &GO) {
    list_item it;
    int np = 0;
    
    forall(it, SO) {
      ++np;
      if (GO[it]->type() != Point2D_t || np > 3) {
        SO.init_iterator();
        return false;
      }
    }

    if (np < 3)
      return false;

    return true;
}

ResultObjects three_points_circle::binder(const SelectedObjects &SO, const GraphicObjects &GO, World *, int, int) {
    ResultObjects result;
    if (!SO.empty()) {			// relies in GeoLab disabling
      list_item it = SO.first();
      Point2D a, b, c;
      a = *((Point2D *)GO[SO[it]]->obj);
      b = *((Point2D *)GO[SO[SO.succ(it)]]->obj);
      c = *((Point2D *)GO[SO[SO.succ(SO.succ(it))]]->obj);
      result.push(circle_by_three_points(a, b, c));
    }
    return result;
}
    
Algorithm* three_points_circle_link() {
    return new three_points_circle();
}


